//
//  gfx.c
//  FF6Tools
//
//  Created by Evan Owen on 12/7/14.
//  Copyright (c) 2014 Evan Owen. All rights reserved.
//

// gfx_render: for simple graphics, function to draw 8-bit indexed graphics with a 32-bit rgb palette
//     input: byte* graphics, dword* palette, int height, int width
//     output: dword* image
// gfx_decode: function to convert any tile format to 8-bit indexed OR convert any palette format to 32-bit rgb
//     input: void* src, int length (in bytes), int format
//     output: void* dest
// gfx_encode: function to convert 8-bit indexed to any tile format OR convert 32-bit rgb to any palette format
//     input: void* src, int length (in bytes), int format
//     output: void* dest
// gfx_render_ppu: for layered graphics
//     input: PPU ppu
//     output: dword* graphics
// PPU structure
//     layer structure [4]
//         height/width (int)
//         vertical/horizontal scroll position (int)
//         8-bit indexed graphics (byte*)
//         tile data for each layer (word*)
//         enabled in main/sub-screen (bool)
//         addition/subtraction enabled (bool)
//     32-bit rgb palette (dword*)
//     addition/subtraction toggle (bool)
//     half-add/sub toggle (bool)

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "gfx.h"

typedef unsigned char  uint8;
typedef unsigned short uint16;
typedef unsigned int   uint32;

#define MAKELONG(a,b) (dword)((a & 0xFFFF) | ((b & 0xFFFF) << 16))
#define MAKEWORD(a,b) (word)((a & 0xFF) | ((b & 0xFF) << 8))
#define HIWORD(a) (word)((a >> 16) & 0xFFFF)
#define LOWORD(a) (word)(a & 0xFFFF)
#define HIBYTE(a) (byte)((a >> 8) & 0xFF)
#define LOBYTE(a) (byte)(a & 0xFF)
#define HINYBBLE(a) (byte)((a >> 4) & 0x0F)
#define LONYBBLE(a) (byte)(a & 0x0F)
#define BANKBYTE(a) (byte)((a >> 16) & 0xFF)

void render_tile(uint32* dest, const uint8* gfx, const uint32* pal, int ppl) {

    uint32* pixel_ptr;
    uint8 color;
    
    for (int line = 0; line < 8; line++) {
        pixel_ptr = dest;
        color = *gfx++; if (color) *pixel_ptr = pal[color]; pixel_ptr++;
        color = *gfx++; if (color) *pixel_ptr = pal[color]; pixel_ptr++;
        color = *gfx++; if (color) *pixel_ptr = pal[color]; pixel_ptr++;
        color = *gfx++; if (color) *pixel_ptr = pal[color]; pixel_ptr++;
        color = *gfx++; if (color) *pixel_ptr = pal[color]; pixel_ptr++;
        color = *gfx++; if (color) *pixel_ptr = pal[color]; pixel_ptr++;
        color = *gfx++; if (color) *pixel_ptr = pal[color]; pixel_ptr++;
        color = *gfx++; if (color) *pixel_ptr = pal[color]; pixel_ptr++;
        dest += ppl;
    }
}

uint32 math_none(uint32 c1, uint32 c2) {
    return c1;
}

uint32 math_add(uint32 c1, uint32 c2) {
    int b1 = c1 & 0x000000FF;
    int g1 = c1 & 0x0000FF00;
    int r1 = c1 & 0x00FF0000;
    int b2 = c2 & 0x000000FF;
    int g2 = c2 & 0x0000FF00;
    int r2 = c2 & 0x00FF0000;
    int b = b1 + b2; if (b > 0x000000FF) b = 0x000000FF;
    int g = g1 + g2; if (g > 0x0000FF00) g = 0x0000FF00;
    int r = r1 + r2; if (r > 0x00FF0000) r = 0x00FF0000;
    int a = 0xFF000000;
    return b | g | r | a;
}

uint32 math_half_add(uint32 c1, uint32 c2) {
    if (c2 == 0) {
        return c1;
    }
    int c = ((c1 & 0x00FEFEFE) + (c2 & 0x00FEFEFE)) >> 1;
    c += (c1 & c2 & 0x00010101);
    c += 0xFF000000;
    return c;
}

uint32 math_subtract(uint32 c1, uint32 c2) {
    int c = 0;
    int b1 = c1 & 0x000000FF;
    int g1 = c1 & 0x0000FF00;
    int r1 = c1 & 0x00FF0000;
    int a1 = c1 & 0xFF000000;
    int b2 = c2 & 0x000000FF;
    int g2 = c2 & 0x0000FF00;
    int r2 = c2 & 0x00FF0000;
    if (r1 > r2) c += (r1 - r2);
    if (g1 > g2) c += (g1 - g2);
    if (b1 > b2) c += (b1 - b2);
    c += a1;
    return c;
}

uint32 math_half_subtract(uint32 c1, uint32 c2) {
    int b1 = c1 & 0x000000FF;
    int g1 = c1 & 0x0000FF00;
    int r1 = c1 & 0x00FF0000;
    int a1 = c1 & 0xFF000000;
    int b2 = (c2 & 0x000000FE) >> 1;
    int g2 = (c2 & 0x0000FE00) >> 1;
    int r2 = (c2 & 0x00FE0000) >> 1;
    int b = b1 - b2; if (b < 0x00000000) b = 0;
    int g = g1 - g2; if (g < 0x00000100) g = 0;
    int r = r1 - r2; if (r < 0x00010000) r = 0;
    int a = a1;
    return b | g | r | a;
}

#define DRAW_PIXEL \
    if (z > z_buffer[pixel_ptr]) { \
        color_index = *gfx; \
        if (color_index) { \
            _color = pal[color_index];\
            if (_color) {\
                sub_color = sub[pixel_ptr]; \
                if (sub_color) { \
                    dest[pixel_ptr] = math(_color, sub_color); \
                } else { \
                    dest[pixel_ptr] = _color; \
                } \
                z_buffer[pixel_ptr] = z; \
            } \
        } \
    } \
    gfx++; \
    pixel_ptr++;

#define DRAW_PIXEL_H \
    if (z > z_buffer[pixel_ptr]) { \
        color_index = *gfx; \
        if (color_index) { \
            _color = pal[color_index];\
            if (_color) {\
                sub_color = sub[pixel_ptr]; \
                if (sub_color) { \
                    dest[pixel_ptr] = math(_color, sub_color); \
                } else { \
                    dest[pixel_ptr] = _color; \
                } \
                z_buffer[pixel_ptr] = z; \
            } \
        } \
    } \
    gfx++; \
    pixel_ptr--;

void render_tile_vh(uint32* dest, uint32* sub, uint32* z_buffer, uint32 (*math)(uint32, uint32), const uint8* gfx, const uint32* pal, int ppl, int z, bool vFlip, bool hFlip) {
    
    int line_ptr = vFlip ? 7 * ppl : 0;
    int pixel_ptr;
    uint8 color_index;
    uint32 _color;
    uint32 sub_color;
    
    if (hFlip) {
        for (int line = 0; line < 8; line++) {
            pixel_ptr = line_ptr + 7;
            DRAW_PIXEL_H
            DRAW_PIXEL_H
            DRAW_PIXEL_H
            DRAW_PIXEL_H
            DRAW_PIXEL_H
            DRAW_PIXEL_H
            DRAW_PIXEL_H
            DRAW_PIXEL_H
            line_ptr += (vFlip ? -ppl : ppl);
        }
    } else {
        for (int line = 0; line < 8; line++) {
            pixel_ptr = line_ptr;
            DRAW_PIXEL
            DRAW_PIXEL
            DRAW_PIXEL
            DRAW_PIXEL
            DRAW_PIXEL
            DRAW_PIXEL
            DRAW_PIXEL
            DRAW_PIXEL
            line_ptr += (vFlip ? -ppl : ppl);
        }
    }
}

void gfx_render(void* dest, const void* gfx, const void* pal, int length, int ppl) {
    const uint8* gfx_ptr = gfx;
    const uint8* end = gfx + length;
    uint32* dest_ptr = dest;
    
    int pixel = 0;
    
    while (gfx_ptr < end) {
        render_tile(dest_ptr + pixel, gfx_ptr, pal, ppl);
        gfx_ptr += 64;
        pixel += 8;
        if (pixel >= ppl) {
            pixel = 0;
            dest_ptr += 8 * ppl;
        }
    }
}

void render_layer(void* dest, uint32* sub, uint32* z_buffer, uint32 (*math)(uint32, uint32), struct Layer layer, const uint32* pal, int length, int ppl) {
    uint32* dest_ptr = dest;
    uint32* end = dest + length;
    const uint16* tile_ptr = (uint16*)layer.tiles;
    const uint32* tile_ptr32 = (uint32*)layer.tiles;
    
    uint16 tile_data = 0;
    uint32 tile_data32 = 0;
    const uint8* gfx_ptr = NULL;
    const uint32* pal_ptr = NULL;
    bool vFlip = false;
    bool hFlip = false;
    int z_level = 0;
    
    int pixel = 0;
    int z = 0;
    int row = layer.v_scroll;
    while (row > 0) row -= layer.height;
    while (row < 0) row += layer.height;
    int column = layer.h_scroll;
    while (column > 0) column -= layer.width;
    while (column < 0) column += layer.width;
    
    while (dest_ptr < end) {
        switch (layer.format) {
            case GBA_4BPP_TILE:
                tile_data = tile_ptr[column + row * layer.width];
                gfx_ptr = layer.gfx + ((tile_data & 0x03FF) << 6);
                pal_ptr = pal + ((tile_data & 0x7000) >> 8);
                z_level = (tile_data & 0x8000) >> 15;
                vFlip = tile_data & 0x0800;
                hFlip = tile_data & 0x0400;
                break;
                
            case GBA_2BPP_TILE:
                tile_data = tile_ptr[column + row * layer.width];
                gfx_ptr = layer.gfx + ((tile_data & 0x03FF) << 6);
                pal_ptr = pal + ((tile_data & 0x7000) >> 10);
                z_level = (tile_data & 0x8000) >> 15;
                vFlip = tile_data & 0x0800;
                hFlip = tile_data & 0x0400;
                break;

            case SNES_4BPP_TILE:
                tile_data = tile_ptr[column + row * layer.width];
                gfx_ptr = layer.gfx + ((tile_data & 0x03FF) << 6);
                pal_ptr = pal + ((tile_data & 0x1C00) >> 6);
                z_level = (tile_data & 0x2000) >> 13;
                vFlip = tile_data & 0x8000;
                hFlip = tile_data & 0x4000;
                break;
                
            case SNES_2BPP_TILE:
                tile_data = tile_ptr[column + row * layer.width];
                gfx_ptr = layer.gfx + ((tile_data & 0x03FF) << 6);
                pal_ptr = pal + ((tile_data & 0x1C00) >> 8);
                z_level = (tile_data & 0x2000) >> 13;
                vFlip = tile_data & 0x8000;
                hFlip = tile_data & 0x4000;
                break;

            case SNES_MODE7_TILE:
                tile_data = tile_ptr[column + row * layer.width];
                gfx_ptr = layer.gfx + ((tile_data & 0x00FF) << 6);
                pal_ptr = pal + ((tile_data & 0x0F00) >> 4);
                vFlip = tile_data & 0x8000;
                hFlip = tile_data & 0x4000;
                break;

            case ROMTOOLS_TILE: // vh--zzzz pppppppp mmmmmmmm mmmmmmmm
                tile_data32 = tile_ptr32[column + row * layer.width];
                gfx_ptr = layer.gfx + ((tile_data32 & 0xFFFF) << 6);
                pal_ptr = pal + ((tile_data32 >> 16) & 0xFF);
                z_level = (tile_data32 >> 24) & 0xF;
                hFlip = (tile_data32 >> 30) & 1;
                vFlip = (tile_data32 >> 31) & 1;
                break;

            default:
                printf("Format %d not supported\n", layer.format);
                return;
        }
        z = layer.z[z_level];
        render_tile_vh(dest_ptr + pixel, sub + pixel, z_buffer + pixel, math, gfx_ptr, pal_ptr, ppl, z, vFlip, hFlip);
        pixel += 8;
        column++;
        if (column >= layer.width) column = 0;
        if (pixel >= ppl) {
            pixel = 0;
            dest_ptr += ppl * 8;
            sub += ppl * 8;
            z_buffer += ppl * 8;
            column = layer.h_scroll;
            while (column > 0) column -= layer.width;
            while (column < 0) column += layer.width;
            row++;
            if (row >= layer.height) row = 0;
        }
    }
}

void render_back_area(void* dest, uint32 color, int length) {
    uint32* ptr = (uint32*)dest;
    uint32* end = (uint32*)(dest + length);
    while (ptr < end) {
        *ptr++ = color;
    }
}

void flip_layer(void* dest, int length, int ppl) {
    uint32* dest_ptr = dest;
    uint32* end = (uint32*)(dest + length);
    void* buffer = malloc(length);
    uint32* buf_ptr = (uint32*)(buffer + length) - ppl;
    
    while (dest_ptr < end) {
        memcpy(buf_ptr, dest_ptr, ppl * sizeof(uint32));
        dest_ptr += ppl;
        buf_ptr -= ppl;
    }
    memcpy(dest, buffer, length);

    free(buffer);
}

void gfx_render_ppu(void* dest, PPU ppu, int length, int ppl) {
    
    uint32* sub = malloc(length);
    uint32* z_buffer = malloc(length);
    memset(sub, 0, length);
    memset(z_buffer, 0, length);
    
    // render the back area
    if (ppu.back_area) {
        render_back_area(dest, *(uint32*)ppu.pal, length);
    }
    
    // render subscreen layers
    for (int layer = 0; layer < 4; layer++) {
        if (ppu.layers[layer].sub) {
            render_layer(sub, sub, z_buffer, math_none, ppu.layers[layer], ppu.pal, length, ppl);
        }
    }

    // function pointer for color math
    uint32 (*math)(uint32, uint32) = math_none;
    
    // clear the z-buffer
    memset(z_buffer, 0, length);

    // render main screen layers
    for (int layer = 0; layer < 4; layer++) {
        if (ppu.layers[layer].main) {
            if (ppu.layers[layer].math) {
                if (ppu.subtract) {
                    if (ppu.half)   math = math_half_subtract;
                    else            math = math_subtract;
                } else {
                    if (ppu.half)   math = math_half_add;
                    else            math = math_add;
                }
            } else {
                math = math_none;
            }
            render_layer(dest, sub, z_buffer, math, ppu.layers[layer], ppu.pal, length, ppl);
        }
    }
    
    if (ppu.flipped) {
        flip_layer(dest, length, ppl);
    }
    
    free(sub);
    free(z_buffer);
}

//void gfx_render_ppu_hdma(void* dest, PPU ppu, int length, int ppl) {
//    uint32* dest_ptr = dest;
//    uint32* end = (uint32*)(dest + length);
//    
//    while (dest_ptr < end) {
//        // update hdma channels
//        for (int channel = 0; channel < 8; channel++) {
//            struct HDMA hdma = ppu.hdma[channel];
//            if (hdma.enabled) {
//                
//                if (hdma.line_counter == 0) {
//                    // line counter has expired
//                    hdma.line_counter = *(uint8*)hdma.data++;
//                    
//                } else {
//                    // decrement line counter
//                    hdma.line_counter--;
//                }
//            }
//        }
//        
//        // render each line separately
//        gfx_render_ppu(dest_ptr, ppu, ppl * sizeof(uint32) * 8, ppl);
//        dest_ptr += ppl * 8;
//        
//        // increase vertical scroll for each layer
//        for (int layer = 0; layer < 4; layer++) {
//            ppu.layers[layer].v_scroll++;
//        }
//    }
//}

void decode_linear2bpp(void* dest, const void* src, int length) {
    const uint8* src_ptr = src;
    const uint8* end = src + length;
    uint8* dest_ptr = dest;
    
    uint8 value;
    
    while (src_ptr < end) {
        value = *src_ptr++;
        *dest_ptr++ = value & 0x03;
        value >>= 2;
        *dest_ptr++ = value & 0x03;
        value >>= 2;
        *dest_ptr++ = value & 0x03;
        value >>= 2;
        *dest_ptr++ = value & 0x03;
    }
}

void decode_linear4bpp(void* dest, const void* src, int length) {
    const uint8* src_ptr = src;
    const uint8* end = src + length;
    uint8* dest_ptr = dest;
    
    uint8 value;
    
    while (src_ptr < end) {
        value = *src_ptr++;
        *dest_ptr++ = value & 0x0F;
        value >>= 4;
        *dest_ptr++ = value & 0x0F;
    }
}

void decode_linear8bpp(void* dest, const void* src, int length) {
    memcpy(dest, src, length);
//    const uint8* src_ptr = src;
//    const uint8* end = src + length;
//    uint8* dest_ptr = dest;
//    
//    uint8 value;
//    
//    while (src_ptr < end) {
//        value = *src_ptr++;
//        *dest_ptr++ = value;
//    }
}

void decode_nes1bpp(void* dest, const void* src, int length) {
    const uint8* src_ptr = src;
    const uint8* end = src + length;
    uint8* dest_ptr = dest;
    
    uint8 bitplanes;
    uint8 value;
    
    while (src_ptr < end) {
        for (int row = 0; row < 8; row++) {
            bitplanes = *src_ptr++;
            for (int bit = 0; bit < 8; bit++) {
                value = bitplanes & 0x80;
                value >>= 7;
//                value &= 0x01;
                *dest_ptr++ = value;
                bitplanes <<= 1;
            }
        }
    }
}

void decode_nes2bpp(void* dest, const void* src, int length) {
    const uint8* src_ptr = src;
    const uint8* bp1_ptr;
    const uint8* bp2_ptr;
    const uint8* end = src + length;
    uint8* dest_ptr = dest;
    
    uint16 bitplanes;
    uint16 value;
    
    while (src_ptr < end) {
        bp1_ptr = src_ptr;
        bp2_ptr = src_ptr + 8;
        for (int row = 0; row < 8; row++) {
            bitplanes = MAKEWORD(*bp1_ptr++, *bp2_ptr++);
            for (int bit = 0; bit < 8; bit++) {
                value = bitplanes & 0x8080;
                value >>= 7;
                value |= (value >> 7);
                value &= 0x03;
                *dest_ptr++ = value;
                bitplanes <<= 1;
            }
        }
        src_ptr += 16;
    }
}

void decode_snes2bpp(void* dest, const void* src, int length) {
    const uint16* src_ptr = src;
    const uint16* end = src + length;
    uint8* dest_ptr = dest;
    
    uint16 bitplanes;
    uint16 value;
    
    while (src_ptr < end) {
        for (int row = 0; row < 8; row++) {
            bitplanes = *src_ptr++;
            for (int bit = 0; bit < 8; bit++) {
                value = bitplanes & 0x8080;
                value >>= 7;
                value |= (value >> 7);
                value &= 0x03;
                *dest_ptr++ = value;
                bitplanes <<= 1;
            }
        }
    }
}

void decode_snes3bpp(void* dest, const void* src, int length) {
    const uint16* src_ptr = src;
    const uint16* bp12_ptr;
    const uint8* bp3_ptr;
    const uint16* end = src + length;
    uint8* dest_ptr = dest;
    
    uint32 bitplanes;
    uint32 value;
    
    while (src_ptr < end) {
        bp12_ptr = src_ptr;
        bp3_ptr = (uint8*)(src_ptr + 8);
        for (int row = 0; row < 8; row++) {
            bitplanes = MAKELONG(*bp12_ptr++, *bp3_ptr++);
            for (int bit = 0; bit < 8; bit++) {
                value = bitplanes & 0x808080;
                value >>= 7;
                value |= (value >> 7);
                value |= (value >> 14);
                value &= 0x07;
                *dest_ptr++ = value;
                bitplanes <<= 1;
            }
        }
        src_ptr += 12;
    }
}

void decode_snes4bpp(void* dest, const void* src, int length) {
    const uint16* src_ptr = src;
    const uint16* bp12_ptr;
    const uint16* bp34_ptr;
    const uint16* end = src + length;
    uint8* dest_ptr = dest;

    uint32 bitplanes;
    uint32 value;
    
    while (src_ptr < end) {
        bp12_ptr = src_ptr;
        bp34_ptr = src_ptr + 8;
        for (int row = 0; row < 8; row++) {
            bitplanes = MAKELONG(*bp12_ptr++, *bp34_ptr++);
            for (int bit = 0; bit < 8; bit++) {
                value = bitplanes & 0x80808080;
                value >>= 7;
                value |= (value >> 7);
                value |= (value >> 14);
                value &= 0x0F;
                *dest_ptr++ = value;
                bitplanes <<= 1;
            }
        }
        src_ptr += 16;
    }
}

void decode_snesmode7(void* dest, const void* src, int length) {

    const uint8* src_ptr = src;
    const uint8* end = src + length;
    uint8* dest_ptr = dest;
    
    uint8 value;
    
    while (src_ptr < end) {
        for (int row = 0; row < 8; row++) {
            value = *src_ptr++;
            *dest_ptr++ = value & 0x0F;
            value >>= 4;
            *dest_ptr++ = value;
        }
    }
}

void decode_snesmode7ex(void* dest, const void* src, int length) {
    
    const uint8* src_ptr = src;
    const uint8* end = src + length;
    uint8* dest_ptr = dest;
    
    uint8 value;
    
    while (src_ptr < end) {
        for (int row = 0; row < 8; row++) {
            value = *src_ptr++;
            *dest_ptr = value & 0x0F;
            dest_ptr += 2;
        }
    }
}

//void decode_snes_mode7(void* dest, const void* src, int length) {
//    const uint16* src_ptr = src;
//    const uint16* end = src + length;
//    uint8* dest_ptr = dest;
//    
//    uint32 value;
//    
//    while (src_ptr < end) {
//        value = *src_ptr++;
//        *dest_ptr++ = value & 0x0F;
//        value >>= 4;
//        *dest_ptr++ = value & 0x0F;
//    }
//}

void decode_bgr555(void* dest, const void* src, int length) {
    
    const uint16* src_ptr = src;
    const uint16* end = src + length;
    uint32* dest_ptr = dest;
    
    uint16 bgr555;
    uint32 xrgb8888;
    uint32 r, g, b;
    
    while (src_ptr < end) {
        bgr555 = *src_ptr++;
        r = bgr555 & 0x1F; bgr555 >>= 5;
        g = bgr555 & 0x1F; bgr555 >>= 5;
        b = bgr555 & 0x1F;
        r = (r * 255) / 31;
        g = (g * 255) / 31; g <<= 8;
        b = (b * 255) / 31; b <<= 16;
        xrgb8888 = 0xFF000000 | r | g | b;
        *dest_ptr++ = xrgb8888;
    }
}

void decode_snes_2bpp_tile(void* dest, const void* src, int length) {
    
    const uint16* src_ptr = src;
    const uint16* end = src + length;
    uint32* dest_ptr = dest;
    
    uint16 raw;
    uint32 value;
    
    while (src_ptr < end) {
        raw = *src_ptr++;
        value = 0;
        value |= (uint32)(raw & 0x03FF);
        value |= (uint32)(raw & 0x1C00) << 8;
        value |= (uint32)(raw & 0x2000) << 11;
        value |= (uint32)(raw & 0x4000) << 16;
        value |= (uint32)(raw & 0x8000) << 16;
        *dest_ptr++ = value;
    }
}

void encode_snes_2bpp_tile(void* dest, const void* src, int length) {
    
    const uint32* src_ptr = src;
    const uint32* end = src + length;
    uint16* dest_ptr = dest;
    
    uint32 raw;
    uint16 value;
    
    while (src_ptr < end) {
        raw = *src_ptr++;
        value = 0;
        value |= (uint16)(raw & 0x000003FF);
        value |= (uint16)((raw & 0x001C0000) >> 8);
        value |= (uint16)((raw & 0x01000000) >> 11);
        value |= (uint16)((raw & 0x40000000) >> 16);
        value |= (uint16)((raw & 0x80000000) >> 16);
        *dest_ptr++ = value;
    }
}

void decode_snes_4bpp_tile(void* dest, const void* src, int length) {
    
    const uint16* src_ptr = src;
    const uint16* end = src + length;
    uint32* dest_ptr = dest;
    
    uint16 raw;
    uint32 value;
    
    while (src_ptr < end) {
        raw = *src_ptr++;
        value = 0;
        value |= (uint32)(raw & 0x03FF);
        value |= (uint32)(raw & 0x1C00) << 10;
        value |= (uint32)(raw & 0x2000) << 11;
        value |= (uint32)(raw & 0x4000) << 16;
        value |= (uint32)(raw & 0x8000) << 16;
        *dest_ptr++ = value;
    }
}

void encode_snes_4bpp_tile(void* dest, const void* src, int length) {
    
    const uint32* src_ptr = src;
    const uint32* end = src + length;
    uint16* dest_ptr = dest;
    
    uint32 raw;
    uint16 value;
    
    while (src_ptr < end) {
        raw = *src_ptr++;
        value = 0;
        value |= (uint16)(raw & 0x000003FF);
        value |= (uint16)((raw & 0x00700000) >> 10);
        value |= (uint16)((raw & 0x01000000) >> 11);
        value |= (uint16)((raw & 0x40000000) >> 16);
        value |= (uint16)((raw & 0x80000000) >> 16);
        *dest_ptr++ = value;
    }
}

void decode_snes_sprite_tile(void* dest, const void* src, int length) {
    
    const uint16* src_ptr = src;
    const uint16* end = src + length;
    uint32* dest_ptr = dest;
    
    uint16 raw;
    uint32 value;
    
    while (src_ptr < end) {
        raw = *src_ptr++;
        value = 0;
        value |= (raw & 0x01FF);
        value |= (raw & 0x0E00) << 11;
        value |= (raw & 0x3000) << 12;
        value |= (raw & 0x4000) << 16;
        value |= (raw & 0x8000) << 16;
        *dest_ptr++ = value;
    }
}

void encode_snes_sprite_tile(void* dest, const void* src, int length) {
    
    const uint32* src_ptr = src;
    const uint32* end = src + length;
    uint16* dest_ptr = dest;
    
    uint32 raw;
    uint16 value;
    
    while (src_ptr < end) {
        raw = *src_ptr++;
        value = 0;
        value |= (uint16)(raw & 0x000001FF);
        value |= (uint16)((raw & 0x00700000) >> 11);
        value |= (uint16)((raw & 0x01000000) >> 12);
        value |= (uint16)((raw & 0x40000000) >> 16);
        value |= (uint16)((raw & 0x80000000) >> 16);
        *dest_ptr++ = value;
    }
}

void decode_snes_mode7_tile(void* dest, const void* src, int length) {
    
    const uint8* src_ptr = src;
    const uint8* end = src + length;
    uint32* dest_ptr = dest;
    
    uint8 raw;
    
    while (src_ptr < end) {
        raw = *src_ptr++;
        *dest_ptr++ = (uint32)raw;
    }
}

void encode_snes_mode7_tile(void* dest, const void* src, int length) {
    
    const uint32* src_ptr = src;
    const uint32* end = src + length;
    uint8* dest_ptr = dest;
    
    uint32 raw;
//    uint16 value;
    
    while (src_ptr < end) {
        raw = *src_ptr++;
        *dest_ptr++ = (uint8)raw;
    }
}

void decode_gba_4bpp_tile(void* dest, const void* src, int length) {
    
    const uint16* src_ptr = src;
    const uint16* end = src + length;
    uint32* dest_ptr = dest;
    
    uint16 raw;
    uint32 value;
    
    while (src_ptr < end) {
        raw = *src_ptr++;
        value = 0;
        value |= (uint32)(raw & 0x03FF);
        value |= (uint32)(raw & 0x7000) << 8;
        value |= (uint32)(raw & 0x8000) << 9;
        value |= (uint32)(raw & 0x0400) << 20;
        value |= (uint32)(raw & 0x0800) << 20;
        *dest_ptr++ = value;
    }
}

void encode_gba_4bpp_tile(void* dest, const void* src, int length) {
    
    const uint32* src_ptr = src;
    const uint32* end = src + length;
    uint16* dest_ptr = dest;
    
    uint32 raw;
    uint16 value;
    
    while (src_ptr < end) {
        raw = *src_ptr++;
        value = 0;
        value |= (uint16)(raw & 0x000003FF);
        value |= (uint16)((raw & 0x00F00000) >> 8);
        value |= (uint16)((raw & 0x01000000) >> 9);
        value |= (uint16)((raw & 0x40000000) >> 20);
        value |= (uint16)((raw & 0x80000000) >> 20);
        *dest_ptr++ = value;
    }
}

void gfx_decode(void* dest, const void* src, int length, int format) {
    
    void (*decode)(void*, const void*, int) = NULL;
    
    switch (format) {
            
        // graphics formats
        case LINEAR_8BPP:       decode = decode_linear8bpp; break;
        case LINEAR_2BPP:       decode = decode_linear2bpp; break;
        case LINEAR_4BPP:       decode = decode_linear4bpp; break;
        case NES_1BPP:          decode = decode_nes1bpp; break;
        case NES_2BPP:          decode = decode_nes2bpp; break;
        case SNES_2BPP:         decode = decode_snes2bpp; break;
        case SNES_3BPP:         decode = decode_snes3bpp; break;
        case SNES_4BPP:         decode = decode_snes4bpp; break;
        case SNES_MODE7:        decode = decode_snesmode7; break;
        case SNES_MODE7_EX:     decode = decode_snesmode7ex; break;
        
        // palette formats
        case BGR555:            decode = decode_bgr555; break;
        
        // tile formats
        case SNES_2BPP_TILE:    decode = decode_snes_2bpp_tile; break;
        case SNES_4BPP_TILE:    decode = decode_snes_4bpp_tile; break;
        case SNES_SPRITE_TILE:  decode = decode_snes_sprite_tile; break;
        case SNES_MODE7_TILE:   decode = decode_snes_mode7_tile; break;
        case GBA_4BPP_TILE:     decode = decode_gba_4bpp_tile; break;

        default:
            printf("Format %d not supported\n", format);
            return;
    }
    decode(dest, src, length);
}

void gfx_encode(void* dest, const void* src, int length, int format) {
    
    void (*encode)(void*, const void*, int) = NULL;

    switch (format) {
            
        // tile formats
        case SNES_2BPP_TILE:    encode = encode_snes_2bpp_tile; break;
        case SNES_4BPP_TILE:    encode = encode_snes_4bpp_tile; break;
        case SNES_SPRITE_TILE:  encode = encode_snes_sprite_tile; break;
        case SNES_MODE7_TILE:   encode = encode_snes_mode7_tile; break;
        case GBA_4BPP_TILE:     encode = encode_gba_4bpp_tile; break;
        
        default:
            printf("Format %d not supported\n", format);
            return;
    }
    encode(dest, src, length);
}
