//
//  DialogText.swift
//  FF6Tools
//
//  Created by Evan Owen on 2/3/17.
//  Copyright © 2017 Evan Owen. All rights reserved.
//

import Foundation

class DialogTextEncoding: ROMTextEncoding {
    
    static let hiraganaCharacters = "ぁあぃいぅうぇえぉおかがきぎくぐけげこごさざしじすずせぜそぞただちぢっつづてでとどなにぬねのはばぱひびぴふぶぷへべぺほぼぽまみむめもゃやゅゆょよらりるれろゎわゐゑをんゔゕゖじどこいえも"

    override func format(string: String) -> String {
        
        var formattedString = string
        
        // get all of the escape sequences, sorted from longest to shortest
        let escapeSequences = (encodingTable.allKeys as! [String]).filter{ $0.hasPrefix("\\") }.sorted{ $0.characters.count > $1.characters.count }
        
        for sequence in escapeSequences {
            
            switch sequence {
                
            case "\\n":
                // new line
                formattedString = formattedString.replacingOccurrences(of: sequence, with: "\n") // new line
                
            case "\\page":
                // new page
                formattedString = formattedString.replacingOccurrences(of: sequence, with: "\u{12}") // form feed
                
            case "\\t[":
                // tab
                // use the full-width space if it exists in the encoding table
                let spaceChar: Character = (encodingTable["\u{3000}"] == nil) ? " " : "\u{3000}"
                
                // format for each occurrence of the tab sequence
                while let tabRange = formattedString.range(of: sequence) {
                    var tabString = String(formattedString[tabRange.lowerBound...])

                    // find the start and end of the parameter
                    if let startOfParameter = tabString.range(of: "[")?.upperBound,
                        let endOfParameter = tabString.range(of: "]")?.lowerBound, startOfParameter < endOfParameter {
                        let parameterString = String(tabString[startOfParameter..<endOfParameter])
                        
                        // parse the parameter as an integer (number of spaces)
                        if let parameterValue = Int(object: parameterString as AnyObject?), parameterValue <= 0xFF {
                            
                            // remove the tab sequence and parameter from the formatted string
                            tabString = String(tabString[..<tabString.index(after: endOfParameter)])
                            formattedString.removeSubrange(formattedString.range(of: tabString)!)
                            
                            // add spaces to the formatted string
                            for _ in 0..<parameterValue {
                                formattedString.insert(spaceChar, at: tabRange.lowerBound)
                            }
                        } else {
                            // unable to parse the number of spaces
                            rom.log("Invalid parameter value " + parameterString)
                            formattedString.removeSubrange(tabRange)
                        }
                    } else {
                        // no parameter terminator found
//                        rom.log("Missing parameter terminator")
                        formattedString.removeSubrange(tabRange)
                    }
                }
                
            case "\\choice":
                // multiple choice indicator
                formattedString = formattedString.replacingOccurrences(of: sequence, with: "\u{5}") // enquiry
                
            case let nameSequence where nameSequence.hasPrefix("\\char"):
                // character name
                let scanner = Scanner(string: nameSequence)
                var characterName = ""
                var characterIndex = 0
                scanner.scanString("\\char", into: nil) // ignore leading "\\char"
                if scanner.scanInt(&characterIndex) == false {
                    rom.log("Invalid character name index")
                    formattedString = formattedString.replacingOccurrences(of: sequence, with: "")
                    break
                }
                
                let nameString = rom.value(at: characterIndex, inPropertyWithKey: "characterNames") as! ROMText
                
                if scanner.scanString("_h", into: nil) {
                    // half-width kana
                    var nameBytes = [UInt8](repeating: 0, count: nameString.assembledLength)
                    nameString.encodedData.getBytes(&nameBytes, length: nameBytes.count)
                    
                    for characterByte in nameBytes {
                        var characterByteInt = Int(characterByte)
                        characterByteInt -= (encodingTable["バ"] as? Int) ?? 0x20
                        characterByteInt += (encodingTable["ﾊﾞ"] as? Int) ?? 0x1F20
                        if let halfWidthCharacter = decodingTable[characterByteInt] as? String {
                            characterName.append(halfWidthCharacter)
                        }
                    }
                } else {
                    // full-width letters
                    characterName = nameString.description
                }
                formattedString = formattedString.replacingOccurrences(of: nameSequence, with: characterName)
                
            case let nameSequence where nameSequence.hasPrefix("\\h_char"):
                // character name (half-width kana)
                let scanner = Scanner(string: nameSequence)
                var characterName = ""
                var characterIndex = 0
                scanner.scanString("\\h_char", into: nil) // ignore leading "\\h_char"
                if scanner.scanInt(&characterIndex) {
                    
                    let nameString = rom.value(at: characterIndex, inPropertyWithKey: "characterNames") as! ROMText
                    var nameBytes = [UInt8](repeating: 0, count: nameString.assembledLength)
                    nameString.encodedData.getBytes(&nameBytes, length: nameBytes.count)
                    
                    for characterByte in nameBytes {
                        let characterByteInt = Int(characterByte) + 0x1F00
                        if let halfWidthCharacter = decodingTable[characterByteInt] as? String {
                            characterName.append(halfWidthCharacter)
                        }
                    }
                }
                formattedString = formattedString.replacingOccurrences(of: nameSequence, with: characterName)
                
            case let hiraganaRange where (hiraganaRange.rangeOfCharacter(from: CharacterSet(charactersIn: DialogTextEncoding.hiraganaCharacters)) != nil):
                // ignore half-width hiragan (this needs to be converted somewhere else)
                continue
                
            case let sequenceWithParameter where sequenceWithParameter.hasSuffix("["):
                // escape sequence with parameter
                while let sequenceRange = formattedString.range(of: sequenceWithParameter) {
                    var substring = String(formattedString[sequenceRange.lowerBound...])
                    if let endOfParameter = substring.characters.index(of: "]") {
                        substring = String(substring[..<substring.index(after: endOfParameter)])
                        formattedString.removeSubrange(formattedString.range(of: substring)!)
                    } else {
                        formattedString.removeSubrange(sequenceRange)
                    }
                }
                
            default:
                // remove all other escape sequences
                formattedString = formattedString.replacingOccurrences(of: sequence, with: "")
            }
        }
        return formattedString
    }
}
