//
//  FF6AttackController.swift
//  FF6Tools
//
//  Created by Evan Owen on 1/8/17.
//  Copyright © 2017 Evan Owen. All rights reserved.
//

import Cocoa

class FF6AttackController: NSObject, NSSplitViewDelegate {
    
    @IBOutlet weak var windowController: ROMWindowController!
    @IBOutlet weak var frameView: FF6AttackFrameView!
    @IBOutlet weak var frameEditor: FF6AttackFrameEditor!
    @IBOutlet weak var tilesetView: FF6AttackTilesetView!
    @IBOutlet weak var splitView: NSSplitView!
    @IBOutlet weak var layerSegmentedControl: NSSegmentedControl!
    var rom: FF6ROM? { return windowController?.rom as? FF6ROM }
    var isGBA: Bool { return rom?.system == .GBA ? true : false }

    // layers
    var layers = [FF6AttackLayer]()
    @objc dynamic var selectedLayerIndex = 0 {
        didSet {
            // select the layer property view
//            if selectedLayer.graphicsProperties != nil {
//                windowController.deselectAllObjects()
//                windowController.selectObject(selectedLayer)
//            }
            
            // return if the layer didn't change
            if selectedLayerIndex == oldValue { return }
            
            // select the first frame
            selectedFrame = 0
            
            // reset the frame size and scroll to the origin
            frameView.setFrameSize(frameView.frame.size)
            frameView.scroll(NSZeroPoint)
            
            // redraw all views
            frameView.needsDisplay = true
            frameView.frameLayers.removeAll()
            frameEditor.needsDisplay = true
            tilesetView.needsDisplay = true
        }
    }
    var selectedLayer: FF6AttackLayer { return layers[selectedLayerIndex] }
    var selectedFrame = 0

    var animationProperties: ROMData? {
        willSet {
            stopObserving(object: animationProperties)
        }
        didSet {
            loadAnimation()
            startObserving(object: animationProperties)
            if animationProperties !== oldValue {
                selectedFrame = 0
                frameView.scroll(NSZeroPoint)
            }
            if animationProperties?.identifier == "monsterSpecialAttackProperties" {
                selectedLayerIndex = 2
//            } else {
//                selectedLayerIndex = 0
            }
        }
    }
    
    @IBAction func clickedLayer(_ sender: AnyObject) {
        
        // make sure the layer's script pointer has a valid reference
        if let attackScriptPointer = selectedLayer.attackScriptPointer,
            attackScriptPointer.references?["scriptPointer"] == nil {
            let attackScript = rom?.value(forKey: "attackScript") as? ROMScript
            _ = attackScript?.addReference(forAssembly: attackScriptPointer, encodingName: "attackScript", linkedKeyPath: "scriptPointer")
        }

        // select the layer property view
        windowController.deselectAllObjects()
        if selectedLayer.graphicsProperties != nil {
            windowController.selectObject(selectedLayer)
        }
    }

    func startObserving(object: AnyObject?) {
        
        // return if the object is not selectable or if it is not currently selected
        guard let selected = object as? ROMSelectable else { return }
        
        // properties for ROMSelectable objects
        if let properties = selected.properties {
            for (key, _) in properties {
                guard let propertyIdentifier = key as? String else { continue }
                object?.addObserver(self, forKeyPath: propertyIdentifier, options: NSKeyValueObservingOptions.old, context: &FF6AttackController.context)
            }
        }
    }
    
    func stopObserving(object: AnyObject?) {
        
        // return if the object is not selectable or if it is not currently selected
        guard let selected = object as? ROMSelectable else { return }
        
        // properties for ROMSelectable objects
        if let properties = selected.properties {
            for (key, _) in properties {
                guard let propertyIdentifier = key as? String else { continue }
                object?.removeObserver(self, forKeyPath: propertyIdentifier, context: &FF6AttackController.context)
            }
        }
    }

    static var context = 0
    
    override func observeValue(forKeyPath keyPath: String?, of object: Any?, change: [NSKeyValueChangeKey : Any]?, context: UnsafeMutableRawPointer?) {
        if context != &FF6AttackController.context {
            super.observeValue(forKeyPath: keyPath, of: object, change: change, context: context)
            return
        }
        
        loadAnimation()
    }

    func loadAnimation() {
        
        // when the animation changes
        
        // stop observing previous layers
        for layer in layers {
            stopObserving(object: layer.graphicsProperties)
        }
        
        // redraw all views
        frameView.needsDisplay = true
        frameView.frameLayers.removeAll()
        frameEditor.needsDisplay = true
        tilesetView.needsDisplay = true
        
        guard let animationIdentifier = animationProperties?.identifier else { return }
        
        switch animationIdentifier {
        case "attackAnimationProperties":
            loadAttack()

        case "weaponAnimationProperties", "monsterSpecialAttackProperties":
            loadWeapon()
            
        default:
            return
        }
        
        // start observing new layers
        for layer in layers {
            startObserving(object: layer.graphicsProperties)
        }
        
        // reset the frame size
        frameView.setFrameSize(frameView.frame.size)
    }
    
    func loadWeapon() {
        
        layers.removeAll()
        
        layerSegmentedControl.setLabel("Left Hand", forSegment: 0)
        layerSegmentedControl.setLabel("Right Hand", forSegment: 1)
        layerSegmentedControl.setLabel("Hit", forSegment: 2)

        for layerType in [FF6AttackLayerType.rightHand, FF6AttackLayerType.leftHand, FF6AttackLayerType.hit] {
            
            let layer = FF6AttackLayer()
            layers.append(layer)
            layer.graphicsIndex = 0x7FFF
            layer.type = layerType
            layer.tilesetSize = 32
            
            // get the script index from the animation properties
            var script = animationProperties!.intValue(forKey: layerType.scriptKey)
            if script >= 0x60 && layerType == .hit { script += 0x0200 } // see C1/9DF3
            
            // get the graphics properties
            layer.graphicsProperties = rom?.value(at: script, inPropertyWithKey: "attackGraphicsProperties") as? ROMData
            if layer.graphicsProperties == nil { continue }
            
            // get the script pointer
            layer.attackScriptPointer = rom?.value(at: script, inPropertyWithKey: "attackScriptPointers") as? ROMData
            
            // decode the palette
            let paletteIndex = animationProperties!.intValue(forKey: layer.type.paletteKey)
            let palLength = 8 * MemoryLayout<UInt32>.size
            let pal = UnsafeMutablePointer<UInt8>.allocate(capacity: palLength)
            memset(pal, 0, palLength)
            guard let paletteData = rom?.value(at: paletteIndex, inPropertyWithKey: "attackPalettes") as? ROMData else { continue }
            paletteData.data.withUnsafeBytes{(bytes: UnsafePointer<UInt8>) -> Void in
                gfx_decode(pal, bytes, Int32(paletteData.data.count), BGR555)
            }
            layer.palette = Data(bytes: pal, count: palLength)
            pal.deinitialize(); pal.deallocate(capacity: palLength)
            
            // load the tileset
            guard let tilesetData = rom?.value(forKey: layer.tilesetKey) as? ROMData else { continue }
            
            // decode the graphics
            guard let graphics = rom?.value(forKey: layer.graphicsKey) as? ROMData else { continue }
            let graphicsFormat = isGBA ? LINEAR_4BPP : (layer.is2bpp ? SNES_2BPP : SNES_3BPP)
            let decodedGraphicsLength = 0x10000
            let decodedGraphics = UnsafeMutablePointer<UInt8>.allocate(capacity: decodedGraphicsLength)
            memset(decodedGraphics, 0, decodedGraphicsLength)
            graphics.data.withUnsafeBytes{(bytes: UnsafePointer<UInt8>) -> Void in
                let tileSize = isGBA ? 0x20 : (layer.is2bpp ? 0x10 : 0x18)
                for index in 0..<256 {
                    let tilePropertiesRange = NSMakeRange(layer.tilesetIndex * 64 + index * 2, 2)
                    if NSMaxRange(tilePropertiesRange) >= tilesetData.data.count { break }
                    let tileProperties: UInt16 = tilesetData.data.subdata(in: Range(tilePropertiesRange)!).withUnsafeBytes{ $0.pointee }
                    gfx_decode(decodedGraphics.advanced(by: index * 0x40), bytes.advanced(by: (Int(tileProperties) & 0x3FFF) * tileSize), Int32(tileSize), graphicsFormat)
                    layer.tileset[index] = UInt16(index | (Int(tileProperties) & 0xC000))
                }
            }
            layer.graphics = Data(bytes: decodedGraphics, count: decodedGraphicsLength)
            layer.graphicsIndex = script
            decodedGraphics.deinitialize(); decodedGraphics.deallocate(capacity: decodedGraphicsLength)
        }
    }
    
    func loadAttack() {
        
        layers.removeAll()

        layerSegmentedControl.setLabel("Sprite", forSegment: 0)
        layerSegmentedControl.setLabel("Layer 1", forSegment: 1)
        layerSegmentedControl.setLabel("Layer 3", forSegment: 2)

        for layerType in [FF6AttackLayerType.sprite, FF6AttackLayerType.layer1, FF6AttackLayerType.layer3] {
            
            let layer = FF6AttackLayer()
            layers.append(layer)
            layer.graphicsIndex = 0x7FFF
            layer.type = layerType
            if layerType == .layer1 {
                layer.tilesetSize = 40
            } else {
                layer.tilesetSize = 32
            }
            
            // get the script index from the animation properties
            let script = animationProperties!.intValue(forKey: layer.type.scriptKey)
//            let noGraphics = animationProperties!.boolValue(forKey: layer.type.noGraphicsKey)
            if script == 0x7FFF { continue }
            
            // get the graphics properties
            layer.graphicsProperties = rom?.value(at: script, inPropertyWithKey: "attackGraphicsProperties") as? ROMData
            if layer.graphicsProperties == nil { continue }
            
            // get the script pointer
            layer.attackScriptPointer = rom?.value(at: script, inPropertyWithKey: "attackScriptPointers") as? ROMData

            // decode the palette
            let paletteIndex = animationProperties!.intValue(forKey: layer.type.paletteKey)
            let palLength = 8 * MemoryLayout<UInt32>.size
            let pal = UnsafeMutablePointer<UInt8>.allocate(capacity: palLength)
            memset(pal, 0, palLength)
            guard let paletteData = rom?.value(at: paletteIndex, inPropertyWithKey: "attackPalettes") as? ROMData else { continue }
            paletteData.data.withUnsafeBytes{(bytes: UnsafePointer<UInt8>) -> Void in
                gfx_decode(pal, bytes, Int32(paletteData.data.count), BGR555)
            }
            layer.palette = Data(bytes: pal, count: palLength)
            pal.deinitialize(); pal.deallocate(capacity: palLength)
            
            // load the tileset
            guard let tilesetData = rom?.value(forKey: layer.tilesetKey) as? ROMData else { continue }
            
            // decode the graphics
            guard let graphics = rom?.value(forKey: layer.graphicsKey) as? ROMData else { continue }
            let graphicsFormat = isGBA ? LINEAR_4BPP : (layer.is2bpp ? SNES_2BPP : SNES_3BPP)
            let decodedGraphicsLength = 0x10000
            let decodedGraphics = UnsafeMutablePointer<UInt8>.allocate(capacity: decodedGraphicsLength)
            memset(decodedGraphics, 0, decodedGraphicsLength)
            graphics.data.withUnsafeBytes{(bytes: UnsafePointer<UInt8>) -> Void in
                let tileSize = isGBA ? 0x20 : (layer.is2bpp ? 0x10 : 0x18)
                for index in 0..<256 {
                    let tilePropertiesRange = NSMakeRange(layer.tilesetIndex * 64 + index * 2, 2)
                    if NSMaxRange(tilePropertiesRange) >= tilesetData.data.count { break }
                    let tileProperties: UInt16 = tilesetData.data.subdata(in: Range(tilePropertiesRange)!).withUnsafeBytes{ $0.pointee }
                    gfx_decode(decodedGraphics.advanced(by: index * 0x40), bytes.advanced(by: (Int(tileProperties) & 0x3FFF) * tileSize), Int32(tileSize), graphicsFormat)
                    layer.tileset[index] = UInt16(index | (Int(tileProperties) & 0xC000))
                }
            }
            layer.graphics = Data(bytes: decodedGraphics, count: decodedGraphicsLength)
            layer.graphicsIndex = script
            decodedGraphics.deinitialize(); decodedGraphics.deallocate(capacity: decodedGraphicsLength)
        }
        
        if layers[0].graphicsIndex == 656 {
            loadAttackAdvance(0, toLayer: &layers[0]) // 10,000 needles
        } else if layers[2].graphicsIndex == 652 {
            loadAttackAdvance(1, toLayer: &layers[2]) // tidal wave
        } else if layers[0].graphicsIndex == 662 {
            loadAttackAdvance(2, toLayer: &layers[0]) // excalibur
        } else if layers[0].graphicsIndex == 663 {
            loadAttackAdvance(2, toLayer: &layers[0]) // excalipoor
        } else if layers[0].graphicsIndex == 664 {
            loadAttackAdvance(2, toLayer: &layers[0]) // masamune
        } else if layers[0].graphicsIndex == 666 {
            loadAttackAdvance(2, toLayer: &layers[0]) // enkidu
        } else if layers[0].graphicsIndex == 660 {
            loadAttackAdvance(3, toLayer: &layers[0]) // dark messenger
        }
        
        if layers[1].graphicsIndex == 650 {
            loadEsper(0x019F, toLayer: &layers[1]) // leviathan
        } else if layers[1].graphicsIndex == 658 {
            loadEsper(0x01A0, toLayer: &layers[1]) // diabolus
        } else if layers[1].graphicsIndex == 655 {
            loadEsper(0x01A1, toLayer: &layers[1]) // gigantaur
        } else if layers[1].graphicsIndex == 661 {
            loadEsper(0x01A2, toLayer: &layers[1]) // gilgamesh
        } else if layers[1].graphicsIndex == 665 {
            loadEsper(0x01A3, toLayer: &layers[1]) // enkidu
        } else if layers[1].graphicsIndex == 653 {
            loadEsper(0x004C, toLayer: &layers[1]) // cactuar
        }
        
        // load esper graphics
        switch animationProperties!.intValue(forKey: "initFunction") {
        case 51: loadEsper(0x0181, toLayer: &layers[0]) // ifrit
        case 52: loadEsper(0x0185, toLayer: &layers[0]) // shoat
        case 53: loadEsper(0x0184, toLayer: &layers[0]) // terrato
        case 54: loadEsper(0x0187, toLayer: &layers[0]) // bismark
        case 55: loadEsper(0x0192, toLayer: &layers[1]) // zoneseek
        case 56: loadEsper(0x0180, toLayer: &layers[0]) // ramuh
        case 57: loadEsper(0x0182, toLayer: &layers[0]) // shiva
        case 58: loadEsper(0x0183, toLayer: &layers[1]) // siren
        case 59: loadEsper(0x018E, toLayer: &layers[1]) // alexandr
        case 60: loadEsper(0x0193, toLayer: &layers[0]) // carbunkl
        case 61: loadEsper(0x0191, toLayer: &layers[0]) // kirin
        case 62: loadEsper(0x0197, toLayer: &layers[0]) // unicorn
        case 63: loadEsper(0x019A, toLayer: &layers[0]) // phoenix
        case 64: loadEsper(0x0195, toLayer: &layers[1]) // sraphim
        case 65: loadEsper(0x0199, toLayer: &layers[0]) // starlet
        case 66: loadEsper(0x018D, toLayer: &layers[1]) // bahamut
        case 67: loadEsper(0x0186, toLayer: &layers[0]) // maduin
        case 68: loadEsper(0x0189, toLayer: &layers[1]) // palidor
        case 69: loadEsper(0x018B, toLayer: &layers[0]) // odin
        case 70: loadEsper(0x0198, toLayer: &layers[1]) // fenrir
        case 104: loadEsper(0x0188, toLayer: &layers[1]) // stray
        case 105: loadEsper(0x0196, toLayer: &layers[1]) // golem
//        case 106: loadEsper(0x0190, toLayer: &layers[1]) // ragnarok
        case 107: loadEsper(0x018C, toLayer: &layers[0]) // raiden
        case 108: loadEsper(0x018A, toLayer: &layers[0]) // tritoch
        case 109: loadEsper(0x0194, toLayer: &layers[1]) // phantom
        case 110:
            loadEsper(0x019B, toLayer: &layers[0]) // tigerbreak
            // load sabin's color palette
            guard let spritePalette = rom?.value(at: 0, inPropertyWithKey: "mapSpritePalettes") as? ROMData else { break }
            let decodedPaletteLength = 16 * 4
            let decodedPalette = UnsafeMutablePointer<UInt8>.allocate(capacity: decodedPaletteLength)
            gfx_decode(decodedPalette, (spritePalette.data as NSData).bytes, 32, BGR555)
            layers[0].palette = Data(bytes: UnsafePointer<UInt8>(decodedPalette), count: decodedPaletteLength)
            decodedPalette.deinitialize(); decodedPalette.deallocate(capacity: decodedPaletteLength)
            
        case 116: // crusader
            loadEsper(0x018F, toLayer: &layers[1])
            loadEsper(0x019D, toLayer: &layers[0])
            // 0x018F top dude
            // 0x019C big dude on the right
            // 0x019D insect dude on the left
            
        case 121: // misc. espers loaded to bg1 (see C2/EAA7)
            switch animationProperties!.intValue(forKey: "soundEffect") {
            case 0: loadEsper(0x0182, toLayer: &layers[1])
            case 1: loadEsper(0x0191, toLayer: &layers[1])
            case 2: loadEsper(0x0187, toLayer: &layers[1])
            case 3: loadEsper(0x0193, toLayer: &layers[1])
            case 4: loadEsper(0x0184, toLayer: &layers[1])
            default: break
            }
            
        case 99, 100, 117: // S.Cross, Overcast, H-Bomb
            // load mode 7 battle animation graphics (32x32)
            guard let _ = rom?.value(forKey: "attackGraphicsMode7") as? ROMData else { break }
            
        default: break
        }
    }
    
    func loadEsper(_ esperIndex: Int, toLayer esperLayer: inout FF6AttackLayer) {
        // decode esper graphics
        guard let graphicsProperties = rom?.value(at: esperIndex, inPropertyWithKey: "monsterGraphicsProperties") as? ROMData else { return }

        let graphicsPointer = graphicsProperties.intValue(forKey: "graphicsPointer") << (isGBA ? 4 : 3)
        var paletteIndex = graphicsProperties.intValue(forKey: "palette")
        var is3bpp = false
        if !isGBA {
            // swap palette bytes (SNES only)
            paletteIndex = Int(UInt16(paletteIndex).byteSwapped)
            is3bpp = graphicsProperties.boolValue(forKey: "is3bpp")
        }
        let isLargeMap = graphicsProperties.boolValue(forKey: "largeMap")
        let mapIndex = graphicsProperties.intValue(forKey: "mapIndex")
        
        // image and tile size
        let width = 128
        let height = isLargeMap ? 128 : 64
        let tileSize = is3bpp ? 0x18 : 0x20
        
        // make a buffer for the raw graphics
        let rawLength = Int(height * width / 64) * tileSize
        let rawGraphics = UnsafeMutablePointer<UInt8>.allocate(capacity: rawLength)
        memset(UnsafeMutableRawPointer(rawGraphics), 0, rawLength)
        
        // get the graphics and graphics map
        guard let graphics = rom?.value(forKeyPath: "monsterGraphics.data") as? Data else { return }
        
        var graphicsMap: Data
        if isGBA {
            // gba graphics maps are stored in sequential order, regardless of size
            guard let monsterGraphicsMaps = rom?.value(forKey: "monsterGraphicsMaps") as? ROMData else { return }
            let mapSize = isLargeMap ? 32 : 8
            graphicsMap = monsterGraphicsMaps.data.subdata(in: Range(NSMakeRange(mapIndex << 3, mapSize))!)
        } else {
            // small and large graphics maps are stored separately on snes
            let graphicsMapsKey = isLargeMap ? "largeMonsterGraphicsMaps" : "smallMonsterGraphicsMaps"
            guard let mapData = rom?.value(at: mapIndex, inPropertyWithKey: graphicsMapsKey) as? ROMData else { return }
            graphicsMap = mapData.data
        }
        
        // load the raw graphics to the buffer
        var mapRow: UInt16 = 0
        var moldOffset = 0
        var graphicsOffset = graphicsPointer
        let bytesPerTileRow = isLargeMap ? 2 : 1
        var tile = 0
        for _ in 0..<(height / 8) {
            let moldRange = NSMakeRange(moldOffset, bytesPerTileRow)
            (graphicsMap as NSData).getBytes(&mapRow, range: moldRange)
            if !isLargeMap || !isGBA {
                mapRow = mapRow.byteSwapped
            }
            for _ in 0..<(width / 8) {
                if (mapRow & 0x8000) != 0 {
                    let tileRange = NSMakeRange(graphicsOffset, tileSize)
                    (graphics as NSData).getBytes(rawGraphics + tile * tileSize, range: tileRange)
                    graphicsOffset += tileSize
                }
                mapRow <<= 1
                tile += 1
            }
            moldOffset += bytesPerTileRow
        }
        
        let decodedGraphicsLength = 0x10000
        let decodedGraphics = UnsafeMutablePointer<UInt8>.allocate(capacity: decodedGraphicsLength)
        let tileFormat = isGBA ? LINEAR_4BPP : (is3bpp ? SNES_3BPP : SNES_4BPP)
        memset(decodedGraphics, 0, decodedGraphicsLength)
        gfx_decode(decodedGraphics, rawGraphics, Int32(rawLength), tileFormat)
        esperLayer.graphics = Data(bytes: decodedGraphics, count: decodedGraphicsLength)
        decodedGraphics.deinitialize(); decodedGraphics.deallocate(capacity: decodedGraphicsLength)
        rawGraphics.deinitialize(); rawGraphics.deallocate(capacity: rawLength)
        
        // make the color palette
        guard let monsterPalettes = rom?.value(forKey: "monsterPalettes") as? ROMArray else { return }
        var paletteData: Data
        if is3bpp {
            paletteData = (monsterPalettes[Int(paletteIndex)] as! ROMData).data
        } else {
            let combinedPaletteData = NSMutableData()
            let firstPalette = monsterPalettes[Int(paletteIndex)] as! ROMData
            let secondPalette = monsterPalettes[Int(paletteIndex) + 1] as! ROMData
            combinedPaletteData.append(firstPalette.data)
            combinedPaletteData.append(secondPalette.data)
            paletteData = combinedPaletteData as Data
        }
        
        // convert palette to 32-bit ARGB
        let decodedPaletteLength = 16 * 4
        let decodedPalette = UnsafeMutablePointer<UInt8>.allocate(capacity: decodedPaletteLength)
        gfx_decode(decodedPalette, (paletteData as NSData).bytes, 32, BGR555)
        esperLayer.palette = Data(bytes: UnsafePointer<UInt8>(decodedPalette), count: decodedPaletteLength)
        decodedPalette.deinitialize(); decodedPalette.deallocate(capacity: decodedPaletteLength)
        
        // clear flip data in the tileset
        for tile in 0..<256 {
            if esperLayer.tileset[tile] == 0xFFFF { continue }
            esperLayer.tileset[tile] &= 0x3FFF
        }
        esperLayer.tilesetSize = 48
    }
    
    func loadAttackAdvance(_ attackIndex: Int, toLayer advanceLayer: inout FF6AttackLayer) {
        guard let attackGraphics = rom?.value(at: attackIndex, inPropertyWithKey: "attackGraphicsAdvance") as? ROMData else { return }
        
        // decode the color palette
        let decodedPaletteLength = 16 * 4
        let decodedPalette = UnsafeMutablePointer<UInt8>.allocate(capacity: decodedPaletteLength)
        attackGraphics.data.withUnsafeBytes{(bytes: UnsafePointer<UInt8>) -> Void in
            gfx_decode(decodedPalette, bytes.advanced(by: 0x1000), 32, BGR555)
        }
        advanceLayer.palette = Data(bytes: UnsafePointer<UInt8>(decodedPalette), count: decodedPaletteLength)
        decodedPalette.deinitialize(); decodedPalette.deallocate(capacity: decodedPaletteLength)

        // decode the graphics
        let decodedGraphicsLength = 0x10000
        let decodedGraphics = UnsafeMutablePointer<UInt8>.allocate(capacity: decodedGraphicsLength)
        memset(decodedGraphics, 0, decodedGraphicsLength)
        attackGraphics.data.withUnsafeBytes{(bytes: UnsafePointer<UInt8>) -> Void in
            gfx_decode(decodedGraphics, bytes, Int32(0x1000), LINEAR_4BPP)
        }
        advanceLayer.graphics = Data(bytes: decodedGraphics, count: decodedGraphicsLength)
        decodedGraphics.deinitialize(); decodedGraphics.deallocate(capacity: decodedGraphicsLength)
        
        // clear flip data in the tileset
        for tile in 0..<256 {
            if advanceLayer.tileset[tile] == 0xFFFF { continue }
            advanceLayer.tileset[tile] &= 0x3FFF
        }
    }
    
    // MARK: NSSplitViewDelegate methods
    
    func splitView(_ splitView: NSSplitView, constrainMaxCoordinate proposedMaximumPosition: CGFloat, ofSubviewAt dividerIndex: Int) -> CGFloat {
        let max = splitView.frame.height - CGFloat(selectedLayer.frameHeight * 16 + 40)
        if proposedMaximumPosition > max {
            return max
        }
        return proposedMaximumPosition
    }
    
    func splitView(_ splitView: NSSplitView, constrainMinCoordinate proposedMinimumPosition: CGFloat, ofSubviewAt dividerIndex: Int) -> CGFloat {
        let min = splitView.frame.height - CGFloat(selectedLayer.frameHeight * 16 * 3 + 40)
        if proposedMinimumPosition < min {
            return min
        }
        return proposedMinimumPosition
    }
}
