//
//  FF6AttackFrameEditor.swift
//  FF6Tools
//
//  Created by Evan Owen on 12/19/16.
//  Copyright © 2016 Evan Owen. All rights reserved.
//

import Cocoa

class FF6AttackFrameEditor: ROMView {

    @IBOutlet weak var attackController: FF6AttackController!

    var selectedTiles = FrameTiles(x: 0, y: 0, width: 1, height: 1, frames: [0, 0xFFFF])
    var showCursor = false
    
    var zoom: CGFloat {
        return min(frame.size.height / CGFloat(attackController.selectedLayer.frameHeight * 16), frame.size.width / CGFloat(attackController.selectedLayer.frameWidth * 16))
    }
    
    var viewRect: NSRect {
        let width = CGFloat(attackController.selectedLayer.frameWidth * 16) * zoom
        let height = CGFloat(attackController.selectedLayer.frameHeight * 16) * zoom
        return NSMakeRect((frame.width - width) / 2, (frame.height - height) / 2, width, height)
    }

    override func draw(_ dirtyRect: NSRect) {
        super.draw(dirtyRect)

        // fill the view with the background color
        NSColor.windowBackgroundColor.set()
        NSBezierPath.fill(dirtyRect)
        NSGraphicsContext.current?.imageInterpolation = imageInterpolation
        
        // load the frame data
        let frameOffset = attackController.selectedLayer.frameOffset + attackController.selectedFrame
        guard let frameData = attackController.rom?.value(at: frameOffset, inPropertyWithKey: "attackGraphicsFrames") as? ROMData else { return }
        attackController.selectedLayer.frames = frameData.data
        attackController.selectedLayer.decodeLayout()
        guard let frameImage = attackController.selectedLayer.image else { return }
        
        // draw the frame image
        NSColor.black.set()
        NSBezierPath.fill(viewRect)
        NSGraphicsContext.current?.imageInterpolation = imageInterpolation
        NSGraphicsContext.current?.cgContext.draw(frameImage, in: viewRect)
        
        if showCursor { drawCursor() }
    }
    
    func drawCursor() {
        
        let x = min(selectedTiles.x, attackController.selectedLayer.frameWidth) * 16
        let y = min(selectedTiles.y, attackController.selectedLayer.frameHeight) * 16
        let width = min(selectedTiles.width, attackController.selectedLayer.frameWidth - x / 16) * 16
        let height = min(selectedTiles.height, attackController.selectedLayer.frameHeight - y / 16) * 16
        
        let grayAreaAtLeft = viewRect.origin.x
        let grayAreaAtTop = viewRect.origin.y
        
        // convert the selection to screen coordinates
        let left = round(CGFloat(x) * zoom + grayAreaAtLeft) + 0.5
        let right = round(CGFloat(x + width) * zoom + grayAreaAtLeft) - 0.5
        let top = round(CGFloat(y) * zoom + grayAreaAtTop) + 0.5
        let bottom = round(CGFloat(y + height) * zoom + grayAreaAtTop) - 0.5
        let cursorRect = NSMakeRect(left, frame.height - bottom, right - left, bottom - top)
        
        // stroke 4 rectangles to make the cursor
        NSColor.black.set(); NSBezierPath.stroke(cursorRect)
        NSColor.keyboardFocusIndicatorColor.set();
        NSBezierPath.stroke(cursorRect.insetBy(dx: 1, dy: 1))
        NSColor.white.set(); NSBezierPath.stroke(cursorRect.insetBy(dx: 2, dy: 2))
        NSColor.black.set(); NSBezierPath.stroke(cursorRect.insetBy(dx: 3, dy: 3))
    }
    
    // MARK: NSResponder methods
    
    override var acceptsFirstResponder: Bool { return true }
    
    override func mouseDown(with event: NSEvent) {
        let clickedPoint = convert(event.locationInWindow, from: nil)
        
        guard let clickedTile = tileAt(point: clickedPoint) else { return }
        selectedTiles.x = clickedTile.x
        selectedTiles.y = clickedTile.y
        showCursor = true
        
        var event: NSEvent! = event

        while event != nil {
            
            switch event.type {
            case .leftMouseUp:
                return
                
            case .leftMouseDragged, .leftMouseDown:
                // get the dragged/clicked point
                let draggedPoint = convert(event.locationInWindow, from: nil)
                if !bounds.contains(draggedPoint) {
                    break
                }
                
                guard let draggedTile = tileAt(point: draggedPoint) else { break }

                // continue if the cursor hasn't moved
                if (selectedTiles.x == draggedTile.x) && (selectedTiles.y == draggedTile.y) && (event.type != .leftMouseDown) { break }
                
                selectedTiles.x = draggedTile.x
                selectedTiles.y = draggedTile.y
                
                setTiles(selectedTiles)
                
            default:
                break
            }
            
            // get the next event
            event = window!.nextEvent(matching: [.leftMouseUp, .leftMouseDragged])
        }
    }
    
    func setTiles(_ tiles: FrameTiles) {
        // get the frame data
        let frameOffset = attackController.selectedLayer.frameOffset + attackController.selectedFrame
        guard let frameData = attackController.rom?.value(at: frameOffset, inPropertyWithKey: "attackGraphicsFrames") as? ROMData else { return }
        var frameTiles: [UInt16] = frameData.data.withUnsafeBytes{ [UInt16](UnsafeBufferPointer(start: $0, count: frameData.data.count / 2)) }
        
        for tileIndex in 0..<(tiles.height * tiles.width) {
            let selectionX = (tileIndex % tiles.width)
            let selectionY = (tileIndex / tiles.width)
            let selectionPosition = UInt16(selectionX << 4 | selectionY)
            
            let frameX = selectionX + tiles.x
            let frameY = selectionY + tiles.y
            let framePosition = UInt16(frameX << 4 | frameY)
            
            // skip tiles that go off the edge of the frame
            if frameX >= attackController.selectedLayer.frameWidth { continue }
            if frameY >= attackController.selectedLayer.frameHeight { continue }
            
            // remove all previous tiles at this location
            for (index, currentTile) in frameTiles.enumerated() {
                
                if currentTile == 0xFFFF { continue }
                if framePosition != currentTile & 0xFF { continue }
                
                frameTiles.remove(at: index)
            }
            
            // append the new tile at this location
            for selectionTile in tiles.frames {
                
                if selectionTile == 0xFFFF { continue }
                if selectionPosition != selectionTile & 0xFF { continue }
                
                frameTiles.append((selectionTile & 0xFF00) | framePosition)
            }
        }
        
        // sort the frame tiles
        frameTiles.sort{
            let position0 = ($0 & 0x0F) << 4 | ($0 & 0xF0) >> 4
            let position1 = ($1 & 0x0F) << 4 | ($1 & 0xF0) >> 4
            return position0 < position1
        }
        
        setFrameData(frameIndex: frameOffset, data: Data(bytes: &frameTiles, count: frameTiles.count * 2))
    }
    
    @objc func setFrameData(frameIndex: Int, data: Data) {
        
        guard let frameData = attackController.rom?.value(at: frameIndex, inPropertyWithKey: "attackGraphicsFrames") as? ROMData else { return }
        needsDisplay = true
        
        // return if nothing changed
        if frameData.data == data { return }
        
        // save undo data
        if let undo = attackController.windowController?.undoManager {
            let target = undo.prepare(withInvocationTarget: self) as AnyObject
            target.setFrameData(frameIndex: frameIndex, data: frameData.data)
            if (!undo.isUndoing) {
                undo.setActionName("Change Attack Tiles")
            }
        }
        
        frameData.data = data
        
        let firstVisibleFrame = attackController.selectedLayer.frameOffset
        let lastVisibleFrame = firstVisibleFrame + attackController.selectedLayer.frameCount
        
        if (firstVisibleFrame + attackController.selectedFrame) == frameIndex {
            // redraw the editor view if the selected frame changed
            self.needsDisplay = true
            attackController.frameView.frameLayers[frameIndex - firstVisibleFrame] = nil
            attackController.frameView.needsDisplay = true
        } else if (firstVisibleFrame..<lastVisibleFrame).contains(frameIndex) {
            // redraw only the frame view if the frame was one of the others for this animation
            attackController.frameView.frameLayers[frameIndex - firstVisibleFrame] = nil
            attackController.frameView.needsDisplay = true
        }
    }
    
    override func rightMouseDown(with event: NSEvent) {
        window?.makeFirstResponder(self)
        
        var event: NSEvent! = event
        
        let clickedPoint = convert(event.locationInWindow, from: nil)
        if !bounds.contains(clickedPoint) { return }
        
        showCursor = true
        
        // update the selected tile
        guard let clickedTile = tileAt(point: clickedPoint) else { return }
        let clickedPosition = UInt16((clickedTile.x << 4) | clickedTile.y)
        
        // get the frame data
        let frameOffset = attackController.selectedLayer.frameOffset + attackController.selectedFrame
        guard let frameData = attackController.rom?.value(at: frameOffset, inPropertyWithKey: "attackGraphicsFrames") as? ROMData else { return }
        let frameTiles: [UInt16] = frameData.data.withUnsafeBytes{ [UInt16](UnsafeBufferPointer(start: $0, count: frameData.data.count / 2)) }
        
        selectedTiles.height = 1
        selectedTiles.width = 1
        selectedTiles.x = clickedTile.x
        selectedTiles.y = clickedTile.y
        selectedTiles.frames = [0xFFFF]
        attackController.tilesetView?.selectedTiles = selectedTiles
        attackController.tilesetView.needsDisplay = true

        // get the tile at this location
        for currentTile in frameTiles {
            
            if currentTile == 0xFFFF { continue }
            if clickedPosition != currentTile & 0xFF { continue }

            selectedTiles.frames = [currentTile & 0xFF00, 0xFFFF]

            let tilesetX = Int(currentTile & 0x0700) >> 8
            let tilesetY = Int(currentTile & 0x3800) >> 11
            
            attackController.tilesetView?.selectedTiles = selectedTiles
            attackController.tilesetView.horizontalFlip = (currentTile & 0x4000 != 0)
            attackController.tilesetView.verticalFlip = (currentTile & 0x8000 != 0)
            attackController.tilesetView?.selectedTiles.x = attackController.tilesetView.horizontalFlip ? 7 - tilesetX : tilesetX
            attackController.tilesetView?.selectedTiles.y = attackController.tilesetView.verticalFlip ? 7 - tilesetY : tilesetY
            attackController.tilesetView.needsDisplay = true
            break
        }
        needsDisplay = true
        
        // keep track of where the mouse gets dragged
        var draggedTile = clickedTile
        
        while event != nil {
            
            switch event.type {
            case .rightMouseUp:
                // return if the right mouse button is up
                return
                
            case .rightMouseDragged:
                let draggedPoint = convert(event.locationInWindow, from: nil)
                guard let newTile = tileAt(point: draggedPoint) else { break }
                
                // don't update if the dragged coordinates didn't change
                if newTile == draggedTile { break }
                draggedTile = newTile
                
                // get the selected rectangle
                let minX = max(min(draggedTile.x, clickedTile.x), 0)
                let minY = max(min(draggedTile.y, clickedTile.y), 0)
                let maxX = min(max(draggedTile.x, clickedTile.x), 255)
                let maxY = min(max(draggedTile.y, clickedTile.y), 255)
                let width = maxX - minX + 1
                let height = maxY - minY + 1
                
                // create the new tile selection
                selectedTiles.x = minX
                selectedTiles.y = minY
                selectedTiles.width = width
                selectedTiles.height = height
                selectedTiles.frames = [0xFFFF]
                for y in 0..<height {
                    for x in 0..<width {
                        // append the new tile at this location
                        
                        let framePosition = UInt16((x + minX) << 4 | (y + minY))
                        let selectionPosition = UInt16(x << 4 | y)
                        
                        for frameTile in frameTiles {
                            
                            if frameTile == 0xFFFF { continue }
                            if framePosition != frameTile & 0xFF { continue }
                            
                            selectedTiles.frames.append((frameTile & 0xFF00) | selectionPosition)
                            break
                        }
                    }
                }
                // clear the tileset selection
                attackController.tilesetView.selectedTiles = selectedTiles
                
                needsDisplay = true
                
            default: break
            }
            
            // get the next event
            event = window?.nextEvent(matching: [.rightMouseUp, .rightMouseDragged])
        }
    }
    
    override func mouseMoved(with event: NSEvent) {
        
        let mousePoint = convert(event.locationInWindow, from: nil)

        guard let tile = tileAt(point: mousePoint) else {
            showCursor = false
            return
        }

        showCursor = true
        selectedTiles.x = tile.x
        selectedTiles.y = tile.y
        needsDisplay = true
    }
    
    override func mouseEntered(with theEvent: NSEvent) {
        // show the tile cursor when the mouse enters
        window?.makeFirstResponder(self)
        mouseMoved(with: theEvent)
    }
    
    override func mouseExited(with theEvent: NSEvent) {
        // hide the tile cursor when the mouse exits
        showCursor = false
        needsDisplay = true
    }
    
    override func updateTrackingAreas() {
        for area in trackingAreas {
            removeTrackingArea(area)
        }
        window?.acceptsMouseMovedEvents = true
        
        let opts: NSTrackingArea.Options = [NSTrackingArea.Options.mouseEnteredAndExited, NSTrackingArea.Options.activeAlways]
        let trackingArea = NSTrackingArea(rect: viewRect, options: opts, owner: self, userInfo: nil)
        addTrackingArea(trackingArea)
        super.updateTrackingAreas()
    }
    
    func tileAt(point: NSPoint) -> (x: Int, y: Int)? {
        
        if !viewRect.contains(point) { return nil }
        
        let x = Int((point.x - viewRect.origin.x) / zoom / 16)
        let y = Int((frame.height - (point.y + viewRect.origin.y)) / zoom / 16)

        return (x, y)
    }
}
