//
//  FF6AttackTilesetView.swift
//  FF6Tools
//
//  Created by Evan Owen on 12/10/16.
//  Copyright © 2016 Evan Owen. All rights reserved.
//

import Cocoa

struct FrameTiles {
    var x: Int
    var y: Int
    var width: Int
    var height: Int
    var frames: [UInt16]
}

class FF6AttackTilesetView: ROMView {

    @IBOutlet weak var attackController: FF6AttackController!
    @IBOutlet var flipButton: NSSegmentedControl?
    
    var zoom: CGFloat { return frame.width / 128.0 }
    var horizontalFlip = false {
        didSet {
            flipButton?.setSelected(horizontalFlip, forSegment: 0)
        }
    }
    var verticalFlip = false {
        didSet {
            flipButton?.setSelected(verticalFlip, forSegment: 1)
        }
    }
    var selectedTiles = FrameTiles(x: 0, y: 0, width: 1, height: 1, frames: [0, 0xFFFF]) {
        didSet {
            if selectedTiles.frames == [0xFFFF] {
                flipButton?.setSelected(true, forSegment: 2)
            } else {
                flipButton?.setSelected(false, forSegment: 2)
            }
        }
    }
    var tilesetFrames = [UInt16]()
    
    func updateTilesetFrames() {
        tilesetFrames.removeAll()
        
        for tile in 0..<attackController.selectedLayer.tilesetSize {
            var x = tile % 8
            if horizontalFlip { x = 7 - x }
            var y = tile / 8
            if verticalFlip { y = 7 - y }
            let hFlip = horizontalFlip ? 0x4000 : 0
            let vFlip = verticalFlip ? 0x8000 : 0
            let frameValue = (x << 4) | y | hFlip | vFlip | (tile << 8)
            tilesetFrames.append(UInt16(frameValue))
        }
        
        // append a terminator
        tilesetFrames.append(0xFFFF)
    }
    
    @IBAction func changeFlip(_ sender: AnyObject) {
        // sender is a segmented control
//        guard let layerButton = sender as? NSSegmentedControl else { return }

        if horizontalFlip != flipButton?.isSelected(forSegment: 0) {
            // horizontal flip just changed
            horizontalFlip = flipButton!.isSelected(forSegment: 0)
            selectedTiles.x = 8 - selectedTiles.x - selectedTiles.width
            for (index, oldTile) in selectedTiles.frames.enumerated() {
                if oldTile == 0xFFFF { continue }
                let newTile = (oldTile & 0xFF0F) | (UInt16(selectedTiles.width - 1) << 4 - (oldTile & 0x00F0))
                selectedTiles.frames[index] = newTile ^ 0x4000
            }
            attackController.frameEditor.selectedTiles = selectedTiles
        }
        if verticalFlip != flipButton?.isSelected(forSegment: 1) {
            // vertical flip just changed
            verticalFlip = flipButton!.isSelected(forSegment: 1)
            selectedTiles.y = 8 - selectedTiles.y - selectedTiles.height
            for (index, oldTile) in selectedTiles.frames.enumerated() {
                if oldTile == 0xFFFF { continue }
                let newTile = (oldTile & 0xFFF0) | (UInt16(selectedTiles.height - 1) - (oldTile & 0x000F))
                selectedTiles.frames[index] = newTile ^ 0x8000
            }
            attackController.frameEditor.selectedTiles = selectedTiles
        }
        if flipButton?.isSelected(forSegment: 2) == true {
            // erase
            selectedTiles.frames = [0xFFFF]
            attackController.frameEditor.selectedTiles = selectedTiles
        }

        needsDisplay = true
    }
    
    override func draw(_ dirtyRect: NSRect) {
        super.draw(dirtyRect)

        // fill the view with the background color
        backgroundColor.set()
        NSBezierPath.fill(dirtyRect)
        
        // create a dummy graphics properties
        guard let dummyProperties = attackController.selectedLayer.graphicsProperties?.copy() as? ROMData else { return }
        dummyProperties.range = nil
        dummyProperties.disassembleFrom(attackController.selectedLayer.graphicsProperties!.data)

        // create a tileset image
        let tilesetLayer = FF6AttackLayer()
        tilesetLayer.type = attackController.selectedLayer.type
        tilesetLayer.graphicsProperties = dummyProperties
        tilesetLayer.graphicsIndex = attackController.selectedLayer.graphicsIndex
        tilesetLayer.graphics = attackController.selectedLayer.graphics
        tilesetLayer.palette = attackController.selectedLayer.palette
        tilesetLayer.frameCount = 1
        tilesetLayer.frameWidth = 8
        tilesetLayer.frameHeight = 8
        updateTilesetFrames()
        tilesetLayer.frames = Data(bytes: tilesetFrames, count: tilesetFrames.count * 2)
        tilesetLayer.tileset = attackController.selectedLayer.tileset
        tilesetLayer.decodeLayout()
        guard let tilesetImage = tilesetLayer.image else { return }
        
        // draw the tileset
        let tilesetRect = NSMakeRect(0, frame.height - 128 * zoom, 128 * zoom, 128 * zoom)
        NSGraphicsContext.current?.cgContext.draw(tilesetImage, in: tilesetRect)
        
        // draw the cursor
        drawCursor(NSColor.keyboardFocusIndicatorColor)
    }
    
    func drawCursor(_ highlightColor: NSColor) {
        
        // don't show the cursor if the selection is empty
        if selectedTiles.frames.first == 0xFFFF { return }
        
        // get the selected tiles
        let x = selectedTiles.x
        let y = selectedTiles.y
        let width = selectedTiles.width
        let height = selectedTiles.height
        
        // convert the selection to screen coordinates
        let left = round(CGFloat(x) * 16 * zoom) + 0.5
        let right = round(CGFloat(x + width) * 16 * zoom) - 0.5
        let top = round(CGFloat(y) * 16 * zoom) + 0.5
        let bottom = round(CGFloat(y + height) * 16 * zoom) - 0.5
        var cursorRect = NSMakeRect(left, frame.height - bottom, right - left, bottom - top)
        
        // draw the outside border (black)
        NSColor.black.set()
        NSBezierPath.stroke(cursorRect)
        
        // draw the outer highlight
        cursorRect = cursorRect.insetBy(dx: 1.0, dy: 1.0)
        NSColor.keyboardFocusIndicatorColor.set()
        NSBezierPath.stroke(cursorRect)
        
        // draw the inner highlight (white)
        cursorRect = cursorRect.insetBy(dx: 1.0, dy: 1.0)
        NSColor.white.set()
        NSBezierPath.stroke(cursorRect)
        
        // draw the inner border (black)
        cursorRect = cursorRect.insetBy(dx: 1.0, dy: 1.0)
        NSColor.black.set()
        NSBezierPath.stroke(cursorRect)
    }
    
    // MARK: NSResponder methods

    override func mouseDown(with event: NSEvent) {
        let clickedPoint = convert(event.locationInWindow, from: nil)
        if !bounds.contains(clickedPoint) { return }
        
        let clickedTile = (x: Int(clickedPoint.x / zoom / 16), y: 7 - Int(clickedPoint.y / zoom / 16))
        var draggedTile = clickedTile

        selectTile(x: Int(clickedPoint.x / zoom / 16), y: 7 - Int(clickedPoint.y / zoom / 16))

        while true {
            let nextEvent = window!.nextEvent(matching: [.leftMouseUp, .rightMouseUp, .leftMouseDragged, .rightMouseDragged])!
            
            switch (nextEvent.type) {
            case .leftMouseUp, .rightMouseUp:
                return
                
            case .leftMouseDragged, .rightMouseDragged:
                let draggedPoint = convert(nextEvent.locationInWindow, from: nil)
                let draggedX = Int(min(max(draggedPoint.x / zoom / 16, 0), 15))
                let draggedY = Int(min(max((bounds.height - draggedPoint.y) / zoom / 16, 0), 15))
                if draggedTile == (draggedX, draggedY) { continue }
                draggedTile = (draggedX, draggedY)
                
                // get the selected rectangle
                let minX = min(draggedX, clickedTile.x)
                let minY = min(draggedY, clickedTile.y)
                let maxX = max(draggedX, clickedTile.x)
                let maxY = max(draggedY, clickedTile.y)
                let width = maxX - minX + 1
                let height = maxY - minY + 1
                
                selectTiles(x: minX, y: minY, width: width, height: height)
                
            default:
                break
            }
        }
    }
    
    func tileAt(_ x: Int, _ y: Int) -> UInt16 {

        if verticalFlip {
            if (64 - x - y * 8) > attackController.selectedLayer.tilesetSize { return 0xFFFF }
        } else {
            if (x + y * 8) > attackController.selectedLayer.tilesetSize { return 0xFFFF }
        }
        
        updateTilesetFrames()
        return tilesetFrames.first(where: { Int($0 & 0xF0) >> 4 == x && Int($0 & 0x0F) == y } ) ?? 0xFFFF
    }

    func selectTiles(x: Int, y: Int, width: Int, height: Int) {
        
        selectedTiles.x = min(x, 7)
        selectedTiles.y = min(y, 7)
        selectedTiles.width = min(width, 8 - selectedTiles.x)
        selectedTiles.height = min(height, 8 - selectedTiles.y)
        selectedTiles.frames.removeAll()
        for row in 0..<selectedTiles.height {
            for column in 0..<selectedTiles.width {
                let tilesetTile = tileAt(x + column, y + row)
                if tilesetTile == 0xFFFF { continue }
                let selectionPosition = (column << 4) | row
                selectedTiles.frames.append((tilesetTile & 0xFF00) | UInt16(selectionPosition))
            }
        }
        selectedTiles.frames.append(0xFFFF)
        attackController.frameEditor.selectedTiles = selectedTiles
        needsDisplay = true
    }

    func selectTile(x: Int, y: Int) {
        selectTiles(x: x, y: y, width: 1, height: 1)
    }
    
    override func rightMouseDown(with theEvent: NSEvent) {
        // left and right click have the same effect
        mouseDown(with: theEvent)
    }
}
