//
//  FF6WindowController.swift
//  FF6Tools
//
//  Created by Evan Owen on 8/9/16.
//  Copyright © 2016 Evan Owen. All rights reserved.
//

import Cocoa

class FF6View: ROMView {
    var ff6ROM: FF6ROM { return rom as! FF6ROM }
    var isGBA: Bool { return rom.system == .GBA ? true : false }
}

class FF6WindowController: ROMWindowController {
    
    @IBOutlet var mapEditor: FF6MapEditor!
    @IBOutlet var mapTilesetView: FF6MapTilesetView!
    @IBOutlet var battleEditor: FF6BattleEditor?
    @IBOutlet var vramView: FF6BattleVRAMView!
    @IBOutlet var attackController: FF6AttackController!
    
    var defaultBattleBackground: Int?

    @IBAction func optimizeDTE(_ sender: AnyObject) {
        guard let ff6ROM = rom as? FF6ROM,
            let dialog = rom.value(forKey: "dialog") as? ROMArray else {
            rom.log("Unable to optimize DTE table")
            return
        }
        
        let oldDialogLength = dialog.assembledLength
        
        rom.log("Optimizing DTE table...")
        if ff6ROM.optimizeDTE() == true {
            let newDialogLength = dialog.assembledLength
            rom.log("Optimization successful!")
            let deltaLength = oldDialogLength - newDialogLength
            if deltaLength > 0 {
                rom.log("Dialog decreased in size by \(deltaLength) bytes")
            }
        } else {
            rom.log("Optimization failed!")
        }
    }
    
    override func selectObject(_ object: AnyObject?, identifierOverride: String?) {

        guard let selected = object as? ROMSelectable, let identifier = selected.identifier else { return }
        
        switch identifier {
            
        case "mapProperties":
            
            super.selectObject(object, identifierOverride: identifierOverride)
            guard let mapProperties = object as? ROMData else { break }
            let index = mapProperties.index
            
            defaultBattleBackground = mapProperties.intValue(forKey: "battleBackground")
            
            mapEditor.mapIndex = index
            editorView?.tabViewItem(at: 0).view = mapEditor.parentView
            
            mapTilesetView.needsDisplay = true
            toolboxView?.tabViewItem(at: 0).view = mapTilesetView.parentView
            
            if let mapBattleProbability = rom.value(forKey: "mapBattleProbability") as? ROMData {
                // this doesn't work if you change maps and then undo
                let probabilityProperty = ["byteIndex" : (index / 4), "bitMask" : (0xC0 >> (index % 4))]
                let properties = NSMutableDictionary(dictionary: mapBattleProbability.properties ?? NSDictionary())
                properties[String(format: "probability%d", index)] = probabilityProperty
                mapBattleProbability.properties = properties
                super.selectObject(mapBattleProbability)
            }
            
            for object in rom.mutableArrayValue(forKey: "mapStartupEventPointersAdvance") {
                
                // get the event pointer for each trigger
                guard let pointerAssembly = object as? ROMData else { continue }
                if pointerAssembly.intValue(forKey: "mapIndex") == index {
                    super.selectObject(pointerAssembly)
                    break
                }
            }
            
        case "battleProperties":
            
            super.selectObject(object, identifierOverride: identifierOverride)
            guard let battleProperties = object as? ROMData else { break }
            let index = battleProperties.index

            battleEditor?.battle = index
            editorView?.tabViewItem(at: 0).view = battleEditor?.superview
            
            vramView.battle = index
            toolboxView?.tabViewItem(at: 0).view = vramView.parentView
            
            // update the monster names in the monsterSlots string table
            if let monsterSlotsStringTable = rom.value(forKey: "stringTable-monsterSlots") as? ROMStringTable {
                
                for slot in 1...6 {
                    guard let monsterSlotString = monsterSlotsStringTable[slot - 1] as? ROMString else { continue }
                    
                    monsterSlotString.rawString = String(slot) + ": " + (battleEditor?.monsterInSlot(slot)?.name ?? "<Unknown>")
                }
            }
            
            // select the battle's total experience and GP
            super.selectObject(battleEditor?.battleExperienceGP, identifierOverride: nil)

        case "battleScriptPointers":
            
            super.selectObject(object, identifierOverride: identifierOverride)
            guard let battleScriptPointer = object as? ROMData,
                let battleScript = rom.value(forKey: "battleScript") as? ROMScript,
                battleScriptPointer.references?["scriptPointer"] == nil else { break }

            battleScriptPointer.willChangeValue(forKey: "scriptPointer")
            battleScriptPointer.references = nil
            stopObserving(object: battleScriptPointer)
            battleScriptPointer.setValue(battleScript.range?.lowerBound, forKey: "scriptPointer")
            startObserving(object: battleScriptPointer)
            battleScript.addReference(forAssembly: battleScriptPointer, encodingName: "battleScript", linkedKeyPath: "scriptPointer")
            battleScriptPointer.didChangeValue(forKey: "scriptPointer")

        case "attackAnimationProperties", "weaponAnimationProperties", "monsterSpecialAttackProperties", "weaponAnimationPropertiesAdvance":

            super.selectObject(object, identifierOverride: identifierOverride)
            guard let assembly = object as? ROMAssembly else { break }
            let index = assembly.index
            guard let romArray = assembly.parentArray, index < romArray.count else { break }
            
            attackController.animationProperties = romArray[index] as? ROMData
            editorView?.tabViewItem(at: 0).view = attackController.frameView.parentView
            toolboxView?.tabViewItem(at: 0).view = attackController.tilesetView.parentView

        case "blitzCodes":
            super.selectObject(object, identifierOverride: identifierOverride)
            guard let blitzCode = object as? ROMData else { break }
            let index = blitzCode.index

            // select the attack name
            let attackNames = rom.value(forKey: "attackNames") as? ROMArray
            selectObject(attackNames?[index + 12], identifierOverride: nil)
            
        case "monsterProperties":

            super.selectObject(object, identifierOverride: identifierOverride)
            guard let monsterProperties = object as? ROMData else { break }
            let index = monsterProperties.index

            // update the monster's special attack name in the string table
            if let attacksStringTable = rom.stringTable(forKey: "attackProperties"),
                let specialAttackString = attacksStringTable[0xEF] as? ROMString {
                specialAttackString.rawString = "Special ([rom.monsterSpecialAttackNames.\(index)])"
            }
            
            // make sure the monster script has a valid reference (in case this is a new monster)
            if let monsterScriptPointer = rom.value(at: index, inPropertyWithKey: "monsterScriptPointers") as? ROMData,
                let monsterScript = rom.value(forKey: "monsterScript") as? ROMScript,
                monsterScriptPointer.references?["scriptPointer"] == nil {
                monsterScriptPointer.willChangeValue(forKey: "scriptPointer")
                monsterScriptPointer.references = nil
                monsterScript.addReference(forAssembly: monsterScriptPointer, encodingName: "monsterScript", linkedKeyPath: "scriptPointer")
                monsterScriptPointer.didChangeValue(forKey: "scriptPointer")
            }
            
            // get the initial rage property
            if let initialRages = rom.value(forKey: "initialRages") as? ROMData, index < 256 {
                // TODO: this doesn't work if you change monsters and then undo
                let knownProperty = ["byteIndex" : (index / 8), "bitMask" : (0x01 << (index % 8))]
                let properties = NSMutableDictionary(dictionary: initialRages.properties ?? NSDictionary())
                properties[String(format: "known", index)] = knownProperty
                initialRages.properties = properties
                super.selectObject(initialRages)
            }

        case "loreDescriptions":
            
            super.selectObject(object, identifierOverride: identifierOverride)
            guard let loreDescription = object as? ROMText else { break }
            let index = loreDescription.index

            if let initialLores = rom.value(forKey: "initialLores") as? ROMData, index < 24 {
                // this doesn't work if you change lores and then undo
                let knownProperty = ["byteIndex" : (index / 8), "bitMask" : (0x01 << (index % 8))]
                let properties = NSMutableDictionary(dictionary: initialLores.properties ?? NSDictionary())
                properties[String(format: "known", index)] = knownProperty
                initialLores.properties = properties
                super.selectObject(initialLores)
            }
            let attackNames = rom.value(forKey: "attackNames") as? ROMArray
            selectObject(attackNames?[index + 58], identifierOverride: nil)

        case "itemProperties":
            super.selectObject(object, identifierOverride: identifierOverride)
            guard let itemProperties = object as? ROMData else { break }
            switch itemProperties.intValue(forKey: "type") {
            case 0: showPropertyView(withIdentifier: "toolProperties", fromNib: itemProperties.nib)
            case 1: showPropertyView(withIdentifier: "weaponProperties", fromNib: itemProperties.nib)
            case 2: showPropertyView(withIdentifier: "armorProperties", fromNib: itemProperties.nib)
            case 3: showPropertyView(withIdentifier: "armorProperties", fromNib: itemProperties.nib)
            case 4: showPropertyView(withIdentifier: "armorProperties", fromNib: itemProperties.nib)
            case 5: showPropertyView(withIdentifier: "armorProperties", fromNib: itemProperties.nib)
            case 6: showPropertyView(withIdentifier: "itemProperties", fromNib: itemProperties.nib)
            default: showPropertyView(withIdentifier: "itemProperties", fromNib: itemProperties.nib)
            }
            
        case "triggerProperties":
            super.selectObject(object, identifierOverride: identifierOverride)
            guard let trigger = object as? FF6MapTrigger else { break }
            super.selectObject(trigger.triggerAssembly, identifierOverride: trigger.triggerAssembly.identifier)
            showPropertyView(withIdentifier: trigger.type.identifier, fromNib: trigger.triggerAssembly.nib)

        case "monsterScript.battleScript":
            super.selectObject(object, identifierOverride: identifierOverride)
            guard let battleScriptIndex = object?.value(forKey: "scriptIndex") as? Int else { return }
            super.selectObject(rom?.value(at: battleScriptIndex, inPropertyWithKey: "battleScriptPointers") as AnyObject?, identifierOverride: identifierOverride)
            
        case "battleMonster":
            super.selectObject(object, identifierOverride: identifierOverride)
            guard let battleMonster = object as? FF6BattleMonster else { return }
            if battleMonster.battleProperties !== battleEditor?.battleProperties { return }
            battleEditor?.selectMonsterSlot(battleMonster.slot)

        case "battleCharacter":
            super.selectObject(object, identifierOverride: identifierOverride)
            guard let battleCharacter = object as? FF6BattleCharacter else { return }
            if battleCharacter.auxBattleProperties !== battleEditor?.auxBattleProperties { return }
            battleEditor?.selectCharacterSlot(battleCharacter.slot)
            
        case "characterAIProperties":
            
            if let auxBattleProperties = rom.value(forKey: "auxBattleProperties") as? ROMArray,
                let dummyBattle = auxBattleProperties.createNewAssembly() as? ROMData,
                let characterAI = object as? ROMData {

                // select a dummy battle character
                dummyBattle.setValue(characterAI.index, forKey: "characterAI")
                dummyBattle.setValue(1, forKey: "enableCharacterAI")
                let dummyCharacter = FF6BattleCharacter(rom: rom, auxBattleProperties: dummyBattle, slot: 1)
                super.selectObject(dummyCharacter)
            }
            super.selectObject(object, identifierOverride: identifierOverride)

        case "eventScript.battle":
            if let scriptCommand = object as? ROMCommand {
                let backgroundOverride = scriptCommand.intValue(forKey: "background")
                if backgroundOverride < 63 {
                    defaultBattleBackground = scriptCommand.intValue(forKey: "background")
                }
            }
            super.selectObject(object, identifierOverride: identifierOverride)

        case "attackLayer":
            super.selectObject(object, identifierOverride: identifierOverride)

            // select linked string table entry
            guard let attackLayer = object as? FF6AttackLayer else { break }
            if let stringTable = rom.value(forKey: "stringTable-attackScriptPointers") as? ROMStringTable {
                selectObject(stringTable.stringAtIndex(attackLayer.graphicsIndex), identifierOverride: "attackLayer-string")
            }
            
        case "battleBackgroundProperties":
            
            // select the tile formation
            guard let battleBackgroundProperties = object as? ROMData else {
                super.selectObject(object, identifierOverride: identifierOverride)
                break
            }
            let layoutIndex = battleBackgroundProperties.intValue(forKey: "layout1")
            guard let layout = rom.value(at: layoutIndex, inPropertyWithKey: "battleBackgroundLayouts") as? ROMTileLayout else { break }
            let graphics1 = battleBackgroundProperties.intValue(forKey: "graphics1")
            let graphics2 = battleBackgroundProperties.intValue(forKey: "graphics2")
            let graphics3 = battleBackgroundProperties.intValue(forKey: "graphics3")
            let palette = battleBackgroundProperties.intValue(forKey: "palette")
            let graphicsDefinition = NSMutableDictionary()
            let isGBA = rom.system == .GBA
            if graphics1 != 0xFF {
                graphicsDefinition[isGBA ? "0x4000" : "0x0000"] = "battleBackgroundGraphics[\(graphics1 & 0x7F)]"
            }
            if graphics2 != 0xFF {
                graphicsDefinition[isGBA ? "0x6000" : "0x2000"] = "battleBackgroundGraphics[\(graphics2 & 0x7F)]"
            }
            if graphics3 != 0xFF {
                graphicsDefinition[isGBA ? "0x2000" : "0xE000"] = "battleBackgroundGraphics[\(graphics3 & 0x7F)]"
            }
            layout.graphics = graphicsDefinition
            layout.palette = ["0x140" : "battleBackgroundPalettes[\(palette)]"]
            super.selectObject(layout)

        case "songScriptPointers":
            super.selectObject(object, identifierOverride: identifierOverride)
            guard let songScriptPointer = object as? ROMData else { return }
            let songIndex = songScriptPointer.index
            
            guard let songInstrumentsTable = rom.stringTable(forKey: "songInstruments") else { return }
            guard let songInstruments = rom.value(at: songIndex, inPropertyWithKey: "songInstruments") as? ROMData else { return }
            for instrument in 1...16 {
                let instrumentIndex = songInstruments.intValue(forKey: "instrument\(instrument)")
                let instrumentName = rom.string(atIndex: instrumentIndex, inStringTable: "instrumentBRR") ?? "Instrument \(instrumentIndex)"
                let instrumentString = songInstrumentsTable[31 + instrument] as? ROMString
                instrumentString?.rawString = instrumentName
            }
            
        default:
            super.selectObject(object, identifierOverride: identifierOverride)
        }
    }
    
    override func deselectAllObjects() {
        
        battleEditor?.deselectAll()
        
        super.deselectAllObjects()
    }
}
