//
//  ROMGraphicsView.swift
//  FF6Tools
//
//  Created by Evan Owen on 5/22/17.
//  Copyright © 2017 Evan Owen. All rights reserved.
//

import Cocoa

class ROMGraphicsView: ROMView {

    var selectedGraphics: ROMGraphics?
    var graphicsImage: NSImage?
    var imageSize: NSSize {
        if selectedGraphics == nil { return NSZeroSize }
        let tileCount = selectedGraphics!.data.count / selectedGraphics!.format.bytesPerTile
        let width = selectedGraphics!.wrapWidth * 8
        let height = (tileCount + selectedGraphics!.wrapWidth - 1) / selectedGraphics!.wrapWidth * 8
        return NSMakeSize(CGFloat(width) * zoom, CGFloat(height) * zoom)
    }
    var zoom: CGFloat = 2.0
    
    override func draw(_ dirtyRect: NSRect) {
        super.draw(dirtyRect)

        if let clipView = superview as? NSClipView, let scrollView = clipView.superview as? NSScrollView {
            scrollView.backgroundColor = NSColor.black
        }

        backgroundColor.set()
        NSBezierPath.fill(dirtyRect)

        if graphicsImage == nil {
            guard let iref = selectedGraphics?.image else { return }
            let dummyImage = NSImage(cgImage: iref, size: imageSize) // this is an NSImage scaled to the zoom factor
            graphicsImage = NSImage(size: imageSize)
            graphicsImage?.lockFocus()
            NSGraphicsContext.current?.imageInterpolation = imageInterpolation
            dummyImage.draw(in: NSMakeRect(0, 0, imageSize.width, imageSize.height))
            graphicsImage?.unlockFocus()
            if graphicsImage == nil { return }
        }
        
        NSGraphicsContext.current?.imageInterpolation = imageInterpolation
        graphicsImage?.draw(in: dirtyRect, from: dirtyRect, operation: NSCompositingOperation.sourceOver, fraction: 1.0)
    }
    
    func selectGraphics(graphics: ROMGraphics) {
        graphicsImage = nil
        selectedGraphics = graphics
        needsDisplay = true
        
        if let clipView = superview as? NSClipView {
            setFrameSize(clipView.documentVisibleRect.size)
        }
    }
    
    override func setFrameSize(_ newSize: NSSize) {
        
        if selectedGraphics == nil {
            super.setFrameSize(newSize)
            return
        }
        
        super.setFrameSize(imageSize)
    }
    
    @IBAction func exportImage(_ sender: AnyObject) {
        guard let cgImage = selectedGraphics?.image else { return }
        let bitmapRep = NSBitmapImageRep(cgImage: cgImage)
        let png = bitmapRep.representation(using: NSBitmapImageRep.FileType.png, properties: [:])
        
        let dialog = NSSavePanel()
        if dialog.runModal() == NSApplication.ModalResponse.OK, let url = dialog.url {
            do {
                try png?.write(to: url)
            } catch {
                Swift.print("Error: Unable to write to image file")
            }
        } else {
            return
        }
        
    }
    
    @IBAction func importImage(_ sender: AnyObject) {
        let dialog = NSOpenPanel()
        if dialog.runModal() != NSApplication.ModalResponse.OK { return }
        guard let url = dialog.url, let fileImage = NSImage(contentsOf: url) else { return }
        
        var imageRect = NSMakeRect(0, 0, fileImage.size.width, fileImage.size.height)
        guard let imageRef = fileImage.cgImage(forProposedRect: &imageRect, context: nil, hints: nil) else { return }
        
        // create ans RGBA context
        let width = Int(imageRect.width)
        let height = Int(imageRect.height)
        let bytesPerPixel = 4
        let bitsPerComponent = 8
        let bytesPerRow = bytesPerPixel * width
        let space = CGColorSpaceCreateDeviceRGB()
        let bitmapInfo = CGImageAlphaInfo.premultipliedLast.rawValue
        let context = CGContext(data: nil, width: width, height: height, bitsPerComponent: bitsPerComponent, bytesPerRow: bytesPerRow, space: space, bitmapInfo: bitmapInfo)

        // draw the image in the context and get the raw RGBA data
        context?.draw(imageRef, in: imageRect)
        guard let _ = context?.data else { return }
        
//        // convert the RGBA data to indexed color
//        let attr = liq_attr_create()!
//        let image = liq_image_create_rgba(attr, rgbaData, Int32(width), Int32(height), 0)!
//        
//        var res: OpaquePointer? = OpaquePointer(UnsafeMutablePointer<UInt8>.allocate(capacity: 1024))
//        let bytesPerTile = selectedGraphics?.format.bytesPerTile
//        let colorCount = 1 << (bytesPerTile! / 8)
//
//        liq_set_max_colors(attr, Int32(colorCount))
//        let quantizeError = liq_image_quantize(image, attr, &res)
//        if quantizeError.rawValue != 0 {
//            Swift.print("Error quantizing image: \(quantizeError)")
//            return
//        }
//        
//        let bufferSize = width * height
//        let buffer = UnsafeMutablePointer<UInt8>.allocate(capacity: bufferSize)
//        let remapError = liq_write_remapped_image(res!, image, buffer, bufferSize)
//        if remapError.rawValue != 0 {
//            Swift.print("Error quantizing image: \(remapError)")
//            return
//        }
//        let pal = liq_get_palette(res!)
//
//        // Save the image and the palette now.
//        for(int i=0; i < pal->count; i++) {
//            example_copy_palette_entry(pal->entries[i]);
//        }
//        // You'll need a PNG library to write to a file.
//        example_write_image(example_bitmap_8bpp);
        
//        liq_result_destroy(res!);
//        liq_image_destroy(image);
//        liq_attr_destroy(attr);
//        buffer.deinitialize()
//        buffer.deallocate(capacity: bufferSize)

        graphicsImage = fileImage
        needsDisplay = true
    }
}
