//
//  ROMMemoryView.swift
//  FF6Tools
//
//  Created by Evan Owen on 5/20/17.
//  Copyright © 2017 Evan Owen. All rights reserved.
//

import Cocoa

class ROMMemoryView: NSView {

    @IBOutlet var windowController: ROMWindowController?
    var rom: ROM! { return windowController?.rom }
    
    static var context = 0
    
    override func awakeFromNib() {
        rom.addObserver(self, forKeyPath: "data", options: NSKeyValueObservingOptions.new, context: &ROMMemoryView.context)
    }
    
    deinit {
        rom.removeObserver(self, forKeyPath: "data")
    }
    
    override func observeValue(forKeyPath keyPath: String?, of object: Any?, change: [NSKeyValueChangeKey : Any]?, context: UnsafeMutableRawPointer?) {
        if context != &ROMMemoryView.context {
            super.observeValue(forKeyPath: keyPath, of: object, change: change, context: context)
            return
        }
        
        needsDisplay = true
    }
        
    override func draw(_ dirtyRect: NSRect) {
        super.draw(dirtyRect)

        // start with the entire ROM, then subtract out unused memory
        var unknownMemory = IndexSet(integersIn: 0..<rom.data.count)
        unknownMemory.formSymmetricDifference(rom.unusedMemory)
        
        // then subtract out assemblies in the memory map
        var usedMemory = IndexSet()
        for (_, object) in rom.memoryMap {
            guard let definition = object as? NSDictionary,
                let range = MakeRange(object: definition, keyPath: "range") else { continue }
            if definition["references"] == nil { continue }
            let mappedRange = rom.mapMode.mapRange(range)
            unknownMemory.remove(integersIn: mappedRange)
            usedMemory.insert(integersIn: mappedRange)
        }
        
        // split the memory map into enough chunks to match the width of the view
        let numChunks = Int(frame.width) * 4
        let chunkSize = rom.data.count / numChunks
        
        for (index, chunkStart) in stride(from: 0, to: rom.data.count, by: chunkSize).enumerated() {
            let chunkEnd = chunkStart + chunkSize
            
            let unknownCount = unknownMemory.count(in: chunkStart..<chunkEnd)
            let usedCount = usedMemory.count(in: chunkStart..<chunkEnd)
            let unusedCount = rom.unusedMemory.count(in: chunkStart..<chunkEnd)
            
            switch max(unknownCount, usedCount, unusedCount) {
            case unknownCount: NSColor.green.set()
            case usedCount: NSColor.blue.set()
            case unusedCount: NSColor.white.set()
            default: break
            }
            
            let startPoint = NSMakePoint(CGFloat(index) / 4, 0)
            let endPoint = NSMakePoint(CGFloat(index) / 4, frame.height)
            
            NSBezierPath.strokeLine(from: startPoint, to: endPoint)
        }
    }
}
