//
//  ROMNavigatorController.swift
//  FF6Tools
//
//  Created by Evan Owen on 8/6/16.
//  Copyright © 2016 Evan Owen. All rights reserved.
//

import Cocoa

class ROMNavigatorController: NSObject, NSOutlineViewDelegate, NSOutlineViewDataSource {

    @IBOutlet weak var windowController: ROMWindowController!
    @IBOutlet var navigatorView: NSOutlineView!
    
    var rom: ROM { return windowController.rom }
    
    override func awakeFromNib() {
        navigatorView.action = #selector(ROMNavigatorController.clickedTable(_:))
        navigatorView.target = self
    }
    
    func outlineView(_ outlineView: NSOutlineView, numberOfChildrenOfItem item: Any?) -> Int {
        if item == nil {
            return rom.hierarchy?.count ?? 1
        } else if let header = item as? NSDictionary, let assemblies = header["assemblies"] as? NSArray {
            return assemblies.count
        }
        return 0
    }
    
    func outlineView(_ outlineView: NSOutlineView, isItemExpandable item: Any) -> Bool {
        if let header = item as? NSDictionary, let _ = header["assemblies"] as? NSArray {
            return true
        }
        return false
    }
    
    func outlineView(_ outlineView: NSOutlineView, child index: Int, ofItem item: Any?) -> Any {
        if item == nil {
            if let header = rom.hierarchy?[index] as? NSDictionary {
                return header
            }
        } else if let header = item as? NSDictionary,
            let assemblies = header["assemblies"] as? NSArray {
            
            if let assemblyIdentifier = assemblies[index] as? String {
                return rom.value(forKeyPath: assemblyIdentifier) ?? assemblyIdentifier
                
            } else if let hierarchyDefinition = assemblies[index] as? NSDictionary {
                return hierarchyDefinition
            }
        }
        return "No Assemblies Found"
    }
    
    func outlineView(_ outlineView: NSOutlineView, objectValueFor tableColumn: NSTableColumn?, byItem item: Any?) -> Any? {

        if let definition = item as? NSDictionary {
            if let name = definition["name"] as? String {
                return name
                
            } else if let identifier = definition["identifier"] as? String,
                let assemblyDefinition = rom.memoryMap[identifier] as? NSDictionary {
                return assemblyDefinition["name"] as? String ?? "Unnamed Assembly"
                
            } else if let identifier = definition["identifier"] as? String,
                let stringTable = rom.stringTable(forKey: identifier) {
                return stringTable.name ?? "Unnamed String Table"
                
            } else if let identifier = definition["identifier"] as? String,
                let object = rom.value(forKeyPath: identifier) as? ROMObject {
                return object.name ?? "Unnamed Object"
            }
            
        } else if let object = item as? ROMObject {
            return object.name ?? "Unnamed Assembly"
            
        } else if let string = item as? String {
            return string
        }
        
        return nil
    }
    
    func select(object: Any) {
        
        if let selectedObject = object as? ROMObject {
            if windowController.objectIsSelected(selectedObject) { return }
            windowController.deselectAllObjects()
            windowController.selectObject(selectedObject)
            
        } else if let hierarchyDefinition = object as? NSDictionary,
            let assemblyIdentifier = hierarchyDefinition["identifier"] as? String,
            let selectedObject = rom.value(forKeyPath: assemblyIdentifier) as? ROMObject {
            if windowController.objectIsSelected(selectedObject) { return }
            windowController.deselectAllObjects()
            windowController.selectObject(selectedObject)
        }
    }
    
    @IBAction func clickedTable(_ sender: AnyObject) {
        select(object: navigatorView.item(atRow: navigatorView.clickedRow) as Any)
    }
    
    func outlineViewSelectionDidChange(_ notification: Notification) {

        guard let tableView = notification.object as? NSOutlineView else { return }
        select(object: tableView.item(atRow: tableView.selectedRow) as Any)
    }
    
    func outlineView(_ outlineView: NSOutlineView, isGroupItem item: Any) -> Bool {
        if let definition = item as? NSDictionary, definition["assemblies"] != nil { return true }
        return false
    }
    
    func outlineView(_ outlineView: NSOutlineView, shouldSelectItem item: Any) -> Bool {
        if let definition = item as? NSDictionary, definition["assemblies"] != nil { return true }
        return true
    }
}
