//
//  ROMSettingsController.swift
//  FF6Tools
//
//  Created by Evan Owen on 5/20/17.
//  Copyright © 2017 Evan Owen. All rights reserved.
//

import Cocoa

class ROMSettingsController: NSObject {

    @IBOutlet var windowController: ROMWindowController?
    var rom: ROM! { return windowController?.rom }

    @objc var sizeString: String {
        let size = rom.data.count
        let mbit = Double(size) / 1024 / 1024 * 8
        return String(format: "%d bytes (%.1f MBit)", size, mbit)
    }

    @objc var unmovableString: String {
        // start with the entire ROM, then subtract out unused memory
        var unmovableMemory = IndexSet(integersIn: 0..<rom.data.count)
        unmovableMemory.formSymmetricDifference(rom.unusedMemory)
        
        // then subtract out assemblies in the memory map
        for (_, object) in rom.memoryMap {
            guard let definition = object as? NSDictionary,
                let range = MakeRange(object: definition, keyPath: "range") else { continue }
            if definition["references"] == nil { continue }
            let mappedRange = rom.mapMode.mapRange(range)
            unmovableMemory.remove(integersIn: mappedRange)
        }
        
        return String(format: "%d bytes (%.1f%%)", unmovableMemory.count, Double(unmovableMemory.count) / Double(rom.data.count) * 100.0)
    }

    @objc var movableString: String {
        // then subtract out assemblies in the memory map
        var movableMemory = IndexSet()
        for (_, object) in rom.memoryMap {
            guard let definition = object as? NSDictionary,
                let range = MakeRange(object: definition, keyPath: "range") else { continue }
            if definition["references"] == nil { continue }
            let mappedRange = rom.mapMode.mapRange(range)
            movableMemory.insert(integersIn: mappedRange)
        }
        
        return String(format: "%d bytes (%.1f%%)", movableMemory.count, Double(movableMemory.count) / Double(rom.data.count) * 100.0)
    }

    @objc var unusedString: String {
        let unusedSize = rom.unusedMemory.count
        return String(format: "%d bytes (%.1f%%)", unusedSize, Double(unusedSize) / Double(rom.data.count) * 100.0)
    }
    
    @objc var isExpandable: Bool {
        if rom.system == .SFC && rom.data.count == 0x300000 {
            return true
        } else if rom.system == .GBA {
            return false
        }
        return false
    }
    
    @IBAction func expandROM(_ sender: AnyObject) {
        willChangeValue(forKey: "isExpandable")
        willChangeValue(forKey: "sizeString")
        willChangeValue(forKey: "unusedString")
        willChangeValue(forKey: "unmovableString")
        willChangeValue(forKey: "movableString")
        rom.expand(toLength: rom.data.count + 0x100000)
        didChangeValue(forKey: "isExpandable")
        didChangeValue(forKey: "sizeString")
        didChangeValue(forKey: "unusedString")
        didChangeValue(forKey: "unmovableString")
        didChangeValue(forKey: "movableString")
    }
    
    @IBAction func romMap(_ sender: AnyObject) {
        rom.createROMMap()
    }

    @IBAction func relocateAll(_ sender: AnyObject) {
        
        var allAssemblies = [(assembly: ROMAssembly, length: Int)]()
        for key in rom.memoryMap.allKeys {
            guard let identifier = key as? String else { continue }
            guard let assembly = rom.value(forKey: identifier) as? ROMAssembly else { continue }
            
            // don't relocate anything that has no references or adjacent assemblies
            if assembly.nextAssembly == nil && assembly.previousAssembly == nil && assembly.references == nil { continue }
            allAssemblies.append((assembly, assembly.assembledLength))
            
            // include the pointer table if there is one
            if let pointerTable = assembly.pointerTable,
                pointerTable.range != nil {
                allAssemblies.append((pointerTable, pointerTable.assembledLength))
            }
        }

        willChangeValue(forKey: "isExpandable")
        willChangeValue(forKey: "sizeString")
        willChangeValue(forKey: "unusedString")
        willChangeValue(forKey: "unmovableString")
        willChangeValue(forKey: "movableString")
        let _ = rom.autoRelocate(allAssemblies)
        didChangeValue(forKey: "isExpandable")
        didChangeValue(forKey: "sizeString")
        didChangeValue(forKey: "unusedString")
        didChangeValue(forKey: "unmovableString")
        didChangeValue(forKey: "movableString")
    }
    
    @IBAction func openLog(_ sender: AnyObject) {
        NSWorkspace.shared.open(rom.logURL)
    }
}
