//
//  ROMWindowController.swift
//  FF6Tools
//
//  Created by Evan Owen on 2/28/16.
//  Copyright © 2016 Evan Owen. All rights reserved.
//

import Cocoa

// TODO: move these to a more generic file

func releaseImageData(dataInfo: UnsafeMutableRawPointer?, data: UnsafeRawPointer, length: Int) { }

class FlippedScrollView: NSScrollView {
    
    override var isFlipped: Bool { return true }
}

class FlippedClipView: NSClipView {
    
    override var isFlipped: Bool { return true }
}

class NoScrollView: NSScrollView {
    override func scrollWheel(with theEvent: NSEvent) {
        nextResponder?.scrollWheel(with: theEvent)
    }
}

class ROMPropertyView: NSView {
    @objc var assemblyIdentifier: String?
    @objc var sizeToFit: Bool = false
}

class ROMView: NSView {

    @IBOutlet weak var parentView: NSView?
    @IBOutlet var windowController: ROMWindowController!
    var rom: ROM { return windowController.rom }

    @objc var backgroundColor: NSColor {
        if let userColorData = UserDefaults.standard.data(forKey: backgroundColorKey),
            let userColor = NSUnarchiver.unarchiveObject(with: userColorData) as? NSColor {
            return userColor
        }
        return NSColor.orange
    }
    
    var imageInterpolation: NSImageInterpolation {
        let userInterpolationInt = UserDefaults.standard.integer(forKey: imageInterpolationKey)
        return NSImageInterpolation(rawValue: UInt(userInterpolationInt)) ?? NSImageInterpolation.none
    }

    override func awakeFromNib() {
        super.awakeFromNib()
        NSUserDefaultsController.shared.addObserver(self, forKeyPath: "values." + backgroundColorKey, options: NSKeyValueObservingOptions.new, context: &ROMView.context)
        NSUserDefaultsController.shared.addObserver(self, forKeyPath: "values." + imageInterpolationKey, options: NSKeyValueObservingOptions.new, context: &ROMView.context)
    }
    
    deinit {
        NSUserDefaultsController.shared.removeObserver(self, forKeyPath: "values." + backgroundColorKey)
        NSUserDefaultsController.shared.removeObserver(self, forKeyPath: "values." + imageInterpolationKey)
    }
    
    static var context = 0
    
    override func observeValue(forKeyPath keyPath: String?, of object: Any?, change: [NSKeyValueChangeKey : Any]?, context: UnsafeMutableRawPointer?) {
        if context != &ROMView.context {
            super.observeValue(forKeyPath: keyPath, of: object, change: change, context: context)
            return
        }
        
        // redraw the view any time the user preferences change
        needsDisplay = true
    }
}

class ROMWindowController: NSWindowController, NSSplitViewDelegate, NSWindowDelegate {

    @objc var rom: ROM!
    
    @IBOutlet weak var splitView: NSSplitView?
    @IBOutlet weak var listView: NSTabView?
    @IBOutlet weak var propertyView: NSScrollView?
    @IBOutlet weak var editorView: NSTabView?
    @IBOutlet weak var toolboxView: NSTabView?
    @IBOutlet var logView: NSTextView!

    @IBOutlet weak var navigatorController: ROMNavigatorController?
    @IBOutlet weak var scriptController: ROMScriptController?
    @IBOutlet weak var tableController: ROMTableController?
    @IBOutlet weak var graphicsView: ROMGraphicsView?
    @IBOutlet weak var tileView: ROMTileView?
    @IBOutlet weak var tilesetView: ROMTilesetView?
    
    fileprivate var propertyViews = [String : NSView]()
    fileprivate var selectedObjects = [String : ROMSelectable]()
    var selectedIndex = 0
    
    override var undoManager: UndoManager? { return document?.undoManager }
    
    @IBOutlet weak var settingsWindow: NSWindow!
    
    @IBAction func showSettings(_ sender: AnyObject) {
        window?.beginSheet(settingsWindow, completionHandler: nil)
    }
    
    @IBAction func hideSettings(_ sender: AnyObject) {
        window?.endSheet(settingsWindow)
    }

    deinit {
        deselectAllObjects()
    }
    
    override func windowDidLoad() {
        super.windowDidLoad()

        // Implement this method to handle any initialization after your window controller's window has been loaded from its nib file.
        
        // this is a bullshit problem - I found the answer here: http://prod.lists.apple.com/archives/cocoa-dev/2015/Oct/msg00242.html
        // i fixed this by checking a box in IB - more info here: http://stackoverflow.com/questions/34111483/nssplitview-not-resizing-subviews-after-dynamically-adding-a-subview
//        splitView.arrangesAllSubviews = true
        
        // add the "No Selection" to the dictionary of properties views
        // TODO: create the noSelection view programatically to ensure that it exists
        addPropertyViews(fromNib: "ROMWindowController")
        propertyView?.documentView = propertyViews["noSelection"]
    }
    
    func showPropertyView(withIdentifier identifier: String, fromNib nibName: String?) {
        // get the appropriate property view
        guard let newPropertyView = propertyView(withIdentifier: identifier, fromNib: nibName) else { return }
        
        // set the new view's width to match the old view
        if let oldView = propertyView?.documentView {
            var newSize = NSMakeSize(oldView.frame.size.width, newPropertyView.frame.height)
            
            // change the height of the view to fit the view
            if (newPropertyView as? ROMPropertyView)?.sizeToFit == true {
                newSize.height = oldView.superview!.frame.height
            }
            
            newPropertyView.setFrameSize(newSize)
            newPropertyView.needsLayout = true
        }
        
        // show the new view
        propertyView?.documentView = newPropertyView
    }
    
    func showPropertyView(forObject object: ROMSelectable) {
        // get the appropriate property view
        guard let newPropertyView = propertyViewForObject(object) else { return }
            
        // set the new view's width to match the old view
        if let oldView = propertyView?.documentView {
            var newSize = NSMakeSize(oldView.frame.size.width, newPropertyView.frame.height)
            
            // change the height of the view to fit the view
            if (newPropertyView as? ROMPropertyView)?.sizeToFit == true {
                newSize.height = oldView.superview!.frame.height
            }
            
            newPropertyView.setFrameSize(newSize)
            newPropertyView.needsLayout = true
        }
        
        // show the new view
        propertyView?.documentView = newPropertyView
    }
    
    func propertyView(withIdentifier identifier: String, fromNib nibName: String?) -> NSView? {
        
        if let loadedView = propertyViews[identifier] {
            return loadedView
        } else if nibName != nil {
            addPropertyViews(fromNib: nibName!)
            return propertyViews[identifier]
        }
        return nil
    }
    
    func propertyViewForObject(_ object: AnyObject) -> NSView? {
        guard let selectable = object as? ROMSelectable,
            let identifier = selectable.identifier else { return nil }
        
        if selectedObjects["text"] === object { return propertyViews["text"] }
        if selectedObjects["string"] === object { return propertyViews["string"] }
        if selectedObjects["graphics"] === object { return propertyViews["graphics"] }
        
        return propertyView(withIdentifier: identifier, fromNib: selectable.nib)
    }
    
    func addPropertyViews(fromNib nib: String) {
        var objects: NSArray?
        rom.log("Loading nib file: " + nib)
        Bundle.main.loadNibNamed(NSNib.Name(rawValue: nib), owner: self, topLevelObjects: &objects)
        if objects == nil { return }
        
        // add all top-level NSView objects to the propertiesViews dictionary
        for object in objects! {
            guard let newView = object as? NSView, let identifier = newView.identifier else { continue }
            
            if let propertyView = newView as? ROMPropertyView,
                let assemblyIdentifier = propertyView.assemblyIdentifier {
                propertyViews[assemblyIdentifier] = propertyView
            } else {
                propertyViews[identifier.rawValue] = newView
            }
        }
    }

    // MARK: selection methods
    
    func selectObject(_ object: AnyObject?, identifierOverride: String? = nil) {

        if objectIsSelected(object) {
            // object is already selected
            return
        }
        
        if object == nil {
            // clear selection
            deselectAllObjects()
            
        } else if (selectedObjects.filter{ $0.value === object }.count) != 0 {
            // return if an object with the same identifier is already selected
            return
            
        } else if let string = object as? ROMString, selectedObjects.count == 0 {
            
            // select a single generic ROMString
            willChangeValue(forKey: "string")
            selectedObjects["string"] = string
            startObserving(object: string, identifierOverride: "string")
            didChangeValue(forKey: "string")
            showPropertyView(forObject: string)
            
        } else if let command = object as? ROMCommand {
            
            selectObject(command.script)
            
            // select the new object and start observing it
            willChangeValue(forKey: "scriptCommand")
            selectedObjects["scriptCommand"] = command
            startObserving(object: command, identifierOverride: "scriptComand")
            didChangeValue(forKey: "scriptCommand")
            
            scriptController?.select(command: command)
            showPropertyView(forObject: command)
            
        } else if let text = object as? ROMText, text.nib == nil && selectedObjects.count == 0 {
            
            // select a single generic ROMText
            willChangeValue(forKey: "text")
            selectedObjects["text"] = text
            startObserving(object: text, identifierOverride: "text")
            didChangeValue(forKey: "text")
            
            showPropertyView(forObject: text)

        } else if let graphics = object as? ROMGraphics, graphics.nib == nil && selectedObjects.count == 0 {
            
            // select a single generic ROMGraphics
            willChangeValue(forKey: "graphics")
            selectedObjects["graphics"] = graphics
            startObserving(object: graphics, identifierOverride: "graphics")
            didChangeValue(forKey: "graphics")
            
            showPropertyView(forObject: graphics)

            graphicsView?.selectGraphics(graphics: graphics)
            editorView?.tabViewItem(at: 0).view = graphicsView?.parentView

        } else if let tileLayout = object as? ROMTileLayout, tileLayout.nib == nil && selectedObjects.count == 0 {
            
            // select a single generic ROMTileLayout
            willChangeValue(forKey: "tileLayout")
            selectedObjects["tileLayout"] = tileLayout
            startObserving(object: tileLayout, identifierOverride: "tileLayout")
            didChangeValue(forKey: "tileLayout")
            
            showPropertyView(forObject: tileLayout)
            tileView?.select(tileLayout: tileLayout)
            editorView?.tabViewItem(at: 0).view = tileView?.parentView
            
            tilesetView?.select(tileLayout: tileLayout)
            toolboxView?.tabViewItem(at: 0).view = tilesetView?.parentView

        } else if let stringTable = object as? ROMStringTable {
            
            // select a string table
            tableController?.selectObject(stringTable)
            
            if let tableTabIndex = listView?.indexOfTabViewItem(withIdentifier: "tableTab") {
                listView?.tabViewItem(at: tableTabIndex).view = tableController?.parentView
                listView?.selectTabViewItem(withIdentifier: "tableTab")
            }
            
        } else if let array = object as? ROMArray, selectedObjects.count == 0 {
            
            // select a ROMArray
            tableController?.selectObject(array)
            
            if let tableTabIndex = listView?.indexOfTabViewItem(withIdentifier: "tableTab") {
                listView?.tabViewItem(at: tableTabIndex).view = tableController?.parentView
                listView?.selectTabViewItem(withIdentifier: "tableTab")
            }
            
            // TODO: select the array in the navigator view as well
            
        } else if let script = object as? ROMScript {
            
            // select the script in the list view
            scriptController?.script = script
            if let tableTabIndex = listView?.indexOfTabViewItem(withIdentifier: "scriptTab") {
                listView?.tabViewItem(at: tableTabIndex).view = scriptController?.parentView
                listView?.selectTabViewItem(withIdentifier: "scriptTab")
            }
            
        } else if let object = object as? ROMSelectable, let identifier = identifierOverride ?? object.identifier {
            
            willChangeValue(forKey: identifier)
            selectedObjects[identifier] = object
            startObserving(object: object, identifierOverride: identifier)
            didChangeValue(forKey: identifier)
            
            // sub-assemblies for ROMData objects
            if let romData = object as? ROMData {
                for (_, subAssembly) in romData.subAssemblies {
                    selectObject(subAssembly as AnyObject?)
                }
            }
            
            showPropertyView(forObject: object)
            
        } else {
            rom.log("Unable to select object")
        }
        
        // select linked objects
        if let assembly = object as? ROMAssembly,
            let parentArray = assembly.parentArray {
            let index = parentArray.array.index(of: assembly)
            if index != NSNotFound {
                
                selectedIndex = index
                
                guard let object = parentArray[index] as? ROMSelectable else { return }
                
                // select linked arrays
                if let linkedArrays = parentArray.linkedArrays as? [String] {
                    for linkIdentifier in linkedArrays {
                        guard let linkedObject = rom.value(at: index, inPropertyWithKey: linkIdentifier) as? ROMSelectable else { continue }
                        selectObject(linkedObject)
                    }
                }
                
                // select linked string table entry
                if let stringTableIdentifier = parentArray.linkedStringTable, let assemblyIdentifier = object.identifier,
                    let stringTable = rom.value(forKey: "stringTable-" + stringTableIdentifier) as? ROMStringTable {
                    selectObject(stringTable.stringAtIndex(index), identifierOverride: assemblyIdentifier + "-string")
                }
            }
        }
    }

    var previousSelectionStack = [[String : ROMSelectable]]()
    var nextSelectionStack = [[String : ROMSelectable]]()
    
    @IBAction func changeSelection(sender: AnyObject) {
        guard let button = sender as? NSSegmentedControl else { return }
        if button.selectedSegment == 0 {
            // back
            selectPrevious()
        } else {
            // forward
            selectNext()
        }
    }
    
    func selectNext() {
        if nextSelectionStack.isEmpty { return }
        let nextSelection = nextSelectionStack.last!
        nextSelectionStack.removeLast()
        deselectAllObjects()
        
        for (identifier, object) in nextSelection {
            selectObject(object, identifierOverride: identifier)
        }
    }
    
    func selectPrevious() {
        if previousSelectionStack.isEmpty { return }
        let previousSelection = previousSelectionStack.last!
        deselectAllObjects()
        nextSelectionStack.append(previousSelectionStack.last!)
        previousSelectionStack.removeLast()
        previousSelectionStack.removeLast()
        
        for (identifier, object) in previousSelection {
            selectObject(object, identifierOverride: identifier)
        }
    }
    
    func deselectAllObjects() {
        
        // save the previous selection
        previousSelectionStack.append(selectedObjects)
        
        // clear the selected index
        selectedIndex = 0
        
        for (identifier, object) in selectedObjects {
            deselectObject(object, identifierOverride: identifier)
        }
        
        // set the new view's width to match the old view
        guard let noSelectionView = propertyViews["noSelection"] else { return }
        
        if let oldView = propertyView?.documentView {
            let newSize = NSMakeSize(oldView.frame.size.width, noSelectionView.frame.height)
            noSelectionView.setFrameSize(newSize)
        }
        
        // show the no selection view
        propertyView?.documentView = noSelectionView
    }
    
    func deselectObject(_ object: AnyObject?, identifierOverride: String? = nil) {
        
        // return if the object is not selectable or if it is not currently selected
        guard let selected = object as? ROMSelectable,
            let identifier = identifierFor(object: object) ?? selected.identifier else { return }

        willChangeValue(forKey: identifier)
        stopObserving(object: object, identifierOverride: identifier)
        selectedObjects[identifier] = nil
        didChangeValue(forKey: identifier)
    }
    
    func objectIsSelected(_ object: AnyObject?) -> Bool {
        
        for (_, selectedObject) in selectedObjects {
            if selectedObject === object { return true }
        }
        return false
    }
    
    func identifierFor(object: AnyObject?) -> String? {
        for (identifier, selectedObject) in selectedObjects {
            if selectedObject === object { return identifier }
        }
        return nil
    }
    
    // MARK: observing methods

    func startObserving(object: AnyObject?, identifierOverride: String? = nil) {

        // return if the object is not selectable or if it is not currently selected
        if !objectIsSelected(object) { return }
        guard let selected = object as? ROMSelectable else { return }
        
        // properties for ROMSelectable objects
        if let properties = selected.properties {
            for (key, _) in properties {
                guard let propertyIdentifier = key as? String else { continue }
                object?.addObserver(self, forKeyPath: propertyIdentifier, options: NSKeyValueObservingOptions.old, context: &ROMWindowController.context)
            }
        }
    }

    func stopObserving(object: AnyObject?, identifierOverride: String? = nil) {

        // return if the object is not selectable or if it is not currently selected
        if !objectIsSelected(object) { return }
        guard let selected = object as? ROMSelectable else { return }

        // properties for ROMSelectable objects
        if let properties = selected.properties {
            for (key, _) in properties {
                guard let propertyIdentifier = key as? String else { continue }
                object?.removeObserver(self, forKeyPath: propertyIdentifier, context: &ROMWindowController.context)
            }
        }
    }
    
    func startObservingAllObjects() {
        // temporarily stop observing all objects
        for (identifier, object) in selectedObjects {
            startObserving(object: object, identifierOverride: identifier)
        }
    }
    
    func stopObservingAllObjects() {
        // temporarily stop observing all objects
        for (identifier, object) in selectedObjects {
            stopObserving(object: object, identifierOverride: identifier)
        }
    }

    static var context = 0
    
    override func value(forKey key: String) -> Any? {
        if let object = selectedObjects[key] {
            return object
        } else if key.hasPrefix("stringTable-") {
            if let stringTable = rom.value(forKey: key) as? ROMStringTable {
                return stringTable.stringArray
            }
        }
        return super.value(forKey: key)
    }
    
    override func value(forUndefinedKey key: String) -> Any? {

        return nil
    }
    
    @objc func setProperty(_ propertyName: String, toValue value: Any, assembly: AnyObject) {
        
        // save undo data
        if let undo = undoManager,
            let oldValue = assembly.value(forKeyPath: propertyName) {
            let target = undo.prepare(withInvocationTarget: self) as AnyObject
            if let dataValue = oldValue as? Data {
                rom.log("Old value = \(dataValue.count) bytes (\(propertyName))")
                if (!undo.isUndoing) {
                    undo.setActionName("Edit Data (\(dataValue))")
                }
            } else {
                rom.log("Old value = \(oldValue) (\(propertyName))")
                if (!undo.isUndoing) {
                    undo.setActionName("Edit Property (\(propertyName))")
                }
            }
            target.setProperty(propertyName, toValue: oldValue, assembly: assembly)
        }
        
        // temporarily stop observing all objects so we don't trigger another undo session
        stopObservingAllObjects()
        
        // set the property value
        assembly.setValue(value, forKey: propertyName)
        
        // resume observing all objects
        startObservingAllObjects()
    }
    
    override func observeValue(forKeyPath keyPath: String?, of object: Any?, change: [NSKeyValueChangeKey : Any]?, context: UnsafeMutableRawPointer?) {
        if context != &ROMWindowController.context {
            super.observeValue(forKeyPath: keyPath, of: object, change: change, context: context)
            return
        }
        
        // save undo data
        if let undo = undoManager,
            let oldValue = change?[NSKeyValueChangeKey.oldKey] {
            let target = undo.prepare(withInvocationTarget: self) as AnyObject
            if let dataValue = oldValue as? Data {
                rom.log("Old value = \(dataValue.count) bytes (\(keyPath!))")
                if (!undo.isUndoing) {
                    undo.setActionName("Edit Data (\(dataValue))")
                }
            } else {
                rom.log("Old value = \(oldValue) (\(keyPath!))")
                if (!undo.isUndoing) {
                    undo.setActionName("Edit Property (\(keyPath!))")
                }
            }
            target.setProperty(keyPath!, toValue: oldValue, assembly: object! as AnyObject)
        }
    }
    
    // MARK: NSSplitViewDelegate methods
    
    func splitViewDidResizeSubviews(_ notification: Notification) {
        
        if let currentView = propertyView?.documentView {
            
            // change the view's width to match the superview
            var viewSize = currentView.frame.size
            viewSize.width = currentView.superview!.frame.width

            // change the height of the view to fit the view
            if (currentView as? ROMPropertyView)?.sizeToFit == true {
                viewSize.height = currentView.superview!.frame.height
            }
            
            // set the frame size for the current view
            currentView.setFrameSize(viewSize)
            currentView.needsLayout = true
        }
    }
    
    @IBAction func clickedSplitButton(_ sender: AnyObject) {
        
        if splitView == nil { return }
        guard let button = sender as? NSButton, let buttonIdentifier = button.identifier else { return }
        
        for view in splitView!.subviews {
            guard let viewIdentifier = view.identifier else { continue }
            if buttonIdentifier.rawValue.hasPrefix(viewIdentifier.rawValue) {
                togglePane(view)
            }
        }
    }
    
    func togglePane(_ paneView: NSView) {
        if splitView == nil { return }
        if splitView!.isSubviewCollapsed(paneView) {
            paneView.isHidden = false
            splitView?.display()
        } else {
            paneView.isHidden = true
            splitView?.display()
        }
    }
}
