//
//  SongScript.swift
//  FF6Tools
//
//  Created by Evan Owen on 4/28/17.
//
//

import Foundation

class SongScriptEncoding: ROMScriptEncoding {
    
    override func willDisassembleScript(_ script: ROMScript) {
        
        if script.identifier == "songScript" {
            for (index, object) in script.rom.mutableArrayValue(forKey: "songScriptPointers").enumerated() {
                guard let scriptPointerAssembly = object as? ROMAssembly,
                    let scriptPointer = scriptPointerAssembly.value(forKeyPath: "scriptPointer") as? Int else { continue }
                
                // make a label for each song
                let eventName = script.rom.string(atIndex: index, inStringTable: "songs") ?? ("Song " + String(index))
                let mappedPointer = script.rom.mapMode.mapAddress(scriptPointer)
                script.commandLabels[mappedPointer] = eventName
                
                // create a reference for each song
                let ref = script.addReference(forAssembly: scriptPointerAssembly, encodingName: "songScriptHeader", linkedKeyPath: "scriptPointer", atAbsoluteOffset: mappedPointer)
                ref?.isMapped = false
            }
        } else if script.identifier == "soundEffectScript" {
            for (index, object) in script.rom.mutableArrayValue(forKey: "soundEffectPointers").enumerated() {
                guard let scriptPointerAssembly = object as? ROMAssembly,
                    let scriptPointer = scriptPointerAssembly.value(forKeyPath: "scriptPointer") as? Int else { continue }
                
                // skip blank pointers
                if scriptPointer == 0 { continue }
                
                // make a label for each sound effect
                let absoluteOffset = script.range!.lowerBound + scriptPointer - 0x3000
                let channelString = index % 2 == 0 ? " (A)" : " (B)"
                let eventName = script.rom.string(atIndex: index / 2, inStringTable: "gameSoundEffects") ?? ("Sound Effect " + String(index / 2))
                script.commandLabels[absoluteOffset] = eventName + channelString
                
                // create a reference for each song
                let ref = script.addReference(forAssembly: scriptPointerAssembly, encodingName: "songScript", linkedKeyPath: "scriptPointer", atAbsoluteOffset: absoluteOffset)
                ref?.offset = 0x3000
            }
        }
        
        super.willDisassembleScript(script)
    }

    override func willAssembleCommand(_ command: ROMCommand) {
        
        switch command.identifier! {
            
        case "songScriptHeader.header":
            
            // the next song header marks the end of this song
            let siblings = command.siblings
            var currentIndex = siblings.index(of: command) + 1
            var songLength = command.assembledLength
            while currentIndex < siblings.count {
                guard let currentCommand = siblings[currentIndex] as? ROMCommand else { continue }
                if currentCommand.encodingName == "songScriptHeader" {
                    break
                } else {
                    songLength += currentCommand.assembledLength
                }
                currentIndex += 1
            }

            // set the length and song location values
            command.setValue(songLength - 2, forKey: "length")
            command.setValue(command.range!.upperBound, forKey: "start")
            command.setValue(command.range!.lowerBound + songLength, forKey: "end")
            
        default: break
        }
    }

    override func didDisassembleCommand(_ command: ROMCommand, fromData data: Data) {

        let offset = command.range!.lowerBound

        func fixPointer(_ pointer: Int) -> Int {
            // vectors can't be more than 0x2C00 bytes (this fixes jump addresses that go from, say, 0xC8FE00 to 0xC90000)
            if (offset - pointer) > 0x2C00 {
                return pointer + 0x10000
            } else if (pointer - offset) > 0x2C00 {
                return pointer - 0x10000
            } else if abs(offset - pointer) > 0x2C00 {
                command.rom.log("Invalid pointer")
                return 0
            } else {
                return pointer
            }
        }
        
        func createReference(forKey key: String, encoding: String = "songScript") {
            let pointer = command.intValue(forKey: key) | (offset & 0xFFFF0000)
            command.script.addReference(forAssembly: command, encodingName: encoding, linkedKeyPath: key, atAbsoluteOffset: fixPointer(pointer))
        }

        switch command.identifier! {
        case "songScriptHeader.header":
            
//            createReference(forKey: "start")
            createReference(forKey: "end", encoding: "songScriptHeader")
            for voice in 0..<8 {
                
                // make a label for each voice
                let voiceName = String(format: "%@: Voice %d", command.label, voice)
                let voicePointer = command.intValue(forKey: "voice" + String(voice))
                let bank = command.range!.lowerBound & 0xFFFF0000
                let absolutePointer = fixPointer(voicePointer | bank)
                if command.script.commandLabels[absolutePointer] == nil {
                    // don't overwrite an existing label
                    command.script.commandLabels[absolutePointer] = voiceName
                }

                // and the alternate start pointer if there is one
                let voicePointerAlt = command.intValue(forKey: "voice" + String(voice) + "Alt")
                let absolutePointerAlt = fixPointer(voicePointerAlt | bank)
                if command.script.commandLabels[absolutePointerAlt] == nil {
                    // don't overwrite an existing label
                    command.script.commandLabels[absolutePointerAlt] = voiceName + " (Alternate)"
                }

                // create references for each voice
                createReference(forKey: "voice" + String(voice))
                createReference(forKey: "voice" + String(voice) + "Alt")
            }
            
        case "songScript.repeat":
            // repeat
            var currentOffset = command.range!.upperBound
            while currentOffset < command.script.range!.upperBound {
                guard let repeatCommand = ROMCommand(rom, script: command.script, encodingName: command.childEncodingName, data: data, offset: currentOffset) else { break }
                repeatCommand.parent = command
                command.childCommands.add(repeatCommand)
                currentOffset += repeatCommand.assembledLength
                if repeatCommand.identifier == "songScript.soundEffect" { break }
                if repeatCommand.identifier == "songScript.endRepeat" { break }
                if repeatCommand.identifier == "songScript.jump" { break }
                if repeatCommand.identifier == "songScript.endOfScript" { break }
            }
            if command.childCommands.count == 0 {
                // if the child array is empty, add an end of script command
                guard let endOfScriptCommand = ROMCommand(rom, script: command.script, identifier: "songScript.endRepeat") else { break }
                endOfScriptCommand.parent = command
                command.childCommands.add(endOfScriptCommand)
            }

        case "songScript.jump", "songScript.conditionalJump", "songScript.jumpEnding":
            createReference(forKey: "scriptPointer")
            
        default: break
        }
    }
    
    override func encodingNameForCommandAfter(_ command: ROMCommand) -> String {
        return "songScript"
    }
    
    override func valueForCommand(_ command: ROMCommand, forKey key: String) -> Any? {
        switch command.identifier! {
        case "songScript.note":
            if key == "note" {
                let opcode = command.intValue(forKey: "opcode")
                return opcode / 14
            } else if key == "duration" {
                let opcode = command.intValue(forKey: "opcode")
                return opcode % 14
            }
            
        case "songScript.pan", "songScript.panEnvelope":
            if key == "panString" {
                let pan = command.intValue(forKey: "pan")
                if pan == 0x40 {
                    return "0 (C)"
                } else if pan <= 0x3F {
                    return String(format: "%d (L)", 0x40 - pan)
                } else if pan <= 0x7F {
                    return String(format: "%d (R)", pan - 0x40)
                }
                return "<Invalid pan value>"
            }
            
        case "songScript.pitchEnvelope", "songScript.transpose", "songScript.modifyTranspose", "songScript.detune", "songScript.echoFeedback":
            if key == "pitch" || key == "transpose" || key == "detune" || key == "echoFeedback" {
                guard let intValue = super.valueForCommand(command, forKey: key) as? Int else { break }
                if intValue > 0x7F {
                    return intValue - 0x100
                } else {
                    return intValue
                }
            }
            
        case "songScript.repeat":
            if key == "count" {
                guard let intValue = super.valueForCommand(command, forKey: key) as? Int else { break }
                return intValue + 1
            }
            
        default: break
        }
        return super.valueForCommand(command, forKey: key)
    }
    
    override func setValue(_ value: Any?, forCommand command: ROMCommand, forKey key: String) {
        switch command.identifier! {
        case "songScript.note":
            guard let intValue = value as? Int else { break }
            if key == "note" {
                let opcode = command.intValue(forKey: "opcode")
                let duration = opcode % 14
                super.setValue(intValue * 14 + duration, forCommand: command, forKey: "opcode")
                return
            } else if key == "duration" {
                let opcode = command.intValue(forKey: "opcode")
                let note = opcode / 14
                super.setValue(note * 14 + intValue, forCommand: command, forKey: "opcode")
                return
            }
            
        case "songScript.pan", "songScript.panEnvelope":
            if key == "pan" {
                command.willChangeValue(forKey: "panString")
                super.setValue(value, forCommand: command, forKey: "pan")
                command.didChangeValue(forKey: "panString")
                return
            }
            
        case "songScript.pitchEnvelope", "songScript.transpose", "songScript.modifyTranspose", "songScript.detune", "songScript.echoFeedback":
            if key == "pitch" || key == "transpose" || key == "detune" || key == "echoFeedback" {
                guard let intValue = value as? Int else { break }
                if intValue < 0 {
                    super.setValue(0x100 + intValue, forCommand: command, forKey: key)
                } else {
                    super.setValue(intValue, forCommand: command, forKey: key)
                }
                return
            }
            
        case "songScript.repeat":
            if key == "count" {
                guard let intValue = value as? Int else { break }
                super.setValue(intValue - 1, forCommand: command, forKey: key)
                return
            }
            
        default: break
        }
        super.setValue(value, forCommand: command, forKey: key)
    }
    
    override func descriptionForCommand(_ command: ROMCommand) -> String {
        
        func durationString(key: String) -> String {
            let duration = command.intValue(forKey: key)
            return String(format: "[%d:%02d]", duration / 0x30, duration % 0x30)
        }
        
        func referenceName(_ key: String = "scriptPointer") -> String {
            guard let reference = command.value(forKey: key) as? ROMReference,
                let referencedCommand = reference.referencedAssembly as? ROMCommand else {
                    return "<Invalid Script Pointer>"
            }
            
            return referencedCommand.label
        }
        
        func panString(key: String = "pan") -> String {
            let pan = command.intValue(forKey: key)
            if pan == 0x40 {
                return "0 (C)"
            } else if pan <= 0x3F {
                return String(format: "%d (L)", 0x40 - pan)
            } else if pan <= 0x7F {
                return String(format: "%d (R)", pan - 0x40)
            }
            return "<Invalid pan value>"
        }
        
        func signedIntString(key: String) -> String {
            let value = command.intValue(forKey: key)
            if value == 0 {
                return "0"
            } else if value <= 0x7F {
                return String(format: "+%d", value)
            } else {
                return String(format: "-%d", 0x100 - value)
            }
        }

        switch command.identifier! {
            
        case "songScriptHeader.header":
            return "Song Header"
            
        case "songScript.note":
            let opcode = command.intValue(forKey: "opcode")
            let note = opcode / 14
            let duration = opcode % 14
            let noteNames = ["C", "C♯", "D", "E♭", "E", "F", "F♯", "G", "A♭", "A", "B♭", "B", "Tie", "Rest"]
            let durationNames = [
                "[4:00]",
                "[2:00]",
                "[1:16]",
                "[1:24]",
                "[1:00]",
                "[0:32]",
                "[0:36]",
                "[0:24]",
                "[0:16]",
                "[0:12]",
                "[0:08]",
                "[0:06]",
                "[0:04]",
                "[0:03]"
            ]
            return noteNames[note] + " " + durationNames[duration]

        case "songScript.volume":
            let volume = command.intValue(forKey: "volume")
            return "Set volume to " + String(volume)

        case "songScript.volumeEnvelope":
            let volume = command.intValue(forKey: "volume")
            return "Set volume to " + String(volume) + " over " + durationString(key: "envelope")

        case "songScript.pan":
            return "Set pan to " + panString()
            
        case "songScript.panEnvelope":
            return "Set pan to " + panString() + " over " + durationString(key: "envelope")

        case "songScript.pitchEnvelope":
            let pitch = command.intValue(forKey: "pitch")
            return "Change pitch by " + String(format: "%+d", pitch) + " over " + durationString(key: "envelope")

        case "songScript.vibrato":
            return "Enable vibrato"

        case "songScript.vibratoOff":
            return "Disable vibrato"

        case "songScript.tremolo":
            return "Enable tremolo"
            
        case "songScript.tremoloOff":
            return "Disable tremolo"

        case "songScript.pansweep":
            return "Enable pansweep"
            
        case "songScript.pansweepOff":
            return "Disable pansweep"

        case "songScript.noiseClock":
            let noiseClock = command.intValue(forKey: "noiseClock")
            return "Set noise clock to " + (command.rom.string(atIndex: noiseClock, inStringTable: "noiseClock") ?? String(noiseClock))

        case "songScript.noise":
            switch command.intValue(forKey: "opcode") {
            case 0xD0: return "Enable noise"
            case 0xD1: return "Disable noise"
            default: return "<Invalid song command>"
            }

        case "songScript.pitchMod":
            switch command.intValue(forKey: "opcode") {
            case 0xD2: return "Enable pitch modulation"
            case 0xD3: return "Disable pitch modulation"
            default: return "<Invalid song command>"
            }

        case "songScript.echo":
            switch command.intValue(forKey: "opcode") {
            case 0xD4: return "Enable echo"
            case 0xD5: return "Disable echo"
            default: return "<Invalid song command>"
            }

        case "songScript.octave":
            let octave = command.intValue(forKey: "octave")
            return "Set octave to \(octave)"

        case "songScript.incrementOctave":
            switch command.intValue(forKey: "opcode") {
            case 0xD7: return "Increment octave"
            case 0xD8: return "Decrement octave"
            default: return "<Invalid song command>"
            }

        case "songScript.transpose":
            let transpose = command.intValue(forKey: "transpose")
            return "Set transpose to " + String(format: "%+d", transpose)

        case "songScript.modifyTranspose":
            let transpose = command.intValue(forKey: "transpose")
            return "Modify transpose by " + String(format: "%+d", transpose)

        case "songScript.detune":
            let detune = command.intValue(forKey: "detune")
            return "Set detune to " + String(format: "%+d", detune)

        case "songScript.instrument":
            let instrument = command.intValue(forKey: "instrument")
            return "Set instrument to " + (command.rom.string(atIndex: instrument, inStringTable: "songInstruments") ?? String(instrument))

        case "songScript.adsrAttack":
            let value = command.intValue(forKey: "value")
            return "Set ADSR attack to " + (command.rom.string(atIndex: value, inStringTable: "adsrAttack") ?? String(value))

        case "songScript.adsrDecay":
            let value = command.intValue(forKey: "value")
            return "Set ADSR decay to " + (command.rom.string(atIndex: value, inStringTable: "adsrDecay") ?? String(value))

        case "songScript.adsrSustain":
            let value = command.intValue(forKey: "value")
            return "Set ADSR sustain to " + (command.rom.string(atIndex: value, inStringTable: "adsrSustain") ?? String(value))

        case "songScript.adsrRelease":
            let value = command.intValue(forKey: "value")
            return "Set ADSR release to " + (command.rom.string(atIndex: value, inStringTable: "adsrRelease") ?? String(value))

        case "songScript.adsrReset":
            return "Reset ADSR values to default"

        case "songScript.repeat":
            let count = command.intValue(forKey: "count")
            if count == 1 {
                return "Repeat 1 time"
            } else {
                return "Repeat \(count) times"
            }

        case "songScript.endRepeat":
            return "End Repeat"

        case "songScript.slur":
            switch command.intValue(forKey: "opcode") {
            case 0xE4: return "Enable slur"
            case 0xE5: return "Disable slur"
            default: return "<Invalid song command>"
            }

        case "songScript.drumRoll":
            switch command.intValue(forKey: "opcode") {
            case 0xE6: return "Enable drum roll"
            case 0xE7: return "Disable drum roll"
            default: return "<Invalid song command>"
            }

        case "songScript.tie":
            return "Tie " + durationString(key: "duration")

        case "songScript.soundEffect":
            let soundEffect = command.intValue(forKey: "soundEffect")
            return "Play sound effect " + (command.rom.string(atIndex: soundEffect, inStringTable: "gameSoundEffects") ?? String(soundEffect))

        case "songScript.tempo":
            let tempo = command.intValue(forKey: "tempo")
            return "Set tempo to " + String(tempo) + " bpm"

        case "songScript.tempoEnvelope":
            let tempo = command.intValue(forKey: "tempo")
            return "Set tempo to " + String(tempo) + " over " + durationString(key: "envelope")

        case "songScript.echoVolume":
            let echoVolume = command.intValue(forKey: "echoVolume")
            return "Set echo volume to " + String(echoVolume)
            
        case "songScript.echoVolumeEnvelope":
            let echoVolume = command.intValue(forKey: "echoVolume")
            return "Set echo volume to " + String(echoVolume) + " over " + durationString(key: "envelope")
            
        case "songScript.songVolume":
            let volume = command.intValue(forKey: "volume")
            return "Set song volume to " + String(volume)
            
        case "songScript.jumpEnding":
            let ending = command.intValue(forKey: "ending")
            return "Jump to " + referenceName() + " (ending \(ending))"

        case "songScript.jump":
            return "Jump to " + referenceName()

        case "songScript.echoFeedback":
            let echoFeedback = command.intValue(forKey: "echoFeedback")
            return "Set echo feedback to " + String(echoFeedback) + " over " + durationString(key: "envelope")

        case "songScript.filter":
            let filter = command.intValue(forKey: "filter")
            return "Set filter to " + String(filter) + " over " + durationString(key: "envelope")
            
        case "songScript.incrementOutputCode":
            return "Increment output code"

        case "songScript.clearOutputCode":
            return "Clear output code"

        case "songScript.ignoreSongVolume":
            return "Ignore song volume"

        case "songScript.conditionalJump":
            return "Jump to " + referenceName() + " when signaled by event script"

        case "songScript.endOfScript":
            return "End of script"

        default:
            return super.descriptionForCommand(command)
        }
    }
}
