//
//  ROMArray.swift
//  ROMTools
//
//  Created by Evan Owen on 3/4/15.
//
//

import Foundation

class ROMArray: ROMAssembly {
    
    // MARK: instance properties

    @objc var array = NSMutableArray()
    var count: Int { return array.count }
    let prototype: ROMAssembly
    let entrySize: Int
    var initialCount: Int
    let isShared: Bool
    var linkedArrays: NSArray?
    var linkedStringTable: String?
    var stringTable: NSMutableDictionary?
    override var description: String { return array.description }
    override var assembledLength: Int {
        
        // arrays with fixed length entries are easy
        if entrySize != 0 { return entrySize * array.count }
        
        // assemble variable length arrays and return the length (slow but it works)
        var assembledData = Data()
        assembleTo(&assembledData)
        return assembledData.count
    }

    // MARK: instance methods
    
    override init(_ rom: ROM, definition: NSDictionary? = nil) {
        entrySize = Int(object: definition, keyPath: "entrySize") ?? 0
        initialCount = Int(object: definition, keyPath: "initialCount") ?? 0
        isShared = definition?["isShared"] as? Bool ?? true
        linkedArrays = definition?["linkedArrays"] as? NSArray
        linkedStringTable = definition?["linkedStringTable"] as? String
        stringTable = definition?["stringTable"] as? NSMutableDictionary

        if let prototypeDefinition = definition?["prototype"] as? NSDictionary,
            let assembly = ROMObject.createObject(rom, definition: prototypeDefinition) as? ROMAssembly {
            prototype = assembly
        } else {
            prototype = ROMData(rom)
        }

        super.init(rom, definition: definition)

        if prototype.identifier == nil { prototype.identifier = self.identifier }
        if prototype.nib == nil { prototype.nib = self.nib }
    }
    
    override func createDefinition() -> NSMutableDictionary {
        let definition = super.createDefinition()
        
        // type
        definition["type"] = "ROMArray"

        // entry size
        if entrySize != 0 {
            definition["entrySize"] = entrySize
        }

        // initial count
        if initialCount != 0 {
            definition["initialCount"] = initialCount
        }

        // prototype
        definition["prototype"] = prototype.createDefinition()

        // isShared
        if isShared == false {
            definition["isShared"] = false
        }
        
        // linked arrays
        definition["linkedArrays"] = linkedArrays
        
        // string table
        definition["linkedStringTable"] = linkedStringTable
        if let identifier = self.identifier, let stringTable = rom.subAssemblies["stringTable-" + identifier] as? ROMStringTable {
            // update string table
            definition["stringTable"] = stringTable.createDefinition()
        } else {
            // use existing string table
            definition["stringTable"] = self.stringTable
        }
        
        return definition
    }
    
    override func updateReferences() {
        super.updateReferences()
        
        for object in array {
            guard let subAssembly = object as? ROMAssembly else { continue }
            subAssembly.updateReferences()
        }
    }
    
    func objectAtIndex(_ index: Int) -> AnyObject {
        if (0..<count).contains(index) {
            return array.object(at: index) as AnyObject
        }
        rom.log("Index \(index) beyond bounds [0...\(count - 1)]")
        return 0 as AnyObject
    }
    
    func replaceObjectAtIndex(_ index: Int, withObject object: AnyObject) {
        array.replaceObject(at: index, with: object)
        
    }
    
    func createNewAssembly() -> ROMAssembly? {
        guard let newAssembly = prototype.copy() as? ROMAssembly else { return nil }
        if let text = newAssembly as? ROMText {
            text.string = ""
        } else {
            newAssembly.disassembleFrom(Data(count: entrySize))
        }
        newAssembly.parentArray = self
        return newAssembly
    }
    
//    func appendNewAssembly() -> ROMAssembly? {
//
//        // append new entries to linked arrays first
//        if linkedArrays != nil {
//            for object in linkedArrays! {
//                guard let arrayIdentifier = object as? String else { continue }
//                let linkedArray = rom.valueForKey(arrayIdentifier)
//                if let romArray = linkedArray as? ROMArray where romArray.count == self.count {
//                    romArray.appendNewAssembly()
//                }
//            }
//        }
//        
//        stringTable?.createNewStringAtIndex(self.count)
//
//        // append the new assembly to the array
//        guard let newAssembly = createNewAssembly() else { return nil }
//        array.addObject(newAssembly)
//        
//        return newAssembly
//    }
    
    func insertAssembly(_ assembly: ROMAssembly, atIndex index: Int) {
        if index >= array.count {
            array.add(assembly)
        } else {
            array.insert(assembly, at: index)
        }
        assembly.parentArray = self
    }
    
    func removeAssemblyAtIndex(_ index: Int) {
        if index >= array.count { return }
        if let assembly = array[index] as? ROMAssembly {
            assembly.parentArray = nil
        }
        array.removeObject(at: index)
    }
    
    subscript(index: Int) -> AnyObject {
        get {
            return objectAtIndex(index)
        } set {
            replaceObjectAtIndex(index, withObject: newValue)
        }
    }
    
    // MARK: ROMAssembly methods

    override func assembleTo(_ data: inout Data) {
        
        // update the initial count
        initialCount = count
        
        // start with a blank NSMutableData
        data = Data()
        
        if (pointerTable != nil) {
            // array has a pointer table
            var currentPointer = range!.lowerBound

            // array for keeping track of duplicates
            var duplicates = [Data : Int]()

            // make a new empty pointer table
            pointerTable!.pointers = [Int](repeating: 0, count: count)
            
            // update the pointer offset if there are valid references
            if pointerTable?.pointerOffset != nil && references != nil && range != nil {
                pointerTable!.pointerOffset = range!.lowerBound - (range!.lowerBound % pointerAlign)
            }
            
            for (index, entry) in array.enumerated() {
                // start with an empty NSMutableData with the same length as each entry (which will be zero if entries are variable-size)
                var assemblyData = Data(count: entrySize)
                
                // assemble the data
                guard let assembly = entry as? ROMAssembly else { continue }
                assembly.assembleTo(&assemblyData)
                
                // check if it's a duplicate of something that was already assembled
                if let duplicatePointer = duplicates[assemblyData] {
                    // assembled data is a duplicate
                    pointerTable!.pointers[index] = duplicatePointer
                } else {
                    // assembled data is not a duplicate
                    pointerTable!.pointers[index] = currentPointer
                    if isShared {
                        // only add to duplicates if the array entries can be shared
                        duplicates[assemblyData] = currentPointer
                    }
                    currentPointer += assemblyData.count
                    data.append(assemblyData)
                }
            }
            
            if !isShared {
                // add a pointer to the end of the data if it is not shared
                pointerTable!.pointers.append(currentPointer)
            }
            
        } else {
            // array has no pointer table
            for entry in array {
                var assemblyData = Data(count: entrySize)
                guard let assembly = entry as? ROMAssembly else { continue }
                assembly.assembleTo(&assemblyData)
                data.append(assemblyData)
            }
        }
    }
    
    override func disassembleFrom(_ data: Data) {
        
        // clear the array
        array.removeAllObjects()
        
        var arrayData: Data
        var entryRanges = [Range<Int>]()
        if pointerTable != nil {
            // array has a pointer table
            pointerTable!.disassembleFrom(data)
            
            // make sure that there are enough pointers
            if pointerTable!.pointers.count < initialCount {
                initialCount = pointerTable!.pointers.count
                rom.log("Not enough pointers available - disassembling \(initialCount) entries")
            }
            
            if range == nil {
                // if range is not specified, include the entire data
                range = 0..<data.count
            }
            
            if isShared {
                // shared data, sort the pointers in reverse order
                let entryPointers = pointerTable!.pointers[0..<initialCount]
                let sortedPointers = Array(entryPointers.sorted{ $0 > $1 })
                
                // create a range from each pointer to the next higher pointer
                var pointerRanges = [Int : Range<Int>]()
                var endPointer = range!.upperBound // start at the end of the range
                
                // figure out where each entry starts and ends
                for startPointer in sortedPointers {

                    // skip pointers that have already been added
                    if pointerRanges[startPointer] != nil { continue }

                    // start out with an empty range (just so there's something there if the range ends up being invalid)
                    pointerRanges[startPointer] = 0..<1

                    // skip pointers that are completely outside the assembly's range
                    if (startPointer > range!.upperBound) || (endPointer < range!.lowerBound) { continue }
                    
                    // crop the pointer range to fit inside the assembly's range
                    let start = max(startPointer, range!.lowerBound)
                    let end = min(endPointer, range!.upperBound)
                    
                    // add the new pointer range
                    pointerRanges[startPointer] = start..<end
                    
                    // the end of the next range will be the start of this range
                    endPointer = startPointer
                }
                
                // populate the array of ranges for each entry
                for pointer in entryPointers {
                    entryRanges.append(pointerRanges[pointer]!)
                }
            } else {
                // not shared, each entry goes up to the next consecutive pointer
                for index in 0..<(initialCount - 1) {
                    let start = max(pointerTable!.pointers[index], range!.lowerBound)
                    let end = min(pointerTable!.pointers[index + 1], range!.upperBound)
                    if end >= start {
                        entryRanges.append(start..<end)
                    } else {
                        // if the pointers aren't in order, go up to the next largest pointer
                        let nextEntryStart = pointerTable!.pointers.reduce(range!.upperBound, { ($1 > start) && ($1 < $0) ? $1 : $0 })
                        entryRanges.append(start..<nextEntryStart)
                    }
                }
                
                // the last range goes up to the end of the array's range
                let lastEntryStart = pointerTable!.pointers[initialCount - 1]
                entryRanges.append(lastEntryStart..<range!.upperBound)

            }
            arrayData = data

        } else if range != nil {
            // no pointer table
            // if the range property is valid, copy only the range of data specified
            arrayData = data.subdata(in: range!)
        } else {
            // otherwise, copy all of the data
            arrayData = data
        }
        
        // determine how many entries there are
        if (initialCount == 0) && (entrySize != 0) {
            initialCount = arrayData.count / entrySize
            if arrayData.count % entrySize != 0 {
                rom.log("Data size is not evenly divisible by array entry size")
            }
        }
        
        if pointerTable == nil {
            for index in 0..<initialCount {
                let start = index * entrySize
                let end = start + entrySize
                entryRanges.append(start..<end)
            }
        }

        for i in 0..<initialCount {
            // disassemble the new item
            let entryData = arrayData.subdata(in: entryRanges[i])

            // append a copy of the prototype assembly
            let entry = prototype.copy() as! ROMAssembly
            entry.parentArray = self
            entry.disassembleFrom(entryData)
            entry.range = entryRanges[i]
            array.add(entry)
        }
    }
}
