//
//  ROMAssembly.swift
//  ROMTools
//
//  Created by Evan Owen on 6/13/15.
//
//

import Foundation

class ROMAssembly: ROMObject {
    
    var range: Range<Int>?
    var align: Int
    var pointerAlign: Int
    var pointerTable: ROMPointerTable?
    var references: [String : ROMReference]?
    var parentArray: ROMArray?
    var index: Int { return parentArray?.array.index(of: self) ?? NSNotFound }
    var assembledLength: Int { return 0 }
    var previousAssembly: String?
    var nextAssembly: String?
    
    override init(_ rom: ROM, definition: NSDictionary? = nil) {

        // align
        align = Int(object: definition, keyPath: "align") ?? 1
        if align <= 0 { align = 1 }

        // pointerAlign
        pointerAlign = Int(object: definition, keyPath: "pointerAlign") ?? 1
        if pointerAlign <= 0 { pointerAlign = 1 }

        // previous and next assemblies
        previousAssembly = definition?["previousAssembly"] as? String
        nextAssembly = definition?["nextAssembly"] as? String

        super.init(rom, definition: definition)
        
        // range
        if let range = MakeRange(object: definition, keyPath: "range") {
            self.range = rom.mapMode.mapRange(range)
        } else if let start = Int(object: definition, keyPath: "start"),
            let end = Int(object: definition, keyPath: "start") {
            let range: Range<Int> = start..<end
            self.range = rom.mapMode.mapRange(range)
        }

        // pointer table
        if let pointerTableDefinition = definition?["pointerTable"] as? NSMutableDictionary {
            pointerTableDefinition["identifier"] = "pointerTable-" + (identifier ?? "unnamed")
            pointerTableDefinition["name"] = "Pointers to " + (name ?? "Unnamed Assembly")
            pointerTable = ROMAssembly.createObject(rom, definition: pointerTableDefinition) as? ROMPointerTable
        }

        // references
        if let referenceDefinitions = definition?["references"] as? NSMutableDictionary {
            references = [String : ROMReference]()
            for (key, value) in referenceDefinitions {
                guard let locationString = key as? String else { continue }
                let reference = ROMReference(rom, definition: value as? NSDictionary)
                if let location = Int(object: locationString as AnyObject?) {
                    reference.location = location
                }
                if reference.referencedAssembly == nil {
                    reference.referencedAssembly = self
                }
                references?[locationString] = reference
            }
        }
    }
    
    // TODO: make this a variable instead of a function
    override func createDefinition() -> NSMutableDictionary {
        
        let definition = super.createDefinition()
        
        // type
        definition["type"] = "ROMAssembly"

        // range
        if range != nil {
            let start = rom.mapMode.unmapAddress(range!.lowerBound)
            let end = start + range!.count
            definition["range"] = String(format: "0x%06X-0x%06X", start, end)
        }

        // range
        if range != nil {
            let start = rom.mapMode.unmapAddress(range!.lowerBound)
            let end = start + range!.count
            definition["range"] = String(format: "0x%06X-0x%06X", start, end)
        }

        // alignment
        if align != 1 { definition["align"] = String(format: "0x%X", align) }
        if pointerAlign != 1 { definition["pointerAlign"] = String(format: "0x%X", pointerAlign) }
        
        // previous and next assemblies
        definition["previousAssembly"] = previousAssembly
        definition["nextAssembly"] = nextAssembly
        
        // pointer table
        if pointerTable != nil {
            definition["pointerTable"] = pointerTable!.createDefinition()
        }
        
        // references
        if references != nil {
            let referenceDefinitions = NSMutableDictionary()
            for (_, reference) in references! {
                let referenceDefinition = reference.createDefinition()
                referenceDefinition["referencedAssembly"] = nil
                referenceDefinition["linkedObject"] = nil
                if let location = reference.location {
                    referenceDefinitions[String(format: "0x%06X", location)] = referenceDefinition
                } else if let linkedKeyPath = reference.linkedKeyPath {
                    referenceDefinitions[linkedKeyPath] = referenceDefinition
                }
            }
            definition["references"] = referenceDefinitions
        }
        
        return definition
    }
    
    func assembleTo(_ data: inout Data) { }
    func disassembleFrom(_ data: Data) { }
    
    func relocate(to location: Int) {
        
        // set the new range for this assembly
        range = location..<(location + assembledLength)
        
        if pointerTable != nil {
            // if the assembly has a pointer table, it needs to be re-assembled to update the pointers
            var dummyData = Data()
            assembleTo(&dummyData)
            if pointerTable!.pointerOffset != nil {
                pointerTable!.pointerOffset = location - (location % pointerAlign)
            }
        }
        
        // update references
        updateReferences()
    }
    
    func updateReferences() {
        
        // return if there are no references
        if references == nil { return }
        
        for (_, reference) in references! {
            if var linkedObject = reference.linkedObject as? Data {
                reference.assembleTo(&linkedObject)
                reference.linkedObject = linkedObject as AnyObject?
            } else {
                var linkedObject: Data = rom.data
                reference.assembleTo(&linkedObject)
                rom.data = linkedObject
            }
        }
    }

    override func copy(with zone: NSZone?) -> Any {
        let copy = super.copy(with: zone)
        if let assembly = copy as? ROMAssembly {
            assembly.parentArray = self.parentArray
            return assembly
        }
        return copy
    }
}
