//
//  ROMCommand.swift
//  FF6Tools
//
//  Created by Evan Owen on 8/21/16.
//  Copyright © 2016 Evan Owen. All rights reserved.
//

import Foundation

class ROMCommand: ROMAssembly {
    
    var data = Data()
    var label: String {
        get {
            if let customLabel = script.commandLabels[reference] {
                return customLabel
            } else {
                let absoluteOffset = rom.mapMode.unmapAddress(range!.lowerBound)
                return String(format: "%02X/%04X", absoluteOffset >> 16, absoluteOffset & 0xFFFF)
            }
        }
        set {
            if newValue == "" {
                script.commandLabels[reference] = nil
            } else {
                script.commandLabels[reference] = newValue
            }
        }
    }
    let encodingName: String
    let encoding: ROMScriptEncoding
    var reference: Int
    override var index: Int { return siblings.index(of: self) }
    
    weak var script: ROMScript!
    weak var parent: ROMCommand?
    var childCommands = NSMutableArray()
    var siblings: NSMutableArray { return parent?.childCommands ?? script.childCommands }
    
    override var description: String { return encoding.descriptionForCommand(self) }
    var numLinesForDescription: Int { return encoding.numLinesForCommandDescription(self) }
    var encodingNameForNextCommand: String { return encoding.encodingNameForCommandAfter(self) }
    var childEncodingName: String { return encoding.encodingNameForChildrenOfCommand(self) }
    
    override var assembledLength: Int {
        var lengthIncludingChildren = data.count
        for child in childCommands {
            if let childCommand = child as? ROMCommand {
                lengthIncludingChildren += childCommand.assembledLength
            }
        }
        return lengthIncludingChildren
    }
    
    var previousCommand: ROMCommand? {
        let index = self.index
        if (index == NSNotFound) || (index == 0) { return nil }
        return siblings[index - 1] as? ROMCommand
    }
    
    var nextCommand: ROMCommand? {
        let index = self.index
        if (index == NSNotFound) || (index + 1) >= siblings.count { return nil }
        return siblings[index + 1] as? ROMCommand
    }
    
    // TODO: make this work better (it's slow)
    var longDescription: String {
        
        // start with the command's offset
        let absoluteOffset = rom.mapMode.unmapAddress(range!.lowerBound)
        var longDescription = String(format: "%02X/%04X: ", absoluteOffset >> 16, absoluteOffset & 0xFFFF)
        
        // add the opcode
        let opcode = intValue(forKey: "opcode")
        if opcode <= 0xFF {
            // one-byte opcode
            longDescription += String(format: "%02X  ", opcode)
        } else {
            // two-byte opcode
            longDescription += String(format: "%02X  ", opcode >> 8)
        }
        
        // indent based on how many parent commands there are
        var indentationLevel = longDescription.characters.count
        var currentParent = self.parent
        while currentParent != nil {
            currentParent = currentParent?.parent
            indentationLevel += 4
        }
        
        for (lineIndex, lineString) in self.description.components(separatedBy: "\n").enumerated() {
            if lineIndex == 0 {
                longDescription = longDescription.padding(toLength: indentationLevel, withPad: " ", startingAt: 0) + lineString + "\n"
            } else {
                longDescription += "".padding(toLength: indentationLevel, withPad: " ", startingAt: 0) + lineString + "\n"
            }
        }
        
        if let customLabel = script.commandLabels[reference] {
            let dashLength = (100 - customLabel.characters.count) / 2
            let labelLine = ("".padding(toLength: dashLength, withPad: "-", startingAt: 0) + "[ " + customLabel + " ]").padding(toLength: 100, withPad: "-", startingAt: 0)
            return "\n" + labelLine + "\n\n" + longDescription
        }
        return longDescription
    }
    
    static var nextUnusedReference = -1
    
    // TODO: init with parent: AnyObject? instead of script so that the command knows its parent before being disassembled (this messes up eventScript.jumpDialog commands)
    required init?(_ rom: ROM, script: ROMScript, encodingName: String, data: Data, offset: Int?) {
        
        self.script = script
        self.encodingName = encodingName
        
        // 0.24
        
        guard let encoding = rom.scriptEncodings[encodingName] as? ROMScriptEncoding else {
            rom.log("Invalid encoding: " + encodingName)
            return nil
        }
        self.encoding = encoding
        
        // 0.11
        
        // set the reference to a unique value
        if offset == nil {
            reference = ROMCommand.nextUnusedReference
            ROMCommand.nextUnusedReference -= 1
        } else {
            reference = offset!
        }
        
        // 0.16
        
        // get the opcode and command definition
        var opcode = 0
        let offset = offset ?? 0
        var opcodeRange: Range<Int> = offset..<(offset + 1)
        let _opcodeRange = NSMakeRange(offset, 1)
        var commandDefinition: NSDictionary?
        if opcodeRange.upperBound > data.count {
            rom.log("Invalid command offset: " + String(offset))
            return nil
        }
        (data as NSData).getBytes(&opcode, range: _opcodeRange)
        commandDefinition = encoding[opcode as NSCopying] as? NSDictionary

        // 0.10
        
        // check if a two-byte command is valid
        if commandDefinition == nil && opcodeRange.upperBound + 1 <= data.count {
            let twoByteRange: Range<Int> = offset..<(offset + 2) // NSMakeRange(offset ?? 0, 2)
            let _twoByteRange = NSMakeRange(offset, 2)
            var twoByteOpcode: UInt16 = 0
            (data as NSData).getBytes(&twoByteOpcode, range: _twoByteRange)
            twoByteOpcode = twoByteOpcode.byteSwapped
            if let twoByteDefinition = encoding[Int(twoByteOpcode) as NSCopying] as? NSDictionary {
                commandDefinition = twoByteDefinition
                opcodeRange = twoByteRange
                opcode = Int(twoByteOpcode)
            }
        }
        
        // 0.13
        
        super.init(rom, definition: commandDefinition)
        self.range = opcodeRange
        
        // 0.28
        
        // set the command identifier
        identifier = commandDefinition?["identifier"] as? String ?? String(format: "command_%02X", opcode)
        
        // 0.15
        
        identifier = encodingName + "." + identifier!
        
        // 0.00
        
        // get the default command length
        if let length = Int(object: commandDefinition?["length"] as AnyObject?) {
            range = offset..<(offset + length)
        }
        
        // 0.07
        
        // validate the range
        if range!.upperBound > data.count {
            rom.log("Invalid command range")
            return nil
        }
        
        // 3.83
        
        // disassemble the command
        encoding.willDisassembleCommand(self, fromData: data) // 0.06
        disassembleFrom(data) // 0.12
        encoding.didDisassembleCommand(self, fromData: data) // 3.75 (about half of this is parent commands creating child commands)
        
        // 0.12
        
        // look for any placeholders that reference this command
        if let placeholderReferences = script.commandReferences[reference] as? NSArray {
            
            // set this command as the referenced assembly for each placeholder
            for object in placeholderReferences {
                guard let reference = object as? ROMReference else { continue }
                reference.referencedAssembly = self
            }
        }
        
        // 0.12
        
        // create a reference to this command for the script (this will remove any placeholders)
        script.commandReferences[reference] = self
    }
    
    // use this init function to create new commands with a given opcode
    convenience init?(_ rom: ROM, script: ROMScript, encodingName: String, opcode: Int) {
        
        // get the appropriate encoding table
        guard let encoding = rom.scriptEncodings[encodingName] as? ROMScriptEncoding else {
            rom.log("Invalid script encoding: " + encodingName)
            return nil
        }
        
        // get the default command length
        let length = Int(object: encoding[opcode as NSCopying], keyPath: "length") ?? 1
        
        // create an empty data of the correct length and set the command code
        var dataBytes = [UInt8](repeating: 0, count: length)
        if opcode <= 0xFF {
            // single byte opcode
            dataBytes[0] = UInt8(opcode)
        } else {
            // two-byte opcode
            dataBytes[0] = UInt8(truncatingIfNeeded: opcode >> 8)
            dataBytes[1] = UInt8(truncatingIfNeeded: opcode)
        }
        let dummyData = Data(bytes: UnsafePointer<UInt8>(dataBytes), count: length)
        
        self.init(rom, script: script, encodingName: encodingName, data: dummyData, offset: nil)
    }
    
    // use this init function to create new commands with a given identifier
    convenience init?(_ rom: ROM, script: ROMScript, identifier: String) {
        
        // validate the encoding
        let identifierComponents = identifier.components(separatedBy: ".")
        if identifierComponents.count < 2 {
            rom.log("Invalid command identifier: " + identifier)
            return nil
        }
        
        // get the appropriate encoding table
        let encodingName = identifierComponents[0]
        guard let encoding = rom.scriptEncodings[encodingName] as? ROMScriptEncoding else {
            rom.log("Invalid script encoding: " + encodingName)
            return nil
        }
        
        // get the first opcode definition
        let shortIdentifier = identifierComponents[1]
        guard let definition = encoding[shortIdentifier as NSCopying] as? NSDictionary,
            let opcodes = definition["opcodes"] as? NSArray,
            let firstOpcode = opcodes.firstObject else {
                rom.log("Command has no valid opcodes: " + identifier)
                return nil
        }
        
        // parse the first opcode
        if let opcodeRange = MakeRange(object: firstOpcode as AnyObject?) {
            self.init(rom, script: script, encodingName: encodingName, opcode: opcodeRange.lowerBound)
        } else if let opcode = Int(object: firstOpcode as AnyObject?) {
            self.init(rom, script: script, encodingName: encodingName, opcode: opcode)
        } else {
            rom.log("Invalid command opcode: " + String(describing: firstOpcode))
            return nil
        }
    }
    
    override func createDefinition() -> NSMutableDictionary {
        // this is only for creating dummy ROMAssemblys for copying this command's reference
        let definition = super.createDefinition()
        let start = reference
        let end = start + range!.count
        definition["range"] = String(format: "0x%06X-0x%06X", start, end)
        return definition
    }
    
    override func assembleTo(_ data: inout Data) {
        
        // append the command's data
        data.append(self.data)
        
        // assemble any child commands
        for child in childCommands {
            guard let childCommand = child as? ROMCommand else { continue }
            childCommand.encoding.willAssembleCommand(childCommand)
            childCommand.assembleTo(&data)
            childCommand.encoding.didAssembleCommand(childCommand)
        }
    }
    
    override func disassembleFrom(_ data: Data) {
        // copy data in the command's range
        if range!.upperBound >= data.count {
            range = range!.lowerBound..<data.count
        }
        self.data = data.subdata(in: range!)
    }
    
    override func updateReferences() {
        
        super.updateReferences()
        
        // update references for all child commands
        for object in childCommands {
            guard let childCommand = object as? ROMCommand else { continue }
            childCommand.updateReferences()
        }
        
        // do any encoding-specific updates
        encoding.updateReferencesForCommand(self)
    }
    
    func intValue(forKey key: String) -> Int {
        return value(forKey: key) as? Int ?? 0
    }
    
    func boolValue(forKey key: String) -> Bool {
        return intValue(forKey: key) != 0
    }
    
    override func value(forKey key: String) -> Any? {
        
        if let encodingValue = encoding.valueForCommand(self, forKey: key) {
            return encodingValue
        }
        return super.value(forKey: key)
    }
    
    override func value(forUndefinedKey key: String) -> Any? {
        return nil
    }
    
    override func setValue(_ value: Any?, forKey key: String) {
        encoding.setValue(value, forCommand: self, forKey: key)
    }
    
}
