//
//  ROMPointerTable.swift
//  FF6Tools
//
//  Created by Evan Owen on 8/21/16.
//  Copyright © 2016 Evan Owen. All rights reserved.
//

import Foundation

class ROMPointerTable: ROMAssembly {
    
    // NOTE: the pointer offset and pointers are absolute from the beginning of the ROM (mapped)
    
    // MARK: instance properties
    
    var pointerOffset: Int?
    var pointerSize: Int
    var pointers = [Int]()
    override var assembledLength: Int { return pointers.count * pointerSize } // this will be incorrect if the associated array has changed size
    
    override init(_ rom: ROM, definition: NSDictionary? = nil) {
        
        pointerOffset = Int(object: definition, keyPath: "pointerOffset")
        pointerSize = Int(object: definition, keyPath: "pointerSize") ?? 2
        if pointerOffset != nil {
            // map the pointer offset if it is valid
            pointerOffset = rom.mapMode.mapAddress(pointerOffset!)
        }
        
        super.init(rom, definition: definition)
    }
    
    override func createDefinition() -> NSMutableDictionary {
        let definition = super.createDefinition()
        
        // type
        definition["type"] = "ROMPointerTable"
        
        // pointer offset
        if pointerOffset != nil {
            definition["pointerOffset"] = String(format: "0x%06X", rom.mapMode.unmapAddress(pointerOffset!))
        }
        
        // pointer size
        if pointerSize != 2 {
            definition["pointerSize"] = pointerSize
        }
        
        return definition
    }
    
    // MARK: ROMAssembly methods
    
    override func assembleTo(_ data: inout Data) {
        
        // start with an empty NSMutableData
        let newData = NSMutableData()
        
        // append each pointer, using the specified size
        for pointer in pointers {
            var valueToWrite = pointer
            if pointerOffset != nil {
                // relative pointer
                valueToWrite -= pointerOffset!
            } else {
                // absolute pointer
                valueToWrite = rom.mapMode.unmapAddress(pointer)
            }
            newData.append(&valueToWrite, length: pointerSize)
        }
        data = newData as Data
    }
    
    override func disassembleFrom(_ data: Data) {
        pointers.removeAll()
        var pointerTableData: Data
        if let dataRange = range {
            // if the range property is valid, copy only the range of data specified
            pointerTableData = data.subdata(in: dataRange)
        } else {
            // otherwise, copy all of the data
            rom.log("Pointer table with no range, this should not be...")
            pointerTableData = data
        }
        
        let count = pointerTableData.count / pointerSize
        for i in 0..<count {
            // read each pointer
            let pointerRange = NSMakeRange(i * pointerSize, pointerSize)
            var pointer = 0
            (pointerTableData as NSData).getBytes(&pointer, range: pointerRange)
            if pointerOffset != nil {
                // relative pointer
                pointer += pointerOffset!
            } else {
                // absolute pointer
                pointer = rom.mapMode.mapAddress(pointer)
            }
            pointers.append(pointer)
        }
        
    }
}
