//
//  ROMReference.swift
//  FF6Tools
//
//  Created by Evan Owen on 8/21/16.
//  Copyright © 2016 Evan Owen. All rights reserved.
//

import Foundation

class ROMReference: ROMAssembly {
    
    var referencedAssembly: ROMAssembly? // the assembly to get a location from
    var offset: Int // a fixed amount to add to the reference location
    var shift: Int // the number of bytes to shift (right) the value before writing
    var isCount: Bool // copy the number of items in an array instead of the location
    var isMapped: Bool // determines whether the location value should get mapped based on the ROM's map mode
    
    // object reference
    var linkedObject: AnyObject? // the object to write the reference location to
    var linkedKeyPath: String? // the keyPath of the property to write the reference location to
    var encodingName: String? // for placeholder ROMCommand references only, tells which encoding to use
    
    // direct reference
    var location: Int? // the location in the ROM to write
    var size: Int? // the length of the range in which to writereference location value
    
    override var description: String {
        if let referenceLocation = referencedAssembly?.range?.lowerBound {
            return String(referenceLocation)
        }
        return super.description
    }
    
    override init(_ rom: ROM, definition: NSDictionary?) {
        
        referencedAssembly = definition?["referencedAssembly"] as? ROMAssembly
        offset = Int(object: definition, keyPath: "offset") ?? 0
        shift = Int(object: definition, keyPath: "shift") ?? 0
        isCount = definition?["isCount"] as? Bool ?? false
        isMapped = definition?["isMapped"] as? Bool ?? false
        
        linkedObject = definition?["linkedObject"] as AnyObject?
        linkedKeyPath = definition?["linkedKeyPath"] as? String
        encodingName = definition?["encodingName"] as? String
        
        location = Int(object: definition, keyPath: "location")
        size = Int(object: definition, keyPath: "size")
        
        super.init(rom, definition: definition)
    }
    
    override func createDefinition() -> NSMutableDictionary {
        let definition = super.createDefinition()
        definition["type"] = "ROMReference"
        
        definition["referencedAssembly"] = referencedAssembly
        if offset != 0 { definition["offset"] = offset }
        if shift != 0 { definition["shift"] = shift }
        if isCount { definition["isCount"] = true }
        if isMapped { definition["isMapped"] = true }
        
//        definition["linkedObject"] = linkedObject
        definition["linkedKeyPath"] = linkedKeyPath
        definition["encodingName"] = encodingName
        
        if location != nil { definition["location"] = String(format: "0x%06X", location!) }
        definition["size"] = size
        
        return definition
    }
    
    override func assembleTo(_ data: inout Data) {
        
        guard var referenceValue = referencedAssembly?.range?.lowerBound else { return }
        
        if isMapped == false {
            referenceValue = rom.mapMode.unmapAddress(referenceValue)
        }
        
        if let referencedCommand = referencedAssembly as? ROMCommand {
            // offset values for script commands
            referenceValue -= referencedCommand.script.pointerOffset
        } else if let romArray = referencedAssembly as? ROMArray, isCount == true {
            // for references to an array's count
            referenceValue = romArray.count
        }

        // apply the offset and shift
        referenceValue = (referenceValue + offset) >> (shift * 8)

        // trim off any unused bytes
        if size != nil {
            var referenceData = Data(bytes: &referenceValue, count: size!)
            referenceData.count = MemoryLayout.size(ofValue: referenceValue)
            referenceValue = referenceData.withUnsafeBytes{ $0.pointee }
        }
        
        if let object = linkedObject as? ROMAssembly, let keyPath = linkedKeyPath {
            // assembly reference: write value to linkedObject / linkedIdentifier
            object.setValue(referenceValue, forKeyPath: keyPath)

        } else if let keyPath = linkedKeyPath, let romData = rom.value(forKeyPath: keyPath) as? ROMData, location != nil, size != nil {
            // assembly reference: write value to ROMData
            let referenceRange = NSMakeRange(location!, size!)
            if NSMaxRange(referenceRange) >= romData.data.count {
                rom.log("Invalid range for reference: \(referenceRange)")
                return
            }
            let oldData = NSMutableData(data: romData.data)
            oldData.replaceBytes(in: referenceRange, withBytes: &referenceValue)
            romData.data = oldData as Data
        
        } else if linkedKeyPath != nil {
            // write directly to a key path
            rom.setValue(referenceValue, forKeyPath: linkedKeyPath!)
            
        } else if location != nil && size != nil {
            // direct reference: write value directly to the data
            let referenceRange = NSMakeRange(rom.mapMode.mapAddress(location!), size!)
            if NSMaxRange(referenceRange) >= data.count {
                rom.log("Invalid range for reference: \(referenceRange)")
                return
            }
            let oldData = NSMutableData(data: data)
            oldData.replaceBytes(in: referenceRange, withBytes: &referenceValue)
            data = oldData as Data
        }
    }
}
