//
//  ROMScript.swift
//  ROMTools
//
//  Created by Evan Owen on 4/10/15.
//
//

import Foundation

class ROMScript: ROMAssembly {
    var childCommands = NSMutableArray()
    var encodingName = ""
    var encoding: ROMScriptEncoding?
    var commandReferences = [Int : Any]()
    var commandLabels = [Int : String]()
    override var assembledLength: Int {
        var length = 0
        for object in childCommands {
            guard let command = object as? ROMCommand else { continue }
            length += command.assembledLength
        }
        return length
    }
    var pointerOffset: Int!

    override init(_ rom: ROM, definition: NSDictionary? = nil) {

        if let encodingName = definition?["encoding"] as? String {
            self.encoding = rom.scriptEncodings[encodingName] as? ROMScriptEncoding
            self.encodingName = encodingName
        }
        
        if let labels = definition?["labels"] as? [String : String] {
            for (key, label) in labels {
                guard let reference = Int(expression: key) else { continue }
                commandLabels[rom.mapMode.mapAddress(reference)] = label
            }
        }
        
        super.init(rom, definition: definition)
        
        if let customOffset = Int(object: definition, keyPath: "pointerOffset") {
            if customOffset == 0 {
                // pointer offset is zero for absolute addressing
                pointerOffset = 0
            } else {
                pointerOffset = rom.mapMode.mapAddress(customOffset)
            }
        } else {
            // default pointer offset is the start location of the script
            pointerOffset = range?.lowerBound ?? 0
        }
    }
    
    override func createDefinition() -> NSMutableDictionary {
        let definition = super.createDefinition()
        
        // type
        definition["type"] = "ROMScript"

        // encoding
        definition["encoding"] = encodingName
        
        // labels
        var labels = [String : String]()
        for (reference, label) in commandLabels {
            guard let referencedCommand = commandReferences[reference] as? ROMCommand,
                let offset = referencedCommand.range?.lowerBound else { continue }
            let offsetString = String(format: "0x%06X", rom.mapMode.unmapAddress(offset))
            labels[offsetString] = label
        }
        definition["labels"] = labels
        
        // pointer offset
        if pointerOffset == 0 {
            definition["pointerOffset"] = 0
        } else if pointerOffset != range?.lowerBound {
            definition["pointerOffset"] = String(format: "0x%06X", rom.mapMode.unmapAddress(pointerOffset))
        }
        
        return definition
    }
    
    override func assembleTo(_ data: inout Data) {
        
        encoding?.willAssembleScript(self)
        
        // start with an empty NSMutableData
        data = Data()
        
        // assemble each child command
        for child in childCommands {
            guard let childCommand = child as? ROMCommand else { continue }
            childCommand.encoding.willAssembleCommand(childCommand)
            childCommand.assembleTo(&data)
            childCommand.encoding.didAssembleCommand(childCommand)
        }
    }
    
    override func disassembleFrom(_ data: Data) {
        
        encoding?.willDisassembleScript(self)
        childCommands.removeAllObjects()
        
        if range == nil {
            range = 0..<data.count
        }
        
        // keep track of the current offset and encoding
        var currentOffset = range!.lowerBound
        var currentEncodingName: String? = self.encodingName
        while currentOffset < range!.upperBound {

            if let referencesForThisOffset = commandReferences[currentOffset] as? NSArray {
                
                // look through the references for a placeholder with an encoding
                for object in referencesForThisOffset {
                    guard let reference = object as? ROMReference,
                        let placeHolderEncodingName = reference.encodingName else { continue }
                    currentEncodingName = placeHolderEncodingName
                    
                    // stop as soon as the first placeholder encoding is found
                    break
                }
            }

            if currentEncodingName == nil {
                rom.log("Invalid encoding")
                return
            }
            
            // create the new command
            guard let newCommand = ROMCommand(rom, script: self, encodingName: currentEncodingName!, data: data, offset: currentOffset) else {
                // there was an error disassembling the command
                rom.log("Unable to create command at offset " + String(currentOffset))
                break
            }
            childCommands.add(newCommand)

            // increment the offset based on the length of the new command
            currentOffset += newCommand.assembledLength
            
            // set the encoding for the next command
            currentEncodingName = newCommand.encodingNameForNextCommand
        }
    }
    
    func insertCommand(_ command: ROMCommand, after previousCommand: ROMCommand?, child: Bool = false) {
        insertCommands(NSArray(object: command), after: previousCommand, child: child)
    }

    func insertCommands(_ commands: NSArray, after previousCommand: ROMCommand?, child: Bool = false) {
        
        // get the first index
        var index = 0
        if !child, let previousIndex = previousCommand?.index {
            if previousIndex == NSNotFound { return }
            index = previousIndex + 1
        }

        // add the commands to the parent
        let parent = child ? previousCommand : previousCommand?.parent
        let siblings = parent?.childCommands ?? self.childCommands
        siblings.insert(commands as [AnyObject], at: IndexSet(integersIn: Range(NSMakeRange(index, commands.count)) ?? 0..<1))
        
        for object in commands {
            guard let command = object as? ROMCommand else { continue }
            command.parent = parent
            addCommand(command)
        }
        updateOffsets()
    }
    
    func addCommand(_ command: ROMCommand) {
        
        // get an unused reference if the command is coming from a different script
        if command.script !== self {
            command.script = self
            command.reference = ROMCommand.nextUnusedReference
            ROMCommand.nextUnusedReference -= 1
        }
        commandReferences[command.reference] = command

        // recursively add child commands
        for object in command.childCommands {
            guard let command = object as? ROMCommand else { continue }
            addCommand(command)
        }
    }
    
    func removeCommand(_ command: ROMCommand) {
        removeCommands(NSArray(object: command))
    }

    func removeCommands(_ commands: NSArray) {
        guard let firstCommand = commands.firstObject as? ROMCommand else { return }
        let childCommands = firstCommand.parent?.childCommands ?? self.childCommands
        childCommands.removeObjects(at: indexesOfCommands(commands))
        for object in commands {
            guard let command = object as? ROMCommand else { continue }
            commandReferences[command.reference] = nil
        }
        updateOffsets()
    }
    
    func indexesOfCommands(_ commands: NSArray) -> IndexSet {
        
        // this version is slower, but it allows non-contiguous indexes
//        let indexes = NSMutableIndexSet()
//        if commands.count == 0 { return indexes }
//        guard let firstCommand = commands[0] as? ROMCommand else { return indexes }
//        
//        for object in commands {
//            guard let command = object as? ROMCommand else { continue }
//            if command.parent != firstCommand.parent { continue }
//            indexes.addIndex(command.index)
//        }
//        return indexes
        
        // this version is faster but it assumes that all commands are contiguous
        if commands.count == 0 { return IndexSet() }
        guard let firstCommand = commands.firstObject as? ROMCommand else { return IndexSet() }
        
        return IndexSet(integersIn: Range(NSMakeRange(firstCommand.index, commands.count)) ?? 0..<1)
    }

    @discardableResult func addReference(forAssembly assembly: ROMAssembly, encodingName: String, linkedKeyPath: String) -> ROMReference? {
        guard let relativeOffset = assembly.value(forKey: linkedKeyPath) as? Int else { return nil }
        return addReference(forAssembly: assembly, encodingName: encodingName, linkedKeyPath: linkedKeyPath, atAbsoluteOffset: relativeOffset + pointerOffset)
    }
    
    @discardableResult func addReference(forAssembly assembly: ROMAssembly, encodingName: String, linkedKeyPath: String, atAbsoluteOffset offset: Int) -> ROMReference? {

        let referenceDefinition: NSDictionary =
            ["encodingName" : encodingName,
             "linkedObject" : assembly,
             "linkedKeyPath" : linkedKeyPath,
             "isMapped" : true]
        let reference = ROMReference(rom, definition: referenceDefinition)
        if assembly.references == nil { assembly.references = [String : ROMReference]() }
        assembly.references![linkedKeyPath] = reference
        
        if let referencedCommand = commandReferences[offset] as? ROMCommand {
            // if the referenced command already exists, point the reference at it
            reference.referencedAssembly = referencedCommand
        } else {
            // create the array of placeholders if it doesn't already exist
            let placeholdersForThisOffset = commandReferences[offset] as? NSMutableArray ?? NSMutableArray()
            placeholdersForThisOffset.add(reference)
            commandReferences[offset] = placeholdersForThisOffset
        }
        
        return reference
    }
    
    override func relocate(to location: Int) {
        super.relocate(to: location)
        if pointerOffset == 0 { return }
        pointerOffset = location - (location % pointerAlign)
        updateOffsets()
    }

    override func updateReferences() {
        super.updateReferences()

        // update references for all child commands
        for object in childCommands {
            guard let childCommand = object as? ROMCommand else { continue }
            childCommand.updateReferences()
        }

        // do any encoding-specific updates
        encoding?.updateReferencesForScript(self)
    }

    @discardableResult func updateOffsets(_ parent: ROMCommand? = nil, offset: Int? = nil) -> Int {
        // update the range for each command
        var currentOffset = offset ?? range?.lowerBound ?? 0
        for object in parent?.childCommands ?? self.childCommands {
            // recursively add the length of all child commands
            guard let command = object as? ROMCommand else { continue }
            command.range = currentOffset..<(currentOffset + command.data.count)
            currentOffset = updateOffsets(command, offset: currentOffset + command.data.count)
        }
        return currentOffset
    }
}

class ROMScriptEncoding: ROMObject {
    
    let commandTree = NSMutableDictionary()
    let commandDefinitions = NSMutableDictionary()
    
    required override init(_ rom: ROM, definition: NSDictionary? = nil) {
        
        super.init(rom, definition: definition)
        
        if definition == nil { return }
        
        for (key, value) in definition! {
            
            guard let commandDefinition = value as? NSMutableDictionary,
                let identifier = key as? String,
                let opcodeDefinitions = commandDefinition["opcodes"] as? NSArray else { continue }
            
            // get the property definitions (or create a blank one if there isn't one
            let properties = commandDefinition["properties"] as? NSMutableDictionary ?? NSMutableDictionary()

            // add all of the opcodes to an array
            let opcodes = NSMutableArray()
            for opcodeString in opcodeDefinitions {
                if let opcodeRange = MakeRange(object: opcodeString as AnyObject?) {
                    for opcode in opcodeRange.lowerBound..<opcodeRange.upperBound {
                        opcodes.add(opcode)
                    }
                    
                } else if let opcode = Int(object: opcodeString as AnyObject?) {
                    opcodes.add(opcode)
                }
            }
            
            // continue if there were no opcodes
            if opcodes.count == 0 { continue }

            // add a property definition for "opcode" (the first byte)
            let opcodeDefinition = NSMutableDictionary()
            opcodeDefinition["byteIndex"] = 0
            if let firstOpcode = opcodes[0] as? Int, firstOpcode <= 0xFF {
                opcodeDefinition["bitMask"] = 0xFF
            } else {
                opcodeDefinition["bitMask"] = 0xFFFF
            }
            properties["opcode"] = opcodeDefinition
            commandDefinition["properties"] = properties
            commandDefinition["identifier"] = identifier
            
            // add the definition to the table, keyed to each opcode
            for object in opcodes {
                guard let opcode = object as? Int else { continue }
                commandDefinitions[opcode] = commandDefinition
            }
            
            // also add the definition keyed to the identifier
            commandDefinitions[identifier] = commandDefinition
            
            // add the first opcode to the command tree
            if let name = commandDefinition["name"] as? String {
                if let category = commandDefinition["category"] as? String {
                    // create a category tree if one doesn't already exist
                    let categoryTree = commandTree[category] as? NSMutableDictionary ?? NSMutableDictionary()
                    categoryTree[name] = opcodes.firstObject as? Int
                    commandTree[category] = categoryTree
                } else {
                    // command with no category
                    commandTree[name] = opcodes.firstObject as? Int
                }
            }
        }
    }
    
    override func createDefinition() -> NSMutableDictionary {
        let scriptDefinition = super.createDefinition()
        scriptDefinition["type"] = "ROMScriptEncoding"
        
        for (key, value) in commandDefinitions {
            guard let opcode = key as? Int,
                let definition = value as? NSDictionary,
                let identifier = definition["identifier"] as? String else { continue }
            
            if let existingDefinition = scriptDefinition[identifier] as? NSMutableDictionary,
                var opcodeTable = existingDefinition["opcodes"] as? [AnyObject] {
                opcodeTable.append(opcode as AnyObject)
                opcodeTable.sort{ $0.isLessThan($1) }
                existingDefinition["opcodes"] = opcodeTable
                scriptDefinition[identifier] = existingDefinition
                continue
            }
            let newDefinition = NSMutableDictionary()
            newDefinition["opcodes"] = [opcode]
            newDefinition["length"] = definition["length"]
            newDefinition["name"] = definition["name"]
            newDefinition["category"] = definition["category"]
            newDefinition["nib"] = definition["nib"]
            if let properties = definition["properties"] as? NSMutableDictionary {
                let newProperties = NSMutableDictionary(dictionary: properties)
                newProperties.removeObject(forKey: "opcode")
                if newProperties.count != 0 {
                    newDefinition["properties"] = newProperties
                }
            }
            scriptDefinition[identifier] = newDefinition
        }
        
        return scriptDefinition
    }
    
    subscript(key: NSCopying) -> AnyObject? {
        return commandDefinitions[key] as AnyObject?
    }
    
    func updateReferencesForScript(_ script: ROMScript) { }
    func updateReferencesForCommand(_ command: ROMCommand) { }
    
    func valueForCommand(_ command: ROMCommand, forKey key: String) -> Any? {
        
        if let reference = command.references?[key] {
            return reference
            
        } else if let propertyDefinition = command.properties?[key] as? NSDictionary {
            
            let property = ROMProperty(command.rom, definition: propertyDefinition)
            property.disassembleFrom(command.data)
            return property.property
        }
        return nil
    }
    
    func setValue(_ value: Any?, forCommand command: ROMCommand, forKey key: String) {
        
        if let referenceValue = value as? ROMReference {
            if command.references == nil { command.references = [String : ROMReference]() }
            command.willChangeValue(forKey: key)
            command.references?[key] = referenceValue
            command.didChangeValue(forKey: key)
            
            return
            
        } else if let propertyDefinition = command.properties?[key] as? NSDictionary,
            let intValue = value as? Int {
            
            let property = ROMProperty(command.rom, definition: propertyDefinition)
            property.disassembleFrom(command.data)
            
            // return if the value didn't change
            if property.property == intValue { return }

            command.willChangeValue(forKey: key)
            property.property = intValue
            property.assembleTo(&command.data)
            command.didChangeValue(forKey: key)
            
            return

        } else if let propertyDefinition = command.properties?[key] as? NSDictionary,
            let boolValue = value as? Bool {
            
            let property = ROMProperty(command.rom, definition: propertyDefinition)
            property.disassembleFrom(command.data)
            
            // return if the value didn't change
            if property.property == (boolValue ? 1 : 0) { return }
            
            command.willChangeValue(forKey: key)
            property.property = boolValue ? 1 : 0
            property.assembleTo(&command.data)
            command.didChangeValue(forKey: key)
            
            return

        }
    }

    func willAssembleScript(_ script: ROMScript) { }
    func willDisassembleScript(_ script: ROMScript) { }
//    func didCreateBlankCommand(command: ROMCommand) { } // i moved this stuff to didDisassembleCommand:fromData
    func willAssembleCommand(_ command: ROMCommand) { }
    func didAssembleCommand(_ command: ROMCommand) { }
    func willDisassembleCommand(_ command: ROMCommand, fromData data: Data) { }
    func didDisassembleCommand(_ command: ROMCommand, fromData data: Data) { }
    func encodingNameForCommandAfter(_ command: ROMCommand) -> String { return command.encodingName }
    func encodingNameForChildrenOfCommand(_ command: ROMCommand) -> String { return command.encodingName }
    func numLinesForCommandDescription(_ command: ROMCommand) -> Int { return 1 }
    func descriptionForCommand(_ command: ROMCommand) -> String { return command.identifier! }
}
